#!/bin/bash
# copies the final products of the current project into the repository.

# check whether a failure should prevent promotion from occurring.
if [ -f $FAILURE_FILE ]; then
  echo Postconditions will not promote due to a failure:
  cat $FAILURE_FILE
  . $CLAM_DIR/exit_make.sh
fi

# make sure that we actually did something during the make.
if [ ! -f "$DIRTY_FILE" -a ! -f "$SUBMAKE_FLAG" ]; then
  # nothing was built, seemingly, so we do no promotion.
  exit
fi

# toss the flag files so we don't see them again.
rm -f "$DIRTY_FILE" "$SUBMAKE_FLAG"

# clean up generated resource files after the build.
for i in *.resources; do rm -f "$i"; done

# these variables define the locations for final products.  all of them
# descend from the targets directory.
ROOT=$TARGETS_DIR
INCLUDE_DIR=$ROOT/include
HEADER_DIR=$INCLUDE_DIR/$PROJECT
LIB_DIR=$ROOT/lib
DLL_DIR=$ROOT/dll
EXE_DIR=$ROOT/exe
TEST_ROOT=$ROOT/tests
TEST_DIR=$TEST_ROOT/$PROJECT
#PRODUCT_DIR=$FINAL_DIR

# filters out extraneous messages from "cp".
FILT="bash $CLAM_DIR/copy_filter.sh"

# causes the shell to quit.
DIE=". $CLAM_DIR/exit_make.sh"

if [ "$TYPE" = "library" ]; then

  # make sure the required directories exist.
  if [ ! -d $ROOT ]; then mkdir $ROOT; fi
  if [ ! -d $INCLUDE_DIR ]; then mkdir $INCLUDE_DIR; fi
  if [ ! -d $LIB_DIR ]; then mkdir $LIB_DIR; fi
  if [ ! -d $DLL_DIR ]; then mkdir $DLL_DIR; fi

  # copy headers...
  echo Copying headers to $HEADER_DIR.

  # first whack anything that's already there.
  chmod 777 $HEADER_DIR/* 2>&1 | bash $CLAM_DIR/toss_input.sh
    # make files writable.
  rm -rf $HEADER_DIR/*  # toss existing headers and such.
  # recreate include directory if needed.
  if [ ! -d $HEADER_DIR ]; then mkdir $HEADER_DIR; fi

  # then copy the headers from here into there.
  $FILT *.h $HEADER_DIR || $DIE
  $FILT *.rh $HEADER_DIR || $DIE

  if [ ! -z "$EXTRA_COPIES" ]; then
    echo Copying extra files to $HEADER_DIR.
    echo [$EXTRA_COPIES]
    $FILT $EXTRA_COPIES $HEADER_DIR || $DIE
  fi
  
  # turn off writability in the new header directory.
  chmod 555 $HEADER_DIR/*
  
  if [ -z "$NO_COMPILE" ]; then
    # we ensure that none of the normal products are copied for a non-compiling
    # style of project.

    # copy libs & dlls...
#    echo Copying static libraries to $LIB_DIR.
#    $FILT $PRODUCT_DIR/*.lib $LIB_DIR || $DIE
#    $FILT $PRODUCT_DIR/*.a $LIB_DIR || $DIE
#    # this bit ensures that static libraries get their program database files
#    # (so that debugging is enabled) and their map files.
#    if [ ! -z "`echo $TARGETS | grep \.lib`" ]; then
#      $FILT $PRODUCT_DIR/*.map $LIB_DIR || $DIE
#    fi

    # copy the import libraries for any DLLs.
    if [ ! -z "`find "$DYNAMIC_LIBRARY_DIR" -iname "*.lib"`" ]; then
      echo Moving import libraries to $LIB_DIR.
      mv "$DYNAMIC_LIBRARY_DIR"/*.lib $LIB_DIR
    fi

  fi

elif [ "$TYPE" = "application" ]; then

  # sets up the directory for executable programs and copies any found in the
  # this project's final directory.

  # first make sure the executable directory is around.
  if [ ! -d $EXE_DIR ]; then mkdir $EXE_DIR; fi

  if [ -z "$NO_COMPILE" ]; then
    # we ensure that none of the normal products are copied for a non-compiling
    # style of project.

    # copy anything extra over.
    if [ ! -z "$EXTRA_COPIES" ]; then
      echo Copying extra files to $EXE_DIR.
      echo [$EXTRA_COPIES]
      $FILT $EXTRA_COPIES $EXE_DIR || $DIE
    fi

  fi

elif [ "$TYPE" = "test" ]; then

  # sets up a directory for test programs based on the project name and copies
  # the generated programs into there.

  # first make sure the test program root directory is around.
  if [ ! -d $TEST_ROOT ]; then mkdir $TEST_ROOT; fi

  # create the target directory if it doesn't exist.
  if [ ! -d $TEST_DIR ]; then mkdir $TEST_DIR; fi

  if [ -z "$NO_COMPILE" ]; then
    # we ensure that none of the normal products are copied for a non-compiling
    # style of project.

    # make the files writable.  this is required for some tests' data files,
    # which come in from the build and could be read-only.
    chmod 777 $TEST_DIR/* $TEST_DIR/*/* $TEST_DIR/*/*/* 2>&1 | bash $CLAM_DIR/toss_input.sh

    # copy anything extra over.
    if [ ! -z "$EXTRA_COPIES" ]; then
      echo Copying extra files to $TEST_DIR.
      echo [$EXTRA_COPIES]
      $FILT $EXTRA_COPIES $TEST_DIR || $DIE
    fi

  fi

elif [ "$TYPE" = "hierarchy" ]; then

  # do nothing.
  echo "Done with $PROJECT";

else
  echo "Unknown type for project [$TYPE]; cancelling postconditions!"
  . $CLAM_DIR/exit_make.sh
fi

