#ifndef DEFINITIONS_GROUP
#define DEFINITIONS_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : definitions                                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "outcome.h"

#include <ctype.h>
#include <new>

//! Useful constants and data types used throughout HOOPLE.
/*!
  This defines a set of useful universal constants (for our chosen universe)
  and a set of various aliases for well-known abstract and concrete data
  types.
*/

///////////////////////////////  Constants  ///////////////////////////////////

//! This global function is available to the system at large for branding info.
const char *software_product_name();

#ifndef NIL
  //! The value representing a pointer to nothing, or nothing itself.
  #define NIL 0
#endif

#ifndef NULL
  //! See NIL.
  #define NULL 0
#endif

#ifndef BITS_PER_BYTE
  //! A fundamental constant measuring the number of bits in a byte.
  #define BITS_PER_BYTE 8
#endif

#ifndef PI_APPROX 
  //! An approximation of the fundamental circular constant.
  #define PI_APPROX 3.14159265358
#endif

//////////////////  Data Structures & Functions  //////////////////////////

#ifndef formal
  //! This macro just eats what it's passed; it marks unused formal parameters.
  #define formal(parameter)
#endif

#ifndef byte
  //! A fairly important unit which is seldom defined...
  typedef unsigned char byte;
#endif

#if defined(UNICODE) && defined(__WIN32__)
  //! the flexichar type is always appropriate to hold data for win32 calls.
  typedef wchar_t flexichar;
#else
  // this version simply defangs any conversions.
  typedef char flexichar;
#endif

#ifndef u_int
  //! Abbreviated name for unsigned integers.
  typedef unsigned int u_int;
#endif
#ifndef u_short
  //! Abbreviated name for unsigned short integers.
  typedef unsigned short u_short;
#endif
#ifndef u_long
  //! Abbreviated name for unsigned long integers.
  typedef unsigned long u_long;
#endif

// we define some maximum and minimum values that are helpful.  note that this
// assumes a platform where the "int" type is 32 bits.  also note that this no
// longer supports antiquated platforms like windows 3.x or pc dos.
#ifndef MAXINT
  //! Maximum normal integer value.
  #define MAXINT 0x7fffffff
#endif
#ifndef MININT
  //! Minimum normal integer value.
  #define MININT 0x80000000
#endif
#ifndef MAXLONG
  //! Maximum long integer value.
  #define MAXLONG 0x7fffffff
#endif
#ifndef MINLONG
  //! Minimum long integer value.
  #define MINLONG 0x80000000
#endif
#ifndef MAXCHAR
  //! Maximum byte-based character value.
  #define MAXCHAR 0x7f
#endif
#ifndef MINCHAR
  //! Minimum byte-based character value.
  #define MINCHAR 0x80
#endif
#ifndef MAXSHORT
  //! Maximum short integer value.
  #define MAXSHORT 0x7fff
#endif
#ifndef MINSHORT
  //! Minimum short integer value.
  #define MINSHORT 0x8000
#endif
#ifndef MAXBYTE
  //! Maximum unsigned byte value.
  #define MAXBYTE 0xff
#endif
#ifndef MINBYTE
  //! Minimum unsigned byte value.
  #define MINBYTE 0x00
#endif

#ifdef __BCPLUSPLUS__
  // borland provides the types that we otherwise would define below.
  #include <classlib/defs.h>
  #include <systypes.h>
#else
  // Provide definitions for integers with platform independent specific sizes.
  // Note that these will have to be adjusted for 64 bit platforms.
  typedef char int8;
  typedef unsigned char uint8;
  typedef signed short int16;
  typedef unsigned short uint16;
  typedef signed long int32;
  typedef unsigned long uint32;
#endif

////////////////////////////////////////////////////////////////////////////

// useful time constants.

// the _ms suffix indicates that these are measured in milliseconds.
const int SECOND_ms = 1000;  //!< Number of milliseconds in a second.
const int MINUTE_ms = 60 * SECOND_ms;  //!< Number of milliseconds in a minute.
const int HOUR_ms = 60 * MINUTE_ms;  //!< Number of milliseconds in an hour.
const int DAY_ms = 24 * HOUR_ms;  //!< Number of milliseconds in a day.

// the _s suffix indicates that these are measured in seconds.
const int MINUTE_s = 60;  //!< Number of seconds in a minute.
const int HOUR_s = 60 * MINUTE_s;  //!< Number of seconds in an hour.
const int DAY_s = 24 * HOUR_s;  //!< Number of seconds in a day.

// useful general constants.

const int KILOBYTE = 1024;  //!< Number of bytes in a kilobyte.
const int MEGABYTE = KILOBYTE * KILOBYTE;  //!< Number of bytes in a megabyte.
const int GIGABYTE = MEGABYTE * KILOBYTE;  //!< Number of bytes in a gigabyte.
inline double TERABYTE() { return double(GIGABYTE) * double(KILOBYTE); }
  //!< Number of bytes in a terabyte.
  /*!< Implemented as a function to avoid annoying link errors for double
  floating point constants in some compilers. */

////////////////////////////////////////////////////////////////////////////

// these forwards are so commonly used that they have made it into being
// part of the HOOPLE API.  they are all defined in basis headers.

class byte_array;
class byte_sequence;
class chaos;
class int_array;
class int_set;
class istring;
class istring_object;
class mutex;
class string_array;
class unichar_sequence;

////////////////////////////////////////////////////////////////////////////

//! Macro for defining an auto-generated logging filter value.
/*! This macro defines a filter value that will automatically be generated
as part of the build.  These are used for logging using the log_base
object. */

#define DEFINE_FILTER(NAME, CURRENT_VALUE, INFO_STRING) \
  NAME = CURRENT_VALUE

////////////////////////////////////////////////////////////////////////////

//! Macro for defining an auto-generated event value.
/*! The event macro allows an auto-generated value to be associated with
different events that can occur.  These are different than outcomes since
they may arise due to factors other than function invocation.  Outcomes occur
because functions can return different classes of result values, but events
can be generated by external reality and injected into a program. */

#define DEFINE_EVENT(NAME, CURRENT_VALUE, INFO_STRING) \
  NAME = CURRENT_VALUE

////////////////////////////////////////////////////////////////////////////

//! the "common" structure gathers some popular elements of lower-level classes.

struct common {

  //! these outcomes are returned by classes that use the basic HOOPLE support.
  /*! More complicated classes will need to define their own outcome values to
  describe what occurred during the processing of requests. */
  enum outcomes {
    DEFINE_OUTCOME(OKAY, 0, "Everything is just fine"),
    DEFINE_OUTCOME(NOT_IMPLEMENTED, -1, "The invoked method is unimplemented"),
    DEFINE_OUTCOME(OUT_OF_RANGE, -2, "The value specified was out of bounds"),
    DEFINE_OUTCOME(NOT_FOUND, -3, "The item sought is not present"),
    DEFINE_OUTCOME(BAD_INPUT, -4, "Precondition failure--the parameters were "
        "inappropriate"),
    DEFINE_OUTCOME(BAD_TYPE, -5, "The objects are of incompatible types"),
    DEFINE_OUTCOME(IS_FULL, -6, "There is no room in the storage facility"),
    DEFINE_OUTCOME(IS_EMPTY, -7, "The container is empty currently"),
    DEFINE_OUTCOME(IS_NEW, -8, "The item is new"),
    DEFINE_OUTCOME(EXISTING, -9, "The item was already present"),
    DEFINE_OUTCOME(FAILURE, -10, "A failure has occurred"),
    DEFINE_OUTCOME(OUT_OF_MEMORY, -11, "There is not enough memory for the "
        "request according to the operating system"),
    DEFINE_OUTCOME(ACCESS_DENIED, -12, "The request was denied, possibly by "
        "the operating system"),
    DEFINE_OUTCOME(IN_USE, -13, "The object is already in exclusive use"),
    DEFINE_OUTCOME(UNINITIALIZED, -14, "The object has not been constructed "
        "properly"),
    DEFINE_OUTCOME(TIMED_OUT, -15, "The allowed time has now elapsed"),
    DEFINE_OUTCOME(GARBAGE, -16, "The request or response has been corrupted"),
    DEFINE_OUTCOME(NO_SPACE, -17, "A programmatic limit on storage space "
        "has been reached"),
    DEFINE_OUTCOME(DISALLOWED, -18, "The method denied the request"),
    DEFINE_OUTCOME(INCOMPLETE, -19, "The operation did not finish or the "
        "object is not completed"),
    DEFINE_OUTCOME(NO_HANDLER, -20, "The object type passed in was not "
        "understood by the invoked method"),
    DEFINE_OUTCOME(NONE_READY, -21, "There were no objects available"),
    DEFINE_OUTCOME(INVALID, -22, "That request or object was invalid"),
    DEFINE_OUTCOME(PARTIAL, -23, "The request was only partially finished"),
    DEFINE_OUTCOME(NO_LICENSE, -24, "The software license does not permit"
        "this request"),
    DEFINE_OUTCOME(UNEXPECTED, -25, "This item was unexpected, although "
        "not necessarily erroneous"),
    DEFINE_OUTCOME(ENCRYPTION_MISMATCH, -26, "The request failed due to a "
        "mismatch between encryption expected and encryption provided"),
  };

  static const char *outcome_name(const outcome &to_name);
    //!< Returns a string representation of the outcome "to_name".
    /*!< If "to_name" is unknown, because it's not a member of the
    common::outcomes enum, then a string reporting that is returned. */

  /*! the list positions specify well known places in a linked list, or indeed
  in any structure that has a beginning and an end. */
  enum list_positions {
    HEAD = -298,   //!< the beginning element in the list.
    TAIL,          //!< the final element in the list.
    MIDDLE         //!< an indeterminate place that is neither head nor tail.
  };

  enum how_to_copy { NEW_AT_END, NEW_AT_BEGINNING, DONT_COPY }; 
    //!< An enumeration of ways to shift existing contents in an array.
    /*!< In the context of dynamically resizing an array of data, a range of
    new elements can be added (or removed for that matter) at the end of the
    existing space (NEW_AT_END), at the beginning of the existing space
    (NEW_AT_BEGINNING), or we might not care about the existing data when
    we perform the resize (DONT_COPY).  These methods impose a particular
    shift on the existing contents if we actually already have enough space
    for the new contents.  If there is space before the part of the array
    that's in use and we want NEW_AT_END, then the existing contents are
    jammed up into the front end of the array. */

  //! These are the lowest-level filters defined for logging.
  enum filters {
    DEFINE_FILTER(NEVER_PRINT, -1, "Indicates that the diagnostic entry "
        "should be dropped"),
    DEFINE_FILTER(ALWAYS_PRINT, 0, "Indicates that the diagnostic entry "
        "should always be recorded"),
    DEFINE_FILTER(NETWORK_LOGGING, 1, "Extra logging of network events "
        "will be performed"),
    DEFINE_FILTER(UNUSUAL_LOGGING, 2, "Logging of unusual but not "
        "necessarily erroneous events"),

    // these must be updated if the above ordering or contents change.
    FIRST_LOW_LEVEL_FILTER = NETWORK_LOGGING,
    LAST_LOW_LEVEL_FILTER = UNUSUAL_LOGGING,
  };

  static int_set low_level_filters();
    //!< A set that contains all of the low-level filter values.
};

////////////////////////////////////////////////////////////////////////////

// this declares the program-wide argument storage area.  these hang onto argc
// and argv if a program uses our macros for main.  visual c++ provides these
// (unless the program is linked erroneously, mixing statically and dynamically
// linked libraries), but we define them for unix & linux.  the HOOPLE_MAIN
// macro (see opsystem/startup_code) sets them from the initial program
// arguments, although you can do that manually also.

#ifdef __UNIX__
  extern int __argc;
  extern char **__argv;
  //! this macro allocates space for the command-line storage areas.
  #define DEFINE_ARGC_AND_ARGV int __argc = 0; char **__argv = NIL
#elif defined(__WIN32__)
  #include <stdlib.h>
    // the macros are defined in a fairly hard to untangle way in this header.
  #define DEFINE_ARGC_AND_ARGV 
    // the allocation of these two is already provided by ms-c.
#else
  // guessing this will be okay for other platforms.
  extern int __argc;
  extern char **__argv;
  #define DEFINE_ARGC_AND_ARGV int __argc = 0; char **__argv = NIL
#endif

////////////////////////////////////////////////////////////////////////////

// compiler specific sections lie ahead.

#if defined(__WIN32__) || defined(_MSC_VER)
  // general windows defs.

  // allows older-style borland windows flags to work with mfc.
  #ifndef WIN32
    #define WIN32
  #endif
  #ifndef __WIN32__
    #define __WIN32__
  #endif
  #ifndef _WIN32
    #define _WIN32
  #endif
  #ifndef _Windows
    #define _Windows
  #endif
  #ifndef _WINDOWS
    #define _WINDOWS
  #endif

  #ifdef _MSC_VER
    // support for ms visual c++ compiler.

    // this special block is required for debugging builds to operate correctly.
    #ifdef _AFXDLL
      #ifdef _DEBUG
        // this code is needed for the debugging version of everything, since
        // MFC allocations will barf if mixed with regular allocations.
        #include <afx.h>
        #define new DEBUG_NEW
          // redefine the new operator for debugging.
      #endif
  
      // NOTE: if you are having compilation problems related to a line in
      // a header that has "new" in it, try modifying the file where the
      // problem occurs by putting this after all the #include statements:
      //   #ifdef new
      //     #undef new
      //     #define NEW_TRAP
      //   #endif
      // and by putting this at the end of the file:
      //   #ifdef NEW_TRAP
      //     #undef NEW_TRAP
      //     #define new DEBUG_NEW
      //   #endif
      // this will turn off the DEBUG_NEW macro for that particular file.
    #endif
  
    // turns off annoying complaints from visual c++.
    #pragma warning(disable : 4251 4275 4003 4800 4355 4786 4290 4996 4407)
    #pragma warning(error : 4172)
      // 4251 and 4275 turn off warnings regarding statically linked code
      //    not being marked with dll import/export flags.
      // 4003 turns off warnings about insufficient number of parameters passed
      //    to a macro.
      // 4800 turns off the warning about conversion from int to bool not being
      //    efficient.
      // 4355 turns off the warning re 'this' used in base member init list.
      // 4786 turns off the warning about 'identifier' was truncated to 'number' 
      //    characters in the debug information which frequenly happens when 
      //    STL pair and set templates are expanded.
      // 4172 is made an error because this warning is emitted for a dangerous
      //    condition; the address of a local variable is being returned, making
      //    the returned object junk in almost all cases.

    #ifdef _M_IX86
      #pragma comment(linker,"/manifestdependency:\"type='win32' "   \
        "name='Microsoft.Windows.Common-Controls' "                \
        "version='6.0.0.0' "                                       \
        "processorArchitecture='x86' "                             \
        "publicKeyToken='6595b64144ccf1df' "                       \
        "language='*'\"")
    #endif

    #ifdef _M_AMD64
       #pragma comment(linker,"/manifestdependency:\"type='win32' "   \
        "name='Microsoft.Windows.Common-Controls' "                \
        "version='6.0.0.0' "                                       \
        "processorArchitecture='amd64' "                           \
        "publicKeyToken='6595b64144ccf1df' "                       \
        "language='*'\"")
    #endif

    #ifdef _M_IA64
       #pragma comment(linker,"/manifestdependency:\"type='win32' "   \
        "name='Microsoft.Windows.Common-Controls' "                \
        "version='6.0.0.0' "                                       \
        "processorArchitecture='ia64' "                            \
        "publicKeyToken='6595b64144ccf1df' "                       \
        "language='*'\"")
    #endif

  #endif  // visual c++.
  
#elif defined(__UNIX__)
  #define LOBYTE(to_chop) (byte(to_chop))
    //!< similar to win32 LOBYTE definition used by some code.
  #define HIBYTE(to_chop) (byte(0xff & (u_short(to_chop) >> 8)))
    //!< similar to win32 HIBYTE definition used by some code.
#else
  // otherwise we know nothing about the compiler environment.
  // we'll go ahead and define a bool type in case the compiler hasn't.
  typedef int bool;  // simple bool type for compilers without it.
  enum _boolean_enumeration_ { false, true };
#endif

// fancy footwork supporting external placement of the logging functions
// and other items that must live in the i_library dll rather than a static
// library.  we need those objects to have a single instance in the program.
#ifdef BUILD_OPERATING_SYSTEM
  #define PWL_EXPO HOOPLE_DLL_EXPORT_FUNCTION
#elif defined(USE_HOOPLE_DLLS)
  #define PWL_EXPO HOOPLE_DLL_IMPORT_FUNCTION
#else
  #define PWL_EXPO
#endif

////////////////////////////////////////////////////////////////////////////

#ifndef OMIT_MEMORY_HOOK
  #ifdef __DEBUGGING__
    // overrides for the global new and delete operators when pooling memory.

    void *operator new(size_t size) throw (std::bad_alloc);
    inline void* operator new [] (size_t size) throw (std::bad_alloc)
        { return ::operator new(size); }

    void operator delete(void *ptr);
    inline void operator delete [] (void *ptr) { ::operator delete(ptr); }
  #endif // debugging.
#endif // memory hook.

#endif // outer guard.

