/*****************************************************************************\
*                                                                             *
*  Name   : earth_time                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1999-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "convert_utf.h"
#include "earth_time.h"
#include "istring.h"
#include "mutex.h"
#include "portable.h"
#include "utility.h"

#include <time.h>
#if defined(__WIN32__) || defined(__UNIX__)
  #include <sys/timeb.h>
#endif

using namespace basis;

namespace earth_time {

const int days_in_month[12]
    = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

const int leap_days_in_month[12]
    = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

const int julian_days_in_month[12]
    = { 31, 29, 31, 30, 31, 30, 31, 30, 31, 30, 31, 30 };
//hmmm: is this right?

const int julian_leap_days_in_month[12]
    = { 31, 30, 31, 30, 31, 30, 31, 30, 31, 30, 31, 30 };

////////////////////////////////////////////////////////////////////////////

void clock_time::pack(byte_array &packed_form) const
{
  attach(packed_form, hour);
  attach(packed_form, minute);
  attach(packed_form, second);
  attach(packed_form, millisecond);
  attach(packed_form, microsecond);
}

bool clock_time::unpack(byte_array &packed_form)
{
  if (!detach(packed_form, hour)) return false;
  if (!detach(packed_form, minute)) return false;
  if (!detach(packed_form, second)) return false;
  if (!detach(packed_form, millisecond)) return false;
  if (!detach(packed_form, microsecond)) return false;
  return true;
}

bool clock_time::operator < (const clock_time &to_compare)
{
  if (hour < to_compare.hour) return true;
  if (minute < to_compare.minute) return true;
  if (second < to_compare.second) return true;
  if (millisecond < to_compare.millisecond) return true;
  if (microsecond < to_compare.microsecond) return true;
  return false;
}

bool clock_time::operator == (const clock_time &to_compare)
{
  return (hour == to_compare.hour) 
      && (minute == to_compare.minute)
      && (second == to_compare.second) 
      && (millisecond == to_compare.millisecond) 
      && (microsecond == to_compare.microsecond);
}

istring clock_time::text_form(int how) const
{
  istring to_return;
  text_form(to_return, how);
  return to_return;
}

void clock_time::text_form(istring &to_return, int how) const
{
  if (!how) return;  // enforce use of the default.
  if (how & MILITARY)
    to_return += isprintf("%02d:%02d", hour, minute);
  else {
    int uhr = hour;
    if (uhr > 12) uhr -= 12;
    to_return += isprintf("%2d:%02d", uhr, minute);
  }
  if ( (how & SECONDS) || (how & MILLISECONDS) )
    to_return += isprintf(":%02d", second);
  if (how & MILLISECONDS)
    to_return += isprintf(":%03d", millisecond);
  if (how & MERIDIAN) {
    if (hour >= 12) to_return += "PM";
    else to_return += "AM";
  }
}

////////////////////////////////////////////////////////////////////////////

void day_in_year::pack(byte_array &packed_form) const
{
  attach(packed_form, day_of_year);
  attach(packed_form, byte(day_of_week));
  attach(packed_form, byte(month));
  attach(packed_form, day_in_month);
  attach(packed_form, byte(daylight_saving));
}

bool day_in_year::unpack(byte_array &packed_form)
{
  if (!detach(packed_form, day_of_year)) return false;
  byte temp;
  if (!detach(packed_form, temp)) return false;
  day_of_week = days(temp);
  if (!detach(packed_form, temp)) return false;
  month = months(temp);
  if (!detach(packed_form, day_in_month)) return false;
  if (!detach(packed_form, temp)) return false;
  daylight_saving = bool(temp);
  return true;
}

bool day_in_year::operator < (const day_in_year &to_compare)
{
  if (month < to_compare.month) return true;
  if (day_in_month < to_compare.day_in_month) return true;
  return true;
}

bool day_in_year::operator == (const day_in_year &to_compare)
{
  return (month == to_compare.month)
      && (day_in_month == to_compare.day_in_month);
}

istring day_in_year::text_form(int how) const
{
  istring to_return;
  text_form(to_return, how);
  return to_return;
}

void day_in_year::text_form(istring &to_stuff, int how) const
{
  if (!how) return;  // enforce use of the default.
  if (how & INCLUDE_DAY) to_stuff += istring(day_name(day_of_week)) + " ";
  const char *monat = short_month_name(month);
  if (how & LONG_MONTH) monat = month_name(month);
//hmmm: more formatting, like euro?
  to_stuff += monat;
  to_stuff += isprintf(" %02d", day_in_month);
}

////////////////////////////////////////////////////////////////////////////

void time_locus::pack(byte_array &packed_form) const
{
  attach(packed_form, year);
  clock_time::pack(packed_form);
  day_in_year::pack(packed_form);
}

bool time_locus::unpack(byte_array &packed_form)
{
  if (!detach(packed_form, year)) return false;
  if (!clock_time::unpack(packed_form)) return false;
  if (!day_in_year::unpack(packed_form)) return false;
  return true;
}

istring time_locus::text_form(int t, int d, int y) const
{
  istring to_return;
  text_form(to_return, t, d, y);
  return to_return;
}

bool time_locus::operator < (const time_locus &to_compare)
{
  if (year < to_compare.year) return true;
  if ( (day_in_year &) *this < to_compare) return true;
  if ( (clock_time &) *this < to_compare) return true;
  return false;
}

bool time_locus::operator == (const time_locus &to_compare)
{
  return (year == to_compare.year)
      && ( (day_in_year &) *this == to_compare)
      && ( (clock_time &) *this == to_compare);
}

void time_locus::text_form(istring &to_stuff, int t, int d, int y) const
{
//hmmm: more formatting desired, like european.
  if (!y) {
    text_form(to_stuff, t, d);  // enforce use of the default.
    return;
  }
  // add the day.
  day_in_year::text_form(to_stuff, d);
  to_stuff += " ";
  // add the year.
  if (y & SHORT_YEAR)
    to_stuff += isprintf("%2d", year % 100);
  else
    to_stuff += isprintf("%4d", year);
  // add the time.
  to_stuff += " ";
  clock_time::text_form(to_stuff, t);
}

////////////////////////////////////////////////////////////////////////////

time_locus convert(const tm &to_convert, int ms)
{
  time_locus r;

  // we lack the resolution for this, currently.
  r.microsecond = 0;

  r.second = to_convert.tm_sec;
  r.minute = to_convert.tm_min;
  r.hour = to_convert.tm_hour;
  r.day_in_month = to_convert.tm_mday;
  r.month = months(to_convert.tm_mon);
  r.year = to_convert.tm_year + 1900;
  r.day_of_week = days(to_convert.tm_wday);
  r.day_of_year = to_convert.tm_yday;
  r.daylight_saving = to_convert.tm_isdst;
  r.millisecond = ms;
  return r;
}

time_locus now()
{
  timeb current;
  ftime(&current);
  tm split_time(*localtime(&current.time));
  return convert(split_time, current.millitm);
}

time_locus greenwich_now()
{
  timeb current;
  ftime(&current);
  tm split_time(*gmtime(&current.time));
  return convert(split_time, current.millitm);
}

clock_time time_now() { return now(); }

days day_now() { return now().day_of_week; }

months month_now() { return now().month; }

int year_now() { return now().year; }

day_in_year date_now() { return now(); }

const char *day_name(days to_name)
{
  switch (to_name) {
    case SUNDAY: return "Sunday";
    case MONDAY: return "Monday";
    case TUESDAY: return "Tuesday";
    case WEDNESDAY: return "Wednesday";
    case THURSDAY: return "Thursday";
    case FRIDAY: return "Friday";
    case SATURDAY: return "Saturday";
    default: return "Not_a_day";
  }
}

const char *month_name(months to_name)
{
  switch (to_name) {
    case JANUARY: return "January";
    case FEBRUARY: return "February";
    case MARCH: return "March";
    case APRIL: return "April";
    case MAY: return "May";
    case JUNE: return "June";
    case JULY: return "July";
    case AUGUST: return "August";
    case SEPTEMBER: return "September";
    case OCTOBER: return "October";
    case NOVEMBER: return "November";
    case DECEMBER: return "December";
    default: return "Not_a_month";
  }
}

const char *short_month_name(months to_name)
{
  switch (to_name) {
    case JANUARY: return "Jan";
    case FEBRUARY: return "Feb";
    case MARCH: return "Mar";
    case APRIL: return "Apr";
    case MAY: return "May";
    case JUNE: return "Jun";
    case JULY: return "Jul";
    case AUGUST: return "Aug";
    case SEPTEMBER: return "Sep";
    case OCTOBER: return "Oct";
    case NOVEMBER: return "Nov";
    case DECEMBER: return "Dec";
    default: return "Not";
  }
}

bool set_time(const time_locus &new_time)
{
#ifdef __WIN32__
  SYSTEMTIME os_time;
  os_time.wYear = WORD(new_time.year);
  os_time.wMonth = new_time.month;
  os_time.wDayOfWeek = new_time.day_of_week;
  os_time.wDay = new_time.day_of_year;
  os_time.wHour = new_time.hour;
  os_time.wMinute = new_time.minute;
  os_time.wSecond = new_time.second;
  os_time.wMilliseconds = 0;  // currently unused.

  // get our process token for manipulation.
  HANDLE petoken;
  OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES
      | TOKEN_QUERY, &petoken);
  // get our 
//something or other
// identifier so we can adjust our privileges.
  LUID our_id;
  LookupPrivilegeValue(NULL, to_unicode_temp("SeSystemTimePrivilege"), &our_id);
  // make up a privilege structure for the adjustment.
  TOKEN_PRIVILEGES privs;
  privs.PrivilegeCount = 1;
  privs.Privileges[0].Luid = our_id;
  privs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  // enable system-time privileges.
  AdjustTokenPrivileges(petoken, false, &privs, sizeof(TOKEN_PRIVILEGES),
      NULL, NULL);

  SetLocalTime(&os_time);  // actually set the time.

  // disable the time adjustment privileges again.
  AdjustTokenPrivileges(petoken, true, &privs, sizeof(TOKEN_PRIVILEGES),
      NULL, NULL);

  // let all the main windows know that the time got adjusted.
//do we need to do this ourselves?
  ::PostMessage(HWND_BROADCAST, WM_TIMECHANGE, 0, 0);

//hmmm: make sure this seems right.
  CloseHandle(petoken);

  return true;
#elif defined(__UNIX__)
//no implem yet.
  return false;
#else
  return false;
#endif
}

/*int canonicalize(clock_time &to_fix)
{
}*/

} // namespace.

