/*****************************************************************************\
*                                                                             *
*  Name   : istring                                                           *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "common_templates.h"
#include "function.h"
#include "guards.h"
#include "istring.h"
#include "utility.h"

#ifdef _MSC_VER
  #include <comdef.h>
  #undef strcasecmp 
  #undef strncasecmp 
  #define strcasecmp strcmpi
  #define strncasecmp strnicmp
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#define DEBUG_STRING
  // uncomment for debugging version.

const int LONGEST_SPRINTF = 600;  // the longest a simple sprintf can be here.

const char CASE_DIFFERENCE = char('A' - 'a');
  // the measurement of the difference between upper and lower case.

// this factor is used to bias dynamic sprintfs for cases where the length
// is specified, but the actual string is shorter than that length.
#ifdef EMBEDDED_BUILD
  const int MAX_FIELD_FUDGE_FACTOR = 8;
#else
  const int MAX_FIELD_FUDGE_FACTOR = 64;
#endif

const byte empty_char_star[] = "";
  // used to initialize empty strings.

////////////////////////////////////////////////////////////////////////////

bool istring_comparator(const istring &a, const istring &b) { return a == b; }

////////////////////////////////////////////////////////////////////////////

#define static_class_name() istring("istring")
  // used for the bounds_halt macro.

istring::istring()
: _implementation(new byte_sequence(1, empty_char_star)),
  _held_string((char * const *)_implementation->internal_offset_mem())
 {}

istring::istring(char initial, int repeat)
: _implementation(NIL)
{
  if (!initial) initial = ' ';  // for nulls, we use spaces.
  int new_size = repeat;
  if (non_positive(repeat)) new_size = 0;
  // delayed construction to avoid unnecessary re-allocation.
  _implementation = new byte_sequence(new_size + 1);
  memset(_implementation->access(), initial, new_size);
  _implementation->put(new_size, '\0');
  _held_string = (char * const *)_implementation->internal_offset_mem();
}

istring::istring(const istring &s1)
: _implementation(new byte_sequence(*s1._implementation)),
  _held_string((char * const *)_implementation->internal_offset_mem())
{
}

istring::istring(const char *initial)
: _implementation(NIL)
{
  // we use the size of the string to copy unless the pointer is null.
  int new_size = initial? strlen(initial) : 0;
  _implementation = new byte_sequence(new_size + 1);
  _implementation->put(0, '\0');
  if (!initial) return;  // bail because there's no string to copy.
  strcpy(access(), initial);
  _held_string = (char * const *)_implementation->internal_offset_mem();
}

istring::istring(special_flag flag, const char *initial, ...)
: _implementation(new byte_sequence(1, empty_char_star)),
  _held_string((char * const *)_implementation->internal_offset_mem())
{
  if (!initial) return;
  if ( (flag != UNTERMINATED) && (flag != SPRINTF) ) {
    operator = (istring(istring::SPRINTF, "unknown flag %d", flag));
    return;
  }

  va_list args;
  va_start(args, initial);

  if (flag == UNTERMINATED) {
    // special process for grabbing a string that has no terminating nil.  
    int length = va_arg(args, int);  // get the length of the string out.
    _implementation->reset(length, (byte *)initial);
    (*_implementation) += byte(0);
    va_end(args);
    return;
  }

  // only other flag currently supported is sprintf, so we do that...
  base_sprintf(initial, args);
  va_end(args);
}

istring::~istring() { WHACK(_implementation); _held_string = NIL; }

const istring &istring::empty_string()
{
  static istring __hidden_empty_string;
  return __hidden_empty_string;
}

const byte_sequence &istring::low_level() const { return *_implementation; }

int istring::length() const { return _implementation->length() - 1; }

byte_sequence &istring::get_implem() { return *_implementation; }

char *istring::access() { return (char *)_implementation->access(); }

char istring::get(int index) const { return (char)_implementation->get(index); }

const char *istring::observe() const
{ return (const char *)_implementation->observe(); }

bool istring::operator < (const istring &s2) const
{ return _implementation->operator < (*s2._implementation); }

bool istring::operator > (const istring &s2) const
{ return _implementation->operator > (*s2._implementation); }

bool istring::operator <= (const istring &s2) const
{ return _implementation->operator <= (*s2._implementation); }

bool istring::operator >= (const istring &s2) const
{ return _implementation->operator >= (*s2._implementation); }

bool istring::operator != (const istring &s2) const
{ return !(*this == s2); }

bool istring::operator == (const istring &s2) const
{ return strcmp(observe(), s2.observe()) == 0; }

bool istring::operator == (const char *that) const
{ return strcmp(observe(), that) == 0; }

bool istring::operator != (const char *that) const
{ return strcmp(observe(), that) != 0; }

bool istring::contains(const istring &to_find) const
{ return (find(to_find, 0) < 0) ? false : true; }

istring &istring::operator += (const istring &s1)
{ insert(length(), s1); return *this; }

void istring::shrink()
{
  istring copy_of_this(observe());
  _implementation->swap_contents(*copy_of_this._implementation);
}

istring &istring::sprintf(const char *initial, ...)
{
  va_list args;
  va_start(args, initial);
  istring &to_return = base_sprintf(initial, args);
  va_end(args);
  return to_return;
}

istring &istring::base_sprintf(const char *initial, va_list &args)
{
  reset();
  if (!initial) return *this;  // skip null strings.
  if (!initial[0]) return *this;  // skip empty strings.

  // these accumulate parts of the sprintf format within the loop.
  char flag_chars[23], width_chars[23], precision_chars[23], modifier_chars[23];

  // thanks for the inspiration to k&r page 156.
  for (const char *traverser = initial; *traverser; traverser++) {
#ifdef DEBUG_STRING
    printf("index=%d, char=%c\n", traverser - initial, *traverser);
#endif

    if (*traverser != '%') {
      // not a special character, so just drop it in.
      *this += *traverser;
      continue;
    }
    traverser++; // go to the next character.
#ifdef DEBUG_STRING
    printf("index=%d, char=%c\n", traverser - initial, *traverser);
#endif
    if (*traverser == '%') {
      // capture the "%%" style format specifier.
      *this += *traverser;
      continue;
    }
    bool failure = false;
      // becomes set to true if something didn't match in a necessary area.

    seek_flag(traverser, flag_chars, failure);
    if (failure) {
      *this += '%';
      *this += flag_chars;
      continue;
    }
    seek_width(traverser, width_chars);
    seek_precision(traverser, precision_chars);
    seek_modifier(traverser, modifier_chars);
    get_type_character(traverser, args, *this, flag_chars,
        width_chars, precision_chars, modifier_chars);
  }
  return *this;
}

void istring::seek_flag(const char *&traverser, char *flag_chars, bool &failure)
{
  flag_chars[0] = '\0';
  failure = false;
  bool keep_going = true;
  while (!failure && keep_going) {
    switch (*traverser) {
      case '-': case '+': case ' ': case '\011': case '#':
        flag_chars[strlen(flag_chars) + 1] = '\0';
        flag_chars[strlen(flag_chars)] = *traverser++;
        break;
      default:
        // we found a character that doesn't belong in the flags.
        keep_going = false;
        break;
    }
  }
#ifdef DEBUG_STRING
  if (strlen(flag_chars)) printf("[flag=%s]\n", flag_chars);
  else printf("no flags\n");
#endif
}

void istring::seek_width(const char *&traverser, char *width_chars)
{
  width_chars[0] = '\0';
  bool no_more_nums = false;
  bool first_num = true;
  while (!no_more_nums) {
    char wideness[2] = { *traverser, '\0' };
    if (first_num && (wideness[0] == '0')) {
      strcpy(width_chars, wideness);
      traverser++;
    } else if (first_num && (wideness[0] == '*') ) {
      strcpy(width_chars, wideness);
      traverser++;
      no_more_nums = true;
    } else if ( (wideness[0] <= '9') && (wideness[0] >= '0') ) {
      // a failure?
      strcat(width_chars, wideness);
      traverser++;
    } else no_more_nums = true;
    first_num = false;
  }
#ifdef DEBUG_STRING
  if (strlen(width_chars)) printf("[width=%s]\n", width_chars);
  else printf("no widths\n");
#endif
}

void istring::seek_precision(const char *&traverser, char *precision_chars)
{
  precision_chars[0] = '\0';
  if (*traverser != '.') return;
  strcpy(precision_chars, ".");
  traverser++;
  bool no_more_nums = false;
  bool first_num = true;
  while (!no_more_nums) {
    char preciseness[2] = { *traverser, '\0' };
    if (first_num && (preciseness[0] == '0')) {
      strcat(precision_chars, preciseness);
      traverser++;
      no_more_nums = true;
    } else if (first_num && (preciseness[0] == '*') ) {
      strcat(precision_chars, preciseness);
      traverser++;
      no_more_nums = true;
    } else if ( (preciseness[0] <= '9') && (preciseness[0] >= '0') ) {
      strcat(precision_chars, preciseness);
      traverser++;
    } else no_more_nums = true;
    first_num = false;
  }
#ifdef DEBUG_STRING
  if (strlen(precision_chars)) printf("[precis=%s]\n", precision_chars);
  else printf("no precision\n");
#endif
}

void istring::seek_modifier(const char *&traverser, char *modifier_chars)
{
  modifier_chars[0] = '\0';
  switch (*traverser) {
    case 'F': case 'N': case 'h': case 'l': case 'L': {
        modifier_chars[strlen(modifier_chars) + 1] = '\0';
        modifier_chars[strlen(modifier_chars)] = *traverser++;
      break;
    }
  }
#ifdef DEBUG_STRING
  if (strlen(modifier_chars)) printf("[mod=%s]\n", modifier_chars);
  else printf("no modifiers\n");
#endif
}

void istring::get_type_character(const char * &traverser, va_list &args,
    istring &output_string, const char *flag_chars, const char *width_chars,
    const char *precision_chars, const char *modifier_chars)
{
  char formatting[120];
  strcpy(formatting, "%");
  strcat(formatting, flag_chars);
  strcat(formatting, width_chars);
  strcat(formatting, precision_chars);
  strcat(formatting, modifier_chars);
  char tmposh[2] = { *traverser, '\0' };
  strcat(formatting, tmposh);
#ifdef DEBUG_STRING
  printf("format: %s\n", formatting);
#endif

  enum argument_size { bits_8, bits_16, bits_32, bits_64, bits_80 };
  bool ints_are_32_bits;
#ifdef __WIN32__
  ints_are_32_bits = true;
#elif defined(__OS2__)
  ints_are_32_bits = true;
#elif defined(__MSDOS__)
  ints_are_32_bits = false;
#elif defined(__WIN32__)
  ints_are_32_bits = false;
#else
  ints_are_32_bits = true;
#endif
  argument_size next_argument;
  bool use_dynamic_sprintf = false;  // for dynamic printing of strings only.
  // get the type character first and ensure it's valid.
  switch (*traverser) {
    case 'd': case 'i': case 'o': case 'u': case 'x': case 'X':
      next_argument = bits_16;
      if (ints_are_32_bits) next_argument = bits_32;
      break;
    case 'f': case 'e': case 'g': case 'E': case 'G':
      next_argument = bits_64;
      break;
    case 'c':
      next_argument = bits_8;
      break;
    case 's':
      next_argument = bits_32;
      use_dynamic_sprintf = true;
      break;
    case 'n':
      next_argument = bits_32; //?????
      break;
    case 'p':
      next_argument = bits_32; //????
      break;
    default:
      // this is an error; the character is not recognized, so spew out
      // any characters accumulated so far as just themselves.
#ifdef DEBUG_STRING
      printf("failure in type char: %c\n", *traverser);
#endif
      output_string += formatting;
      return;
  }
/* hmmm: not supported yet.
  if (width_chars && (width_chars[0] == '*')) {
  }
  if (precision_chars && (precision_chars[0] == '*')) {
  }
*/
  if (strlen(modifier_chars)) {
    switch (modifier_chars[0]) {
      case 'N':  // near pointer.
        next_argument = bits_16;
        if (ints_are_32_bits) next_argument = bits_32;
        break;
      case 'F':  // far pointer.
        next_argument = bits_32;
        break;
      case 'h':  // short int.
        next_argument = bits_16;
        break;
      case 'l':  // long.
        next_argument = bits_32;
        break;
      case 'L':  // long double;
        next_argument = bits_80;
        break;
      default:
        // a failure has occurred because the modifier character is not
        // one of the recognized values.  everything is just spewed out.
#ifdef DEBUG_STRING
        printf("failure in modifier: %s\n", modifier_chars);
#endif
        output_string += formatting;
        return;
    }
  }
  // action time: the output string is given a tasty value.
  char temp[LONGEST_SPRINTF];
  char *temp2 = NIL;  // for dynamic only.
  switch (next_argument) {
//hmmm: this switch is where support would need to be added for having two
//      arguments (for the '*' case).
    case bits_8: case bits_16:
      if (ints_are_32_bits) ::sprintf(temp, formatting, va_arg(args, long));
      else ::sprintf(temp, formatting, va_arg(args, int));
      break;
    case bits_32:
      if (use_dynamic_sprintf) {
        // currently we only do dynamic sprintf for strings.
        char *to_print = va_arg(args, char *);
        // check if it's valid and if we really need to do it dynamically.
        if (!to_print) {
          // bogus string; put in a complaint.
          use_dynamic_sprintf = false;
          ::sprintf(temp, "{error:parm=NIL}");
        } else if (strlen(to_print) < LONGEST_SPRINTF - 2) {
          // we're within our bounds, plus some safety room, so just do a
          // regular sprintf.
          use_dynamic_sprintf = false;
          ::sprintf(temp, formatting, to_print);
        } else {
          // it's too long, so we definitely need to do it dynamically.
          temp2 = new char[strlen(to_print) + MAX_FIELD_FUDGE_FACTOR];
          ::sprintf(temp2, formatting, to_print);
        }
      } else ::sprintf(temp, formatting, va_arg(args, void *));
      break;
    case bits_64:
      ::sprintf(temp, formatting, va_arg(args, double));
      break;
    case bits_80:
      ::sprintf(temp, formatting, va_arg(args, long double));
      break;
  }
  if (use_dynamic_sprintf) {
    output_string += temp2;
    delete [] temp2;
  } else output_string += temp;
}

//hmmm: de-redundify this function, which is identical to the constructor.
void istring::reset(special_flag flag, const char *initial, ...)
{
  reset();  // clear the string out.
  if (!initial) return;
  if ( (flag != UNTERMINATED) && (flag != SPRINTF) ) {
    operator = (istring(istring::SPRINTF, "unknown flag %d", flag));
    return;
  }

  va_list args;
  va_start(args, initial);

  if (flag == UNTERMINATED) {
    // special process for grabbing a string that has no terminating nil.  
    int length = va_arg(args, int);  // get the length of the string out.
    _implementation->reset(length, (byte *)initial);
    (*_implementation) += byte(0);
    va_end(args);
    return;
  }

  // only other flag currently supported is sprintf, so we do that...
  base_sprintf(initial, args);
  va_end(args);
}

void istring::pad(int len, char padding)
{
  if (length() >= len) return;
  byte_sequence pad(len - length());
  memset(pad.access(), padding, pad.length());
  operator += (istring(UNTERMINATED, (char *)pad.observe(), pad.length()));
}

void istring::trim(int len)
{
  if (length() <= len) return;
  zap(len, end());
}

istring &istring::operator = (const istring &s1)
{
  if (this != &s1) _implementation->operator = (*s1._implementation);
  return *this;
}

istring &istring::operator = (const char *s1)
{
  reset();
  *this += s1;
  return *this;
}

void istring::zap(int position1, int position2)
{
  bounds_return(position1, 0, end(), );
  bounds_return(position2, 0, end(), );
  _implementation->zap(position1, position2);
}

void istring::to_lower()
{
  for (int i = 0; i < length(); i++)
    if ( (get(i) >= 'A') && (get(i) <= 'Z') )
      _implementation->put(i, char(get(i) - CASE_DIFFERENCE));
}

void istring::to_upper()
{
  for (int i = 0; i < length(); i++)
    if ( (get(i) >= 'a') && (get(i) <= 'z') )
      _implementation->put(i, char(get(i) + CASE_DIFFERENCE));
}

istring istring::lower() const
{
  istring to_return(*this);
  to_return.to_lower();
  return to_return;
}

istring istring::upper() const
{
  istring to_return(*this);
  to_return.to_upper();
  return to_return;
}

void istring::copy(char *array_to_stuff, int how_many) const
{
  if (!array_to_stuff) return;
  array_to_stuff[0] = '\0';
  if ( (how_many <= 0) || (length() <= 0) ) return;
  strncpy(array_to_stuff, observe(), minimum(how_many, int(length())));
  array_to_stuff[minimum(how_many, int(length()))] = '\0';
}

bool istring::iequals(const istring &that) const
{ return strcasecmp(observe(), that.observe()) == 0; }

bool istring::iequals(const char *that) const
{ return strcasecmp(observe(), that) == 0; }

int istring::ifind(char to_find, int position, bool reverse) const
{ return char_find(to_find, position, reverse, false); }

int istring::find(char to_find, int position, bool reverse) const
{ return char_find(to_find, position, reverse, true); }

int istring::find_any(const istring &to_find, int position, bool reverse) const
{ return char_find_any(to_find, position, reverse, true); }

int istring::ifind_any(const istring &to_find, int position, bool reverse) const
{ return char_find_any(to_find, position, reverse, false); }

char simple_lower(char input)
{
  if ( (input <= 'Z') && (input >= 'A') ) return input - CASE_DIFFERENCE;
  return input;
}

int istring::char_find(char to_find, int position, bool reverse,
    bool case_sense) const
{
  if (position < 0) return common::OUT_OF_RANGE;
  if (position > end()) return common::OUT_OF_RANGE;
  if (reverse) {
    for (int i = position; i >= 0; i--) {
      if (case_sense && (get(i) == to_find)) return i;
      else if (simple_lower(get(i)) == simple_lower(to_find)) return i;
    }
  } else {
    if (case_sense) {
      const char *const pos = strchr(observe() + position, to_find);
      if (pos) return int(pos - observe());
    } else {
      for (int i = position; i < length(); i++)
        if (simple_lower(get(i)) == simple_lower(to_find)) return i;
    }
  }
  return common::NOT_FOUND;
}

bool imatches_any(char to_check, const istring &list)
{
  for (int i = 0; i < list.length(); i++)
    if (simple_lower(to_check) == simple_lower(list[i])) return true;
  return false;
}

bool matches_any(char to_check, const istring &list)
{
  for (int i = 0; i < list.length(); i++)
    if (to_check == list[i]) return true;
  return false;
}

int istring::char_find_any(const istring &to_find, int position, bool reverse,
    bool case_sense) const
{
  if (position < 0) return common::OUT_OF_RANGE;
  if (position > end()) return common::OUT_OF_RANGE;
  if (reverse) {
    for (int i = position; i >= 0; i--) {
      if (case_sense && matches_any(get(i), to_find)) return i;
      else if (imatches_any(get(i), to_find)) return i;
    }
  } else {
    for (int i = position; i < length(); i++) {
      if (case_sense && matches_any(get(i), to_find)) return i;
      else if (imatches_any(get(i), to_find)) return i;
    }
  }
  return common::NOT_FOUND;
}

int istring::find(const istring &to_find, int posn, bool reverse) const
{ return str_find(to_find, posn, reverse, true); }

int istring::ifind(const istring &to_find, int posn, bool reverse) const
{ return str_find(to_find, posn, reverse, false); }

int istring::str_find(const istring &to_find, int posn, bool reverse,
    bool case_sense) const
{
  bounds_return(posn, 0, end(), common::OUT_OF_RANGE);
  if (!to_find.length()) return common::BAD_INPUT;

  // skip some steps by finding the first place that the first character of
  // the string resides in our string.
  if (case_sense) posn = find(to_find[0], posn, reverse);
  else posn = ifind(to_find[0], posn, reverse);
  if (posn < 0) return common::NOT_FOUND;
  
//hmmm: there is a better way to do this loop in terms of the number of
//      comparisons performed.  knuth morris pratt algorithm?
  if (case_sense) {
//hmmm: this could use strncmp too?
    if (reverse) {
      if (posn > length() - to_find.length())
        posn = length() - to_find.length();
      for (int i = posn; i >= 0; i--)
        if (!memcmp((void *)&observe()[i], (void *)to_find.observe(),
              to_find.length()))
          return i;
    } else {
      const int find_len = to_find.length();
      const int str_len = length();
      const char first_char = to_find[0];
      bounds_return(posn, 0, str_len - find_len, common::OUT_OF_RANGE);
      for (int i = posn - 1;
          ( ( (i = find(first_char, i + 1)) >= 0)
            && (str_len - i >= find_len) ); /*no increment*/) {
        if (!memcmp((void *)&observe()[i], (void *)to_find.observe(),
              to_find.length()))
          return i;
      }
    }
  } else {
    // not case-sensitive.
    if (reverse) {
      if (posn > length() - to_find.length())
        posn = length() - to_find.length();
      for (int i = posn; i >= 0; i--)
        if (!strncasecmp(&observe()[i], to_find.observe(), to_find.length()))
          return i;
    } else {
      bounds_return(posn, 0, length() - to_find.length(), common::OUT_OF_RANGE);
      for (int i = posn; i < length() - to_find.length() + 1; i++)
        if (!strncasecmp(&observe()[i], to_find.observe(), to_find.length()))
          return i;
    }
  }
  return common::NOT_FOUND;
}

istring istring::operator + (const istring &s1) const
{
  istring to_return(*this);
  to_return += s1;
  return to_return;
}

char &istring::operator [] (int position)
{
  if (position < 0) position = 0;
  if (position > end()) position = 0;
  byte &found = _implementation->use(position);
  char &to_return = *((char *)(&found));
  return to_return;
}

const char &istring::operator [] (int position) const
{
  if (position < 0) position = 0;
  if (position > end()) position = 0;
  const byte &found = _implementation->get(position);
  const char &to_return = *((const char *)(&found));
  return to_return;
}

int istring::convert(int default_value) const
{
  if (!length()) return default_value;
  int to_return;
  int fields = sscanf(observe(), "%d", &to_return);
  if (fields < 1) return default_value;
  return to_return;
}

long istring::convert(long default_value) const
{
  if (!length()) return default_value;
  long to_return;
  int fields = sscanf(observe(), "%ld", &to_return);
  if (fields < 1) return default_value;
  return to_return;
}

float istring::convert(float default_value) const
{
  if (!length()) return default_value;
  float to_return;
  int fields = sscanf(observe(), "%f", &to_return);
  if (fields < 1) return default_value;
  return to_return;
}

double istring::convert(double default_value) const
{
  if (!length()) return default_value;
  double to_return;
  int fields = sscanf(observe(), "%lf", &to_return);
  if (fields < 1) return default_value;
  return to_return;
}

istring &istring::operator += (const char *s1)
{
  if (!s1) return *this;
  int len = length();
  _implementation->insert(len, strlen(s1));
  memory_assign((char *)&_implementation->operator [](len),
      s1, strlen(s1));
  return *this;
}

istring &istring::operator += (char s1)
{
  int len = length();
  _implementation->insert(len, 1);
  _implementation->put(len, s1);
  return *this;
}

bool istring::compare(const istring &to_compare, int start_first,
  int start_second, int count) const
{
  bounds_return(start_first, 0, end(), false);
  bounds_return(start_second, 0, to_compare.end(), false);
  bounds_return(start_first + count, start_first, length(), false);
  bounds_return(start_second + count, start_second, to_compare.length(), false);
  return !memcmp((void *)&observe()[start_first],
      (void *)&to_compare.observe()[start_second], count);
}

void istring::substring(istring &target, int start, int bender) const
{
  #define func "substring"
  target.reset();
  if (bender < start) return;
  const int last = end();  // final position that's valid in the string.
  bounds_halt(start, 0, last, );
  bounds_halt(bender, 0, last, );
  target.reset(UNTERMINATED, observe() + start, bender - start + 1);
  #undef func
}

istring istring::substring(int start, int end) const
{
  istring to_return;
  substring(to_return, start, end);
  return to_return;
}

void istring::insert(int position, const istring &to_insert)
{
  bounds_return(position, 0, length(), );
  if (this == &to_insert) {
    istring copy_of_me(to_insert);
    insert(position, copy_of_me);  // not recursive because no longer == me.
  } else {
    _implementation->insert(position, to_insert.length());
    _implementation->overwrite(position, *to_insert._implementation,
        to_insert.length());
  }
}

bool istring::replace(const istring &tag, const istring &replacement)
{
  int where = find(tag);
  if (negative(where)) return false;
  zap(where, where + tag.end());
  insert(where, replacement);
  return true;
}

bool istring::replace_all(const istring &to_replace, const istring &new_string)
{
  bool did_any = false;
  for (int i = 0; i < length(); i++) {
    int indy = find(to_replace, i);
    if (negative(indy)) break;  // get out since there are no more matches.
    i = indy;  // update our position to where we found the string.
    zap(i, i + to_replace.length() - 1);  // remove old string.
    insert(i, new_string);  // plug the new string into the old position.
    i += new_string.length() - 1;  // jump past what we replaced.
    did_any = true;
  }
  return did_any;
}

bool istring::replace_all(char to_replace, char new_char)
{
  bool did_any = false;
  for (int i = 0; i < length(); i++) {
    if (get(i) == to_replace) {
      put(i, new_char);
      did_any = true;
    }
  }
  return did_any;
}

bool istring::matches(const istring &match_list, char to_match)
{
  for (int i = 0; i < match_list.length(); i++)
    if (to_match == match_list.get(i)) return true;
  return false;
}

void istring::strip(const istring &strip_list, how_to_strip way)
{
  if (way & FROM_FRONT)
    while (length() && matches(strip_list, get(0)))
      zap(0, 0);

  if (way & FROM_END)
    while (length() && matches(strip_list, get(end())))
      zap(end(), end());
}

void istring::pack(byte_array &target) const
{ basis::attach(target, (char *)_implementation->observe()); }

bool istring::unpack(byte_array &source)
{ return basis::detach(source, *this); }

int istring::icompare(const istring &to_compare, int length_in) const
{
  if (!length_in) return 0;  // nothing is equal to nothing.
  int real_len = length_in;
  // if they're passing a negative length, we use the full length.
  if (negative(length_in)) real_len = length();
  // if we have no length, make the obvious returns now.
  if (!real_len) return to_compare.length()? -1 : 0;
  // if the second string is empty, it's always less than the non-empty.
  if (!to_compare.length()) return 1;
  int to_return = strncasecmp(observe(), to_compare.observe(), real_len);
  if (negative(length_in) && !to_return
      && (to_compare.length() > length()) ) {
    // catch special case for default length when the two are equal except
    // that the second string is longer--this means the first is less than
    // second, not equal.
    return -1;
  } else
    return to_return;
}

int istring::icompare(const char *first, const char *second, int length)
{
  int len1 = strlen(first);
  int len2 = strlen(second);
  if (!length) return 0;  // no characters are equal to none.
  if (!len1 && !len2) return 0;  // equal as empty.
  if (!len1 && len2) return -1;  // first < second.
  if (len1 && !len2) return 1;  // first > second.
  if (positive(length)) {
    len1 = minimum(length, len1);
    len2 = minimum(length, len2);
  }
  for (int i = 0; i < len1; i++) {
    if (i > len2 - 1) return 1;  // first > second, had more length.
    if (simple_lower(first[i]) < simple_lower(second[i]))
      return -1;  // first < second.
    if (simple_lower(first[i]) > simple_lower(second[i]))
      return 1;  // first > second.
  }
  // at this point we know second is equal to first up to the length of
  // first.
  if (len2 > len1) return -1;  // second was longer and therefore greater.
  return 0;  // equal.
}

#undef static_class_name

////////////////////////////////////////////////////////////////////////////

isprintf::isprintf() : istring() {}

isprintf::isprintf(const istring &s) : istring(s) {}

isprintf::isprintf(const char *initial, ...)
: istring()
{
  if (!initial) return;
  va_list args;
  va_start(args, initial);
  base_sprintf(initial, args);
  va_end(args);
}

