#ifndef LOG_BASE_CLASS
#define LOG_BASE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : log_base                                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "object_base.h"

//! Provides an abstract base for logging mechanisms.
/*!
  This assists greatly in generating diagnostic entries polymorphically and in
  enabling different logging mechanisms to be plugged in easily.  Note that all
  of the functions are defined virtually which enables overriding pretty much
  all of the base functionality.  Use this wisely, if ever.
*/ 

class log_base : public virtual object_base
{
public:
  log_base();  //!< Constructor initializes the log base members.

  log_base(const log_base &to_copy);  //!< copy constructor.

  virtual ~log_base();  //!< Virtual destructor deallocates everything.

  log_base &operator =(const log_base &to_copy); //!< assignment operator.

  //! Filter values are used to choose what kinds of logged items to allow.
  /*! the first service offered by the base class is filtering.  filters provide
  a way to reconfigure what is logged; anything in the current filter set
  WILL be logged when that filter is used in the log() method.  anything
  not in the set (except ALWAYS_PRINT) will not be logged.  the set of
  filters can be modified except that ALWAYS_PRINT will always be included
  in the set and NEVER_PRINT will never be included. */
  enum KNOWN_FILTERS {
    // these two filters are predefined and their meaning cannot be changed.
    ALWAYS_PRINT = common::ALWAYS_PRINT,  //!< Always allows the logging.
    NEVER_PRINT = common::NEVER_PRINT,  //!< Never allows the logging.
  };

  virtual void add_filter(int new_filter);
    //!< Adds a member to the filter set.
    /*!< The filter set is used to check all extended filter values passed to
    log and print.  if the special filters of ALWAYS_PRINT or NEVER_PRINT are
    added, then either everything will be logged or nothing will be. */
  virtual void remove_filter(int old_filter);
    //!< Removes a member from the filter set.

  virtual bool member(int filter_to_check);
    //!< Returns true if the "filter_to_check" is a member of the filter set.
    /*!< If "filter_to_check" is ALWAYS_PRINT, this always returns true.  If
    the value is NEVER_PRINT, false is always returned. */

  virtual void clear_filters();
    //!< Resets the filter set to be empty.

  virtual int_set filter_set();
    //!< Returns the current filter set's contents.

  //! Line endings specify how the endings of logged lines are handled.
  /*! the second service offered by this base class is the notion of line
  endings.  on ms-dos, every line in a file has a carriage return (CR) and
  a line feed (LF) appended to the line.  on unix, only a line feed (LF)
  character is used.  thus, if you want the logging entries to be logged
  on separate lines, you can use LF_AT_END or CRLF_AT_END to cause that.
  if you want to make up a logging line out of multiple parts, then use
  NO_ENDING before printing the first part (saving the original setting for
  eol() temporarily), print every part but the last part, change the ending
  back to its original value, and print the final part.
  NOTE: it is up to the derived classes to implement the use of the
  appropriate line ending; they are the implementors of the log() method
  and so must do whatever is right for the current ending (the get_ending()
  method can help). */

  enum line_ending {
    LF_AT_END = -15,  //!< Unix standard is LF_AT_END ("\n").
    CRLF_AT_END,  //!< DOS standard is CRLF_AT_END ("\r\n").
    NO_ENDING  //!< No additional characters will be added to line endings.
  };

  virtual line_ending eol();
    //!< observes how line endings are to be printed.
  virtual void eol(line_ending to_set);
    //!< modifies how line endings are to be printed.

  virtual istring get_ending();
    //!< returns a string for the current ending.
  virtual void get_ending(istring &to_end);
    //!< appends a string for the current ending to "to_end".

  // the real core of the log_base is the log() method, which all derived
  // classes MUST provide.

  virtual outcome log(const istring &info, int filter = ALWAYS_PRINT) = 0;
    //!< writes the information in "info" to the log using the "filter" value.
    /*!< the "filter" value must be checked to see if it is in the current set
    of allowed filters (using member()).  this function _must_ be provided by
    derived classes and it _must_ check the filter before printing if it wants
    to be in compliance with log_base protocols.  it also must implement the
    line_ending appropriately.  the return value will often be a member of the
    common outcomes, but derived classes may see fit to return other things.
    note that an object derived from log_base should be thread-safe and protect
    its own member attributes through synchronization. */

  outcome log(const char *info, int filter = ALWAYS_PRINT);
    //!< synonym for log that provides char * support.
  inline log_base &operator <<(const istring &info) { log(info); return *this; }
    //!< synonym for log that mimics a stream interface and takes an istring.
  inline log_base &operator <<(const char *info) { log(info); return *this; }
    //!< synonym for log that mimics a stream interface and takes char *.

  // helpful methods for dealing with line endings.

  static const char *string_for_ending(line_ending ending);
    //!< returns the string form for the "ending" value.
  static line_ending pick_ending_for_platform();
    //!< provides the appropriate ending on the current OS platform.
  static const char *platform_ending();
    //!< provides the characters that make up this platform's line ending.

  // nitty-gritty special functions...

  void stuff_filter_set(const int_set &new_set);
    //!< this non-virtual method simply accesses the internal list.
    /*!<  it allows derived objects to change the filter list without
    confusion about which object implements the call. */

private:
  mutex *_lock;  //!< our synchronization support.
  int_set *_filter_set;  //!< set of values that are allowed to print.
  line_ending _ending;  //!< the current printing style.
};

////////////////////////////////////////////////////////////////////////////

// these macros can assist in logging.  they rely on the base class for
// logging services.  note that it is often convenient to customize these
// to yield a simpler macro name per project, such as PRINT or LOG.

//! Logs a string "to_log" on "the_logger" using the "filter".
/*! The filter is checked before the string is allowed to come into
existence, which saves allocations when the item would never be printed
out anyway. */
#define FILTER_LOG(the_logger, to_log, filter) { \
  if (the_logger.member(filter)) { \
    the_logger.log(to_log, filter); \
  } \
}

//! Logs a string at the emergency level, meaning that it always gets logged.
#define EMERGENCY_LOG(the_logger, to_log) \
  FILTER_LOG(the_logger, to_log, log_base::ALWAYS_PRINT)

//! Corresponding functions for including the time and date in the log entry.
/*! These require visibility of basis/utility.h in files using the macro. */
#define STAMPED_FILTER_LOG(the_logger, to_log, filter) { \
  if (the_logger.member(filter)) { \
    istring temp_log = to_log; \
    if (temp_log.length()) \
      temp_log.insert(0, timestamp(true, true)); \
    the_logger.log(temp_log, filter); \
  } \
}
//! Time-stamped logging that will always be printed.
#define STAMPED_EMERGENCY_LOG(the_logger, to_log) \
  STAMPED_FILTER_LOG(the_logger, to_log, log_base::ALWAYS_PRINT)

//! Class specific logging method that uses a filter.
/* These add a class and function name to the log entry.
These also requires basis/utility.h. */
#define CLASS_FILTER_LOG(the_logger, to_log, filter) { \
  if (the_logger.member(filter)) { \
    istring temp_log = to_log; \
    if (temp_log.length()) { \
      temp_log.insert(0, timestamp(true, true)); \
      BASE_FUNCTION(func); \
      temp_log += " ["; \
      temp_log += function_name; \
      temp_log += "]"; \
    } \
    the_logger.log(temp_log, filter); \
  } \
}
//! Class specific logging method that always prints.
#define CLASS_EMERGENCY_LOG(the_logger, to_log) \
  CLASS_FILTER_LOG(the_logger, to_log, log_base::ALWAYS_PRINT)

//! Logs information that includes specific class instance information.
/*! This use the instance name of the object, which can include more
information than the simple class name. */
#define INSTANCE_FILTER_LOG(the_logger, to_log, filter) { \
  if (the_logger.member(filter)) { \
    istring temp_log = to_log; \
    if (temp_log.length()) { \
      temp_log.insert(0, timestamp(true, true)); \
      BASE_INSTANCE_FUNCTION(func); \
      temp_log += " ["; \
      temp_log += function_name; \
      temp_log += "]"; \
    } \
    the_logger.log(temp_log, filter); \
  } \
}
//! Logs with class instance info, but this always prints.
#define INSTANCE_EMERGENCY_LOG(the_logger, to_log) \
  INSTANCE_FILTER_LOG(the_logger, to_log, log_base::ALWAYS_PRINT)

#endif

