/*****************************************************************************\
*                                                                             *
*  Name   : mutex                                                             *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

// NOTE: we are explicitly avoiding use of new and delete here because this
//       class is needed by our memory allocation object, which would be
//       providing the new and delete methods.

#include "mutex.h"
#include "portable.h"
#include "utility.h"

#include <stdlib.h>

#ifdef __UNIX__
  #include <pthread.h>
#endif

mutex_base::mutex_base()
{
  construct();
}

mutex_base::~mutex_base()
{
  destruct();
}

void mutex_base::construct()
{
#ifdef __WIN32__
  _os_mutex = (CRITICAL_SECTION *)malloc(sizeof(CRITICAL_SECTION));
  InitializeCriticalSection((LPCRITICAL_SECTION)_os_mutex);
#elif defined(__UNIX__)
  pthread_mutexattr_t attr;
  pthread_mutexattr_init(&attr);
  int ret = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE_NP);
  if (ret != 0) {
//printf("failed to initialize mutex attributes!\n"); fflush(NIL);
  }
  _os_mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init((pthread_mutex_t *)_os_mutex, &attr);
  pthread_mutexattr_destroy(&attr);
#else
  #pragma error("no implementation of mutexes for this OS yet!")
#endif
}

void mutex_base::destruct()
{
  defang();
}

void mutex_base::defang()
{
  if (!_os_mutex) return;  // already defunct.
#ifdef __WIN32__
  DeleteCriticalSection((LPCRITICAL_SECTION)_os_mutex);
  free(_os_mutex);
#elif defined(__UNIX__)
  pthread_mutex_destroy((pthread_mutex_t *)_os_mutex);
  free(_os_mutex);
#else
  #pragma error("no implementation of mutexes for this OS yet!")
#endif
  _os_mutex = 0;
}

void mutex_base::lock()
{
  if (!_os_mutex) return;
#ifdef __WIN32__
  EnterCriticalSection((LPCRITICAL_SECTION)_os_mutex);
#elif defined(__UNIX__)
  pthread_mutex_lock((pthread_mutex_t *)_os_mutex);
#else
  #pragma error("no implementation of mutexes for this OS yet!")
#endif
}

void mutex_base::unlock()
{
  if (!_os_mutex) return;
#ifdef __WIN32__
  LeaveCriticalSection((LPCRITICAL_SECTION)_os_mutex);
#elif defined(__UNIX__)
  pthread_mutex_unlock((pthread_mutex_t *)_os_mutex);
#else
  #pragma error("no implementation of mutexes for this OS yet!")
#endif
}

