#ifndef MUTEX_CLASS
#define MUTEX_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : mutex                                                             *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "auto_synch.h"
  // we include auto_synch here because it is so often used with mutexes.

//! A simple primitive class that encapsulates OS support for mutual exclusion.
/*!
  The word "mutex" is an abbreviation for "mutual exclusion".  The mutex
  provides a simple synchronization object that supports the programming of
  critical sections.  It is guaranteed to be safe for threads, but it is only
  useful within one application rather than between multiple applications.
  The mutex_base is hardly ever used directly; instead the mutex class should
  be used.
*/

class mutex_base : public synchronizer_base
{
public:
  mutex_base();  //!< Constructs a new mutex.

  virtual ~mutex_base();
    //!< Destroys the mutex.  It should not be locked upon destruction.

  //! Constructor for use with malloc/free instead of new/delete.
  void construct();
  //! Destructor for use with malloc/free instead of new/delete.
  void destruct();

  void lock();
    //!< Clamps down on the mutex, if possible.
    /*!< Otherwise the current thread is blocked until the mutex is unlocked. */

  void unlock();
    //!< Gives up the possession of the mutex.

  virtual void establish_lock() { lock(); }
    //!< Satisfies base class requirements for locking.
  virtual void repeal_lock() { unlock(); }
    //!< Satisfies base class requirements for unlocking.

  void defang();
    //!< Removes the underlying OS synchronization primitive.
    /*!< This method renders this mutex object inoperable.  This is useful
    when the reason for the lock has vanished, but the mutex object cannot be
    deleted yet.  Sometimes it may still be referred to, but there is no
    longer any critical section to be protected. */

private:
  void *_os_mutex;  //!< OS version of the mutex.

  mutex_base(const mutex_base &);  //!< not allowed.
  mutex_base &operator =(const mutex_base &);  //!< not allowed.
};

////////////////////////////////////////////////////////////////////////////

//! The object that will actually be used for mutual exclusion.
/*! See mutex_base for the real implementation. */
class mutex : public mutex_base, public virtual object_base
{
public:
  mutex() {}  //!< Constructs a new mutex.
  virtual ~mutex() {}  //!< Destroys this mutex.
  IMPLEMENT_CLASS_NAME("mutex");
};

#endif

