#ifndef PORTABILITY_GROUP
#define PORTABILITY_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : portability definitions                                           *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1994-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "istring.h"

//! @file portable.h Support for platform independence.
/*! @file portable.h
  These definitions, inclusions and types are aimed at allowing our source
  code to remain independent of the underlying windowing and operating
  systems.
*/

// forward.
class char_star_array;  // see basis/common_templates.

// gnarly headers that are needed for certain types of compilation...

#ifdef __UNIX__
  #include <unistd.h>
#endif
#ifndef NO_XWINDOWS
  #ifdef __XWINDOWS__
    #include <Intrinsic.h>
    #include <StringDefs.h>
    #include <Xm/Xm.h>
    #include <Xlib.h>
  #endif
#endif
#ifdef __WIN32__
  #ifndef STRICT
    #define STRICT
  #endif
  #ifndef _AFXDLL
    // include ms-windows headers only if we're not doing mfc; mfc has its own
    // special way of including the headers.
    #include <windows.h>
  #else
    #include <afx.h>
    #include <afxwin.h>
  #endif
  class _bstr_t;
#endif

// forward.
struct timeval;
class version;

// wrapper classes defined in the sequel...
//
//   application_instance -- this program's application object; used mainly
//       in ms-windows.
//
//   window_handle -- a wrapper for the root of all objects that can be
//       referred to in the relevant windowing system.

#ifdef __UNIX__
  // the application_instance class is implemented very simply for unix;
  // it's a stand-in since unix apps don't refer to an instance handle like
  // ms-windows does.
  typedef void *application_instance;

  // some definitions to quiet complaints from win32-based code.
  typedef u_int DWORD;
  typedef void *HANDLE;

  //temp; these just mimic function prototypes coming from non-portable code.
  typedef void *ATOM;
  typedef void *BITMAPINFO;
  typedef void *HBITMAP;
  typedef void *HDC;
  typedef void *RGBQUAD;
  #define LoadBitmap(a, b) (a)
  #define __stdcall 

  // portability definitions for the X windowing system.
  #ifndef NO_XWINDOWS  // protects regions with similar names.
  #ifdef __XWINDOWS__
    typedef Widget window_handle;
      //!< main way to access a windowing system object.  should be a pointer.

    typedef Colormap window_colormap;
      //!< the windowing system representation of a pallette or colormap.

    typedef XColor window_color;
      //!< the system representation of color.

    typedef XmString window_string;
      //!< a special windowing system dependent kind of character string.

//is that really fixed?
    const int MAXIMUM_COLOR_INTENSITY = 65535;
      // largest valid value a color component (R, G or B) can have.
  #else
    // no x-windows.
    typedef void *window_handle;
  #endif
  #endif
#endif

#ifdef __WIN32__
  typedef HINSTANCE application_instance;
    // our moniker for an application's guts.

  typedef HWND window_handle;

  #include "convert_utf.h"

  #ifdef _AFXDLL
    // these two helper functions support MFC Cstrings.  they were moved outside
    // the istring class to support smaller builds that don't require MFC (such
    // as for ATL components).
    inline istring convert(const CString &original)
            { return istring(from_unicode_temp(original)); }
    inline CString convert(const istring &original)
            { return CString(to_unicode_temp(original)); }
  #endif

  // conversions between this and the _bstr_t object.  this supports ATL
  // strings being used interchangeably with istrings.
  istring convert_bstr(const _bstr_t &original);
  _bstr_t convert_bstr(const istring &original);

#endif

// now the actual portable namespace...

namespace portable {

  // assorted useful system functions:

  version get_OS_version();
    //!< returns the operating system's version information.
    /*!< for linux, this has: major/minor/release/revision as components.
    for ms-windows, this has: major/minor/platform_ID/build_number. */

  u_int process_id();
    //!< returns the process id for this task, if that's relevant on the OS.

  istring current_directory();
    //!< returns the current directory as reported by the operating system.

  istring application_name();
    //!< returns the full name of the current application.

  istring module_name(const void *module_handle = NIL);
    //!< returns the name of the module for the "module_handle" where supported.
    /*!< if the handle is NIL, then the program name is returned. */

  u_int system_uptime();
    //!< the length of time (in milliseconds) since this computer booted.

  void sleep_ms(u_int msec);
    //!< a system independent name for a forced snooze measured in milliseconds.
    /*!< the application will do nothing on the current thread for the amount of
    time specified.  on some systems, if "msec" is zero, then the sleep will
    yield the current thread's remaining timeslice back to other threads. */

  istring env_string(const istring &variable_name);
    //!< looks up the "variable_name" in the current environment variables.
    /*!< this returns the value for "variable_name" as it was found in the
    operating system's environment variables that are defined at this point
    in time for the user and process.  the returned string will be empty if no
    variable under that name could be found. */

  bool set_environ(const istring &variable_name, const istring &value);
    //!< adds or creates "variable_name" in the environment.
    /*!< changes the current set of environment variables by adding or
    modifying the "variable_name".  its new value will be "value". */

  u_int system_error();
    //!< gets the most recent system error reported on this thread.

  istring system_error_text(u_int error_to_show);
    //!< returns the OS's string form of the "error_to_show".
    /*!< this often comes from the value reported by system_error(). */

  istring null_device();
    //!< returns the name of the system's NULL device.
    /*!< this device is a black hole where output can be sent, never to be
    seen again.  this is /dev/null on unix and null: on win32. */

  char_star_array break_line(istring &app, const istring &parameters);
    //!< prepares an "app" to launch with the "parameters" (via exec).
    /*!< this breaks the strings for an application named "app" and its
    "parameters" into an array of char * that is appropriate for the execv
    function. */

  enum process_launch_flags {
    AWAIT_APP_EXIT = 0x1,
      //!< stays in the function until the launched application has exited.
    RETURN_IMMEDIATELY = 0x2,
      //!< starts the application and comes right back to the caller.
    HIDE_APP_WINDOW = 0x4
      //!< launches the application invisibly if possible.
  };

  u_int launch_process(const istring &app_name, const istring &command_line,
          int flag, u_int &child_id);
    //!< starts an application using the "app_name" as the executable to run.
    /*!< the "command_line" is the set of parameters to be passed to the app.
    the return value is OS specific but can be identified with
    system_error_text().  usually a zero return means success and non-zero
    codes are errors.  if the AWAIT flag is used, then the return value will
    be the launched process's own exit value.  the thread or process_id of the
    launched process is stored in "child_id" if appropriate.  the "flag" is an
    XORed value from the process_launch_flags that dictates how the app is to
    be started. */

  timeval fill_timeval_ms(int milliseconds);
    //!< returns a timeval system object that represents the "milliseconds".
    /*!< if "milliseconds" is zero, then the returned timeval will
    represent zero time passing (rather than infinite duration as some
    functions assume). */

////////////////////////////////////////////////////////////////////////////

  // Unix and Linux specific items are included here.
  #ifdef __UNIX__
    istring get_cmdline_from_proc();
      //!< returns the command line that /proc has recorded for our process.

    char *itoa(int to_convert, char *buffer, int radix);
      //!< turns the integer "to_convert" into a string stored in the "buffer".
      /*!< this is needed on linux since there doesn't seem to be a builtin
      version of it.  the buffer must be long enough to hold the number's
      textual representation!  the buffer's length must also account for the
      chosen "radix" (the base for the number system, where 10 is decimal,
      16 is hexadecimal, etc). */

    #define RGB(r, g, b) (b + (g << 8) + (r << 16))
      //!< no idea if that's even approximately right.
  #endif

  // ms-windows of more modern types, i.e. win32.
  #ifdef __WIN32__

//hmmm: is there an equivalent to this for unix?
    bool is_address_valid(const void *address, int size_expected,
            bool writable = false);
      //!< checks that the address specified is a valid pointer.
      /*! also tests that the address's allocated space has the
      "size_expected".  if "writable" is true, then the pointer is
      checked for being writable as well as readable. */

    #define BROADCAST_HANDLE HWND_BROADCAST

    bool has_win95_style_shell();
      //!< this returns true if the operating system is using the "new" shell.
      /*!< the new shell is version 4.0 which was first seen in windows 95.
      if true is returned, then it indicates that the operating system is
      windows 9x or later or that it is Windows NT 4.0 or later. */

    enum known_operating_systems { WIN_95, WIN_NT, WIN_2K, WIN_XP, UNKNOWN_OS };
    known_operating_systems determine_OS();
      //!< returns the operating system that seems to be running currently.
      /*!< note that WIN_95 also covers windows 98 and windows ME. */

    istring rc_string(u_int id, application_instance instance);
      //!< acquires a string from a particular "instance".
      /*!< the "id" is the resource identifier for the desired string in
      that "instance". */

    istring rc_string(u_int id);
      //!< simplified from above function by not needing an "instance".
      /*!< the "current" dynamic library or executable's resources are sought
      for the "id". */

    void show_wait_cursor();
      //!< changes the cursor to the "wait" look, which is usually an hourglass.
    void show_normal_cursor();
      //!< changes the cursor to "normal", which is usually a pointing arrow.

    extern application_instance _i_handle;
      //!< dll_root.cpp defines this for all dlls.
    #define DEFINE_INSTANCE_HANDLE application_instance portable::_i_handle = 0
      //!< some applications may need this to use rc_string and others.
      /*!< if this macro is used, then the code is impervious to future
      changes in how the instance handle works. */
    #define SET_INSTANCE_HANDLE(value) portable::_i_handle = value
      //!< use this to change the instance handle for this dll or exe.
      /*!< note that this should be done only once and by the main thread. */
    #define GET_INSTANCE_HANDLE() portable::_i_handle
      //!< a handy macro that frees one from knowing the name of the handle.

  #else // not win32.

    // turn off a couple of macros that might show up in code but which are
    // meaningless in non-win32 programs.
    #define DEFINE_INSTANCE_HANDLE
    #define SET_INSTANCE_HANDLE(value)
    #define GET_INSTANCE_HANDLE() 0

  #endif // win32.

} // namespace.

#endif // outer guard.

