#ifndef SET_CLASS
#define SET_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : set                                                               *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "array.h"

namespace basis {

//! Emulates a mathematical set, providing several standard set operations.

template <class contents>
class set : public array<contents>
{
public:
  //! Constructs a set with "num" elements, copying them from "init".
  /*! Be very careful to ensure that the array "init" has sufficient length
  for "num" elements to be copied from it. */
  inline set(int num = 0, const contents *init = NIL,
      u_short flags = array<contents>::EXPONE)
  : array<contents>(num, init, flags) {}

  inline ~set() {}  //!< Destroys any storage held for the set.

  inline int elements() const { return this->length(); }
    //!< Returns the number of elements in this set.

  inline bool empty() const { return this->elements() == 0; }
    //!< Returns true if the set has no elements.
  inline bool non_empty() const { return this->elements() != 0; }
    //!< Returns true if the set has some elements.

  inline void clear() { this->reset(); }  //!< Empties out this set.

  bool member(const contents &to_test) const;
    //!< Returns true if the item "to_test" is a member of this set.
  
  bool add(const contents &to_add);
    //!< Adds a new element "to_add" to the set.
    /*!< This always succeeds, but will return true if the item was not
    already present. */

  inline set &operator += (const contents &to_add)
      { this->add(to_add); return *this; }
    //!< An algebraic operator synonym for add() that operates on the contents.
  inline set &operator += (const set &to_add)
      { this->unionize(to_add); return *this; }
    //!< An algebraic operator synonym for add() that operates on a set.

  bool remove(const contents &to_remove);
    //!< Removes the item "to_remove" from the set.
    /*!< If it was not present, false is returned and the set is unchanged. */

  inline set &operator -= (const contents &to_zap)
      { this->remove(to_zap); return *this; }
    //!< An algebraic operator synonym for remove that operates on the contents.
  inline set &operator -= (const set &to_zap)
      { this->differentiate(to_zap); return *this; }
    //!< An algebraic operator synonym for remove that operates on a set.

  set set_union(const set &union_with) const;
    //!< Implements the set union of "this" with "union_with".
    /*!< This returns the set formed from the union of "this" set with the set
    specified in "union_with".  (unfortunately, the name "set_union" must
    be used to distinguish from the C keyword "union".) */

  void unionize(const set &union_with);
    //!< Makes "this" set a union of "this" and "union_with".

  inline set operator + (const set &uw) const { return this->set_union(uw); }
    //!< A synonym for set_union.

  set intersection(const set &intersect_with) const;
    //!< Returns the intersection of "this" with the set in "intersect_with".

  inline set operator * (const set &iw) const { return this->intersection(iw); }
    //!< A synonym for intersection.

  set difference(const set &differ_with) const;
    //!< Returns the difference of this with "differ_with".
    /*!< Difference is defined as the subset of elements in "this" that are not
    also in "differ_with". */

  void differentiate(const set &differ_with);
    //!< Makes "this" set equal to the difference of "this" and "differ_with".
    /*!< That is, after the call, "this" will only contain elements that were
    not also in "differ_with". */

  inline set operator - (const set &dw) const { return this->difference(dw); }
    //!< A synonym for difference.

  int find(const contents &to_find) const;
    //!< Returns the integer index of the item "to_find" in this set.
    /*!< This returns a negative number if the index cannot be found.  Note
    that this only makes sense within our particular implementation of set as
    an array. */

  //! Zaps the entry at the specified "index".
  /*! This also treats the set like an array.  The index must be within the
  bounds of the existing members. */
  inline bool remove_index(int index)
      { return this->zap(index, index) == common::OKAY; }
};

} // namespace.

#endif

