/*****************************************************************************\
*                                                                             *
*  Name   : check_version                                                     *
*  Author : Chris Koeritz                                                     *
*  Credits: John Stockton (for original windows get_version support)          *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "common_templates.h"
#include "convert_utf.h"
#include "guards.h"
#include "istring.h"
#include "portable.h"
#include "utility.h"
#include "version_checker.h"
#include "version_record.h"

#ifdef _MSC_VER
  #include <direct.h>
  #include <winver.h>
#elif defined(__BCPLUSPLUS__)
  #include <dir.h>
  #include <ver.h>
#endif

#ifdef __WIN32__
  // ensures that we handle the data properly regardless of unicode settings.
  #ifdef UNICODE
    #define render_ptr(ptr) from_unicode_temp( (UTF16 *) ptr)
  #else
    #define render_ptr(ptr) ( (char *) ptr)
  #endif
#endif

///////////////////////////////////////////////////////////////////////////

version_checker::version_checker(const istring &library_file_name,
    const version &expected_version, const istring &version_complaint)
: _library_file_name(new istring(library_file_name)),
  _expected_version(new version(expected_version)),
  _version_complaint(new istring(version_complaint))
{}

version_checker::~version_checker()
{
  WHACK(_library_file_name);
  WHACK(_expected_version);
  WHACK(_version_complaint);
}

istring version_checker::text_form() const
{
  return istring(class_name()) + ": library_file_name=" + *_library_file_name
      + ", expected_version=" + _expected_version->text_form()
      + ", complaint_message=" + *_version_complaint;
}

bool version_checker::good_version() const
{
  version version_found = get_version(*_library_file_name);
  if (version_found.compatible(*_expected_version)) return true;  // success.
  complain_wrong_version(*_library_file_name, *_expected_version,
      version_found);
  return false;
}

bool version_checker::loaded(const istring &library_file_name)
{
#ifdef __WIN32__
  return bool(get_handle(library_file_name) != 0); 
#else
//temp code. 
  return true;
#endif
}

void *version_checker::get_handle(const istring &library_file_name)
{
#ifdef __WIN32__
  return GetModuleHandle(to_unicode_temp(library_file_name));
#else
//hmmm: there really isn't this concept on OSes that i'm aware of.
  return 0;
#endif
}

istring version_checker::get_name(void *to_find)
{ return portable::module_name(to_find); }

bool version_checker::retrieve_version_info(const istring &filename,
    byte_array &to_fill)
{
  to_fill.reset();  // clear the buffer.

////  istring kludge = filename;  // MS doesn't do consts at interface.  argh!

  // determine the required size of the version info buffer.
  int required_size;
#ifdef __WIN32__
  u_long module_handle;  // filled with the dll or exe handle.
  required_size = GetFileVersionInfoSize(to_unicode_temp(filename), &module_handle);
#else
  required_size = 0;
#endif
  if (!required_size) return false;
  to_fill.reset(required_size);  // resize the buffer.
  
  // read the version info into our buffer.
  bool success = false;
#ifdef __WIN32__
  success = GetFileVersionInfo(to_unicode_temp(filename), module_handle,
      required_size, to_fill.access());
#else
  success = false;
#endif
  return success;
}

bool version_checker::get_language(byte_array &version_chunk,
    u_short &high, u_short &low)
{
  high = 0;
  low = 0;
#ifdef __WIN32__
  // determine the language that the version's written in.
  u_int data_size;
  void *pointer_to_language_structure;
  // query the information from the version blob.
  if (!VerQueryValue(version_chunk.access(),
      to_unicode_temp("\\VarFileInfo\\Translation"),
      &pointer_to_language_structure, &data_size))
    return false;
  // get the low & high shorts of the structure.
  high = LOWORD(*(unsigned int *)pointer_to_language_structure);
  low = HIWORD(*(unsigned int *)pointer_to_language_structure);
#else
  high = 0;
  low = 0;
#endif

  return true;
}

version version_checker::get_version(const istring &filename)
{
  byte_array version_info_found(0, NIL);
  if (!retrieve_version_info(filename, version_info_found))
    return version(0, 0, 0, 0);

  u_short high, low;  // holds the language of the version data.
  if (!get_language(version_info_found, high, low))
    return version(0, 0, 0, 0);

  // retrieve the file version from version info using the appropriate
  // language.
  istring root_key(istring::SPRINTF, "\\StringFileInfo\\%04x%04x", high, low);
  istring file_version_key(root_key + istring("\\FileVersion"));

  istring version_string;
#ifdef __WIN32__
  byte *file_version_pointer;
  u_int data_size;
  if (!VerQueryValue(version_info_found.access(),
      to_unicode_temp(file_version_key),
      (LPVOID *)&file_version_pointer, &data_size))
    return version(0, 0, 0, 0);
  version_string = render_ptr(file_version_pointer);
  // clean any spaces out of the string; people sometimes format these
  // very badly.
  for (int i = 0; i < version_string.length(); i++) {
    if (version_string[i] == ' ') {
      version_string.zap(i, i);
      i--;  // skip back a beat.
    }
  }
#else
  return version(0, 0, 0, 0);
//tmp.
#endif
  return version::from_text(version_string);
}

bool version_checker::get_record(const istring &filename, 
    version_record &to_fill)
{
  to_fill = version_record();
  byte_array version_info_found(0, NIL);
  if (!retrieve_version_info(filename, version_info_found))
    return false;

  u_short high, low;  // holds the language of the version data.
  if (!get_language(version_info_found, high, low))
    return false;

  // set the root key for all accesses of the version chunk.
  istring root_key(istring::SPRINTF, "\\StringFileInfo\\%04x%04x", high, low);

  // reports whether all lookups succeeded or not.
  bool total_success = true;

  // the various version pieces are retrieved...

#ifdef __WIN32__
  u_int data_size;
  void *data_pointer;

  // file version.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\FileVersion")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.file_version = version::from_text(render_ptr(data_pointer));

  // company name.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\CompanyName")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.company_name = render_ptr(data_pointer);

  // file description.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\FileDescription")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.description = render_ptr(data_pointer);

  // internal name.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\InternalName")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.internal_name = render_ptr(data_pointer);

  // copyright info.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\LegalCopyright")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.copyright = render_ptr(data_pointer);

  // trademark info.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\LegalTrademarks")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.trademarks = render_ptr(data_pointer);

  // original file name.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\OriginalFilename")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.original_name = render_ptr(data_pointer);

  // product name.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\ProductName")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.product_name = render_ptr(data_pointer);

  // product version.
  if (!VerQueryValue(version_info_found.access(), to_unicode_temp(root_key
      + istring("\\ProductVersion")), &data_pointer, &data_size))
    total_success = false;
  else
    to_fill.product_version = version::from_text(render_ptr(data_pointer));
#else
  // hmmm: chunks missing in version check.
#endif

  return total_success;
}

void version_checker::complain_wrong_version(const istring &library_file_name,
    const version &expected_version, const version &version_found) const
{
  istring to_show("There has been a Version Mismatch: The module \"");
  // use embedded module handle to retrieve name of dll or exe.
  istring module_name = get_name(GET_INSTANCE_HANDLE());
  if (!module_name) module_name = "Unknown";
  to_show += module_name;
  to_show += istring("\" cannot load.  This is because the file \"");
  to_show += *_library_file_name;
  to_show += istring("\" was expected to have a version of [");

  to_show += expected_version.flex_text_form(version::DOTS);

  to_show += istring("] but it instead had a version of [");
  to_show += version_found.flex_text_form(version::DOTS);

  to_show += istring("].  ");
  to_show += *_version_complaint;
#ifdef __UNIX__
  continuable_error("version checking", "failure", to_show.s());
#elif defined(__WIN32__)
  MessageBox(0, to_unicode_temp(to_show), to_unicode_temp("version_checking::failure"),
      MB_OK);
#endif
}

void version_checker::complain_cannot_load(const istring &lib_file) const
{
  istring to_show("There has been a failure in Version Checking: The file \"");
  to_show += lib_file;
  to_show += istring("\" could not be loaded or found.  ");
  to_show += *_version_complaint;
  continuable_error("version checking", "loading dll", to_show.s());
}

