#ifndef AMORPH_IMPLEMENTATION
#define AMORPH_IMPLEMENTATION

/*****************************************************************************\
*                                                                             *
*  Name   : amorph                                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1989-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "amorph.h"

#include <basis/array.cpp>
#include <basis/function.h>
#include <basis/guards.h>
#include <basis/utility.h>

//#define DEBUG_AMORPH
  // uncomment to enable more testing, as well as complaints on errors.

#undef static_class_name
#define static_class_name() istring("amorph")
  // used in bounds_halt macro.

#undef AMO_ALERT
#ifdef DEBUG_AMORPH
  #include <basis/istring.h>
  #define AMO_ALERT(a, b, c) continuable_error(a, b, c)
  #define CHECK_FIELDS check_fields(func)
#else
  #define AMO_ALERT(a1, a2, a3) {}
  #define CHECK_FIELDS {}
#endif

////////////////////////////////////////////////////////////////////////////

template <class contents>
amorph<contents>::amorph(int elements)
: array<contents *>(elements, NIL, array<contents *>::SIMPLE_COPY
    | array<contents *>::EXPONE),
  _fields_used(0)
{
  #define func "constructor"
  set_nil(0, elements - 1);
  CHECK_FIELDS;
  #undef func
}

template <class contents>
amorph<contents>::~amorph()
{
  #define func "destructor"
  CHECK_FIELDS;
  clear_all();
  #undef func
}

template <class contents>
void amorph<contents>::set_nil(int start, int end)
{
  for (int i = start; i <= end; i++)
    array<contents *>::put(i, (contents *)NIL);
}

template <class contents>
void amorph<contents>::check_fields(const char *where) const
{
  #define func "check_fields"
  int counter = 0;
  for (int i = 0; i < elements(); i++)
    if (array<contents *>::get(i)) counter++;
  if (_fields_used != counter)
    AMO_ALERT("amorph", isprintf("check_fields for %s", where).s(),
        "error in _fields_used count");
  #undef func
}

template <class contents>
void amorph<contents>::reset(int new_maximum)
{
  #define func "reset"
  CHECK_FIELDS;
  adjust(new_maximum);
  clear_all();
  #undef func
}

template <class contents>
void amorph<contents>::clear_all()
{
  #define func "clear_all"
  CHECK_FIELDS;
  for (int i = 0; i < elements(); i++) clear(i);
  #undef func
}

template <class contents>
outcome amorph<contents>::append(const contents *data)
{
  #define func "append"
  CHECK_FIELDS;
  adjust(elements() + 1);
  return put(elements() - 1, (contents *)data);
  #undef func
}

template <class contents>
const contents *amorph<contents>::get(int field) const
{
  #define func "get"
  CHECK_FIELDS;
  bounds_return(field, 0, elements() - 1, NIL);
  return array<contents *>::observe()[field];
  #undef func
}

template <class contents>
void amorph<contents>::adjust(int new_maximum)
{
  #define func "adjust"
  CHECK_FIELDS;
  if (new_maximum < 0) return;  // bad input here.
  int old_max = elements();
  if (new_maximum == old_max) return;  // nothing to do.
  if (new_maximum < old_max) {
    // removes the elements beyond the new size of the amorph.
    zap(new_maximum, old_max - 1);
    // we're done tuning it.
    return;
  }

  // we get to here if we need more space than we used to.
  int new_fields = new_maximum - old_max;

  array<contents *>::insert(old_max, new_fields);
  for (int i = old_max; i < new_maximum; i++) {
    array<contents *>::put(i, NIL);
  }
  #undef func
}

template <class contents>
outcome amorph<contents>::insert(int position, int lines_to_add)
{
  #define func "insert"
  CHECK_FIELDS;
  bounds_return(position, 0, elements(), common::OUT_OF_RANGE);
  outcome outcome = array<contents *>::insert(position, lines_to_add);
  if (outcome != common::OKAY) return common::OUT_OF_RANGE;
  set_nil(position, position + lines_to_add - 1);
  return common::OKAY;
  #undef func
}

template <class contents>
outcome amorph<contents>::zap(int start_index, int end_index)
{
  #define func "zap"
  CHECK_FIELDS;
  bounds_return(start_index, 0, elements() - 1, common::OUT_OF_RANGE);
  bounds_return(end_index, 0, elements() - 1, common::OUT_OF_RANGE);
  if (end_index < start_index) return common::OKAY;
  for (int i = start_index; i <= end_index;  i++) clear(i);
  outcome outcome = array<contents *>::zap(start_index, end_index);
  return (outcome==common::OKAY? common::OKAY : common::OUT_OF_RANGE);
  #undef func
}

template <class contents>
outcome amorph<contents>::put(int field, const contents *data)
{
  #define func "put"
  CHECK_FIELDS;
  bounds_return(field, 0, elements() - 1, common::OUT_OF_RANGE);
  contents *to_whack = acquire(field);
  WHACK(to_whack);
  if (data) {
    array<contents *>::access()[field] = (contents *)data;
    _fields_used++; 
  }
  return common::OKAY;
  #undef func
}

template <class contents>
int amorph<contents>::find_empty(outcome &outcome) const
{
  #define func "find_empty"
  CHECK_FIELDS;
  if (_fields_used == elements()) { outcome = common::IS_FULL; return 0; }
  for (int i = 0; i < elements(); i++)
    if (!array<contents *>::get(i)) { outcome = common::OKAY; return i; }
  AMO_ALERT("amorph", "empty", "_fields_used is incorrect");
  return common::IS_FULL;
  #undef func
}

template <class contents>
const contents *amorph<contents>::next_valid(int &field) const
{
  #define func "next_valid"
  CHECK_FIELDS;
  bounds_return(field, 0, elements() - 1, NIL);
  for (int i = field; i < elements(); i++)
    if (array<contents *>::get(i)) {
      field = i;
      return array<contents *>::get(i);
    }
  return NIL;
  #undef func
}

template <class contents>
outcome amorph<contents>::clear(int field)
{
  #define func "clear"
  CHECK_FIELDS;
  return this->put(field, NIL);
  #undef func
}

template <class contents>
contents *amorph<contents>::acquire(int field)
{
  #define func "acquire"
  CHECK_FIELDS;
  contents *to_return = borrow(field);
  if (to_return) {
    _fields_used--;
    array<contents *>::access()[field] = NIL;
  }
  return to_return;
  #undef func
}

template <class contents>
int amorph<contents>::find(const contents *to_locate, outcome &outcome)
{
  #define func "find"
  CHECK_FIELDS;
  if (!_fields_used) { outcome = common::NOT_FOUND; return 0; }
  for (int i = 0; i < elements(); i++) {
    if (array<contents *>::get(i) == to_locate) {
      outcome = common::OKAY;
      return i; 
    }
  }
  outcome = common::NOT_FOUND;
  return 0;
  #undef func
}

template <class contents>
contents *amorph<contents>::borrow(int field)
{
  #define func "borrow"
  CHECK_FIELDS;
  bounds_return(field, 0, elements() - 1, NIL);
  return array<contents *>::access()[field];
  #undef func
}

template <class contents>
void amorph<contents>::swap_contents(amorph<contents> &other)
{
  #define func "swap_contents"
  CHECK_FIELDS;
  int hold_fields = _fields_used;
  _fields_used = other._fields_used;
  other._fields_used = hold_fields;
  this->array<contents *>::swap_contents(other);
  #undef func
}

#undef static_class_name

#endif

