#ifndef DEEP_ARRAY_IMPLEMENTATION
#define DEEP_ARRAY_IMPLEMENTATION

/*****************************************************************************\
*                                                                             *
*  Name   : deep_array                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2006-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "amorph.cpp"
#include "deep_array.h"

#define static_class_name() istring("deep_array")

template <class contents>
deep_array<contents>::deep_array(int number)
: _real_list(new amorph<contents>(number))
{
}

template <class contents>
deep_array<contents>::deep_array(const deep_array<contents> &copy_from)
: _real_list(new amorph<contents>(copy_from.length()))
{
  for (int i = 0; i < copy_from.length(); i++) {
    const contents *found = copy_from.restore(i);
    if (found) put(i, *found);
  }
}

template <class contents>
deep_array<contents>::~deep_array()  
{ WHACK(_real_list); }

template <class contents>
void deep_array<contents>::reset(int number)
{ _real_list->reset(number); }

template <class contents>
int deep_array<contents>::length() const
{ return _real_list->elements(); }

template <class contents>
int deep_array<contents>::last() const
{ return _real_list->elements() - 1; }

template <class contents>
outcome deep_array<contents>::insert(int index, int new_indices)
{ return _real_list->insert(index, new_indices); }

template <class contents>
outcome deep_array<contents>::zap(int start, int end)
{ return _real_list->zap(start, end); }

//hmmm: this is a no-op.  is that a problem?
template <class contents>
outcome deep_array<contents>::shrink() { return common::OKAY; }

template <class contents>
void deep_array<contents>::swap_contents(deep_array<contents> &other)
{ _real_list->swap_contents(*other._real_list); }

template <class contents>
outcome deep_array<contents>::resize(int number, common::how_to_copy way)
{
  if (number < 0) number = 0;  // stop that noise right now.
  if (number == length()) return common::OKAY;
  if (way == common::DONT_COPY) {
    // they're not keeping any contents so this is just a reset.
    reset(number);
    return common::OKAY;
  }

  if (way == common::NEW_AT_END) {
    // some stuff got lopped off or added on the end so we'll just adjust.
    _real_list->adjust(number);
    return common::OKAY;
  }
  // the way was new at beginning.  let's see if we're shrinking or not.
  int old_len = length();
  if (old_len > number) {
    // the size of the array is shrinking.  keep the suffix.
    _real_list->zap(0, old_len - number - 1);
  } else {
    // the size is increasing.  add new stuff at the beginning.
    _real_list->insert(0, number - old_len);
  }
  return common::OKAY;
}

template <class contents>
deep_array<contents> &deep_array<contents>::operator =
    (const deep_array<contents> &copy_from)
{
  if (this == &copy_from) return *this;
  reset(copy_from.length());
  for (int i = 0; i < copy_from.length(); i++) {
    const contents *found = copy_from.restore(i);
    if (found) put(i, *found);
    else store(i, NIL);
  }
  return *this;
}

template <class contents>
const contents &deep_array<contents>::get(int index) const
{
  #define func "get"
  bounds_halt(index, 0, this->last(), bogonic<contents>());
  const contents *to_return = _real_list->get(index);
  if (to_return) return *to_return;
  // we can't just store a new item if the spot is missing, since this is
  // a const method.
  return bogonic<contents>();
  #undef func
}

template <class contents>
contents &deep_array<contents>::use(int index)
{
  #define func "use"
  bounds_halt(index, 0, this->last(), bogonic<contents>());
  if (!_real_list->borrow(index)) store(index, new contents);
  return *_real_list->borrow(index);
  #undef func
}

template <class contents>
outcome deep_array<contents>::store(int index, contents *to_store)
{
  #define func "store"
  bounds_halt(index, 0, this->last(), common::BAD_INPUT);
  _real_list->put(index, to_store);
  return common::OKAY;
  #undef func
}

template <class contents>
const contents *deep_array<contents>::restore(int index) const
{
  #define func "restore"
  bounds_halt(index, 0, this->last(), NIL);
  return _real_list->get(index);
  #undef func
}

template <class contents>
contents *deep_array<contents>::acquire(int index)
{
  #define func "acquire"
  bounds_halt(index, 0, this->last(), NIL);
  return _real_list->acquire(index);
  #undef func
}

template <class contents>
outcome deep_array<contents>::put(int index, const contents &to_put)
{
  #define func "put"
  bounds_halt(index, 0, this->last(), common::BAD_INPUT);
  if (!_real_list->get(index)) _real_list->put(index, new contents(to_put));
  else *_real_list->borrow(index) = to_put;
  return common::OKAY;
  #undef func
}

template <class contents>
deep_array<contents> deep_array<contents>::concatenation
    (const deep_array<contents> &to_concatenate) const
{
  deep_array<contents> to_return = *this;
  to_return.concatenate(to_concatenate);
  return to_return;
}

template <class contents>
deep_array<contents> deep_array<contents>::concatenation
    (const contents &to_concatenate) const
{
  deep_array<contents> to_return = *this;
  to_return.concatenate(to_concatenate);
  return to_return;
}

template <class contents>
deep_array<contents> &deep_array<contents>::concatenate
    (const deep_array<contents> &to_concatenate)
{
  if (!to_concatenate.length()) return *this;
  if (this == &to_concatenate)
    return concatenate(deep_array<contents>(*this));
  int old_len = length();
  _real_list->adjust(old_len + to_concatenate.length());
  for (int i = 0; i < to_concatenate.length(); i++) {
    const contents *found = to_concatenate.restore(i);
    if (found) put(old_len + i, *found);
    else store(old_len + i, NIL);
  }
  return *this;
}

template <class contents>
deep_array<contents> &deep_array<contents>::concatenate
    (const contents &to_concatenate)
{
  _real_list->adjust(length() + 1);
  put(last(), to_concatenate);
  return *this;
}

template <class contents>
deep_array<contents> deep_array<contents>::subarray(int start, int end) const
{
  #define func "subarray"
  if (start > end) return deep_array<contents>();
  bounds_return(start, 0, this->last(), deep_array<contents>());
  bounds_return(end, 0, this->last(), deep_array<contents>());
  deep_array<contents> to_return(end - start + 1);
  for (int i = start; i <= end; i++) {
    const contents *found = restore(i);
    if (found) to_return.put(i - start, *found);
  }
  return to_return;
  #undef func
}

template <class contents>
outcome deep_array<contents>::overwrite(int index,
    const deep_array<contents> &write_with, int count)
{
  #define func "overwrite"
  if (!count) return common::OKAY;
  if ( (this == &write_with) || !this->length() || !write_with.length()
      || negative(count) )
    return common::BAD_INPUT;
  bounds_halt(index, 0, this->last(), common::OUT_OF_RANGE);
  if (count > write_with.length()) count = write_with.length();
  if (index > this->length() - count) count = this->length() - index;
  for (int i = index; i < index + count; i++) {
    const contents *found = write_with.restore(i - index);
    if (found) put(i, *found);
    else store(i, NIL);
  }
  return common::OKAY;
  #undef func
}

#undef static_class_name

#endif // implementation guard.
