#ifndef DEEP_ARRAY_CLASS
#define DEEP_ARRAY_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : deep_array                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2006-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "amorph.h"

//! This object mimics a C array but tracks its contents as deep pointers.
/*!
  This object is intended to provide an interface similar to the array
  template in basis, but it instead stores all contents as pointers rather
  than as a contiguous array.  This saves a lot of copying costs for
  complicated objects.  A method is provided that allows object pointers 
  to be stored directly (see store(), restore() and acquire()).
*/

template <class contents>
class deep_array
{
public:
  deep_array(int number = 0);
    //!< Constructs a deep_array with room for "number" objects.

  deep_array(const deep_array<contents> &copy_from);
    //!< copies the contents & sizing information from "copy_from".

  ~deep_array();  //!< destroys the memory allocated for the objects.

  bool valid() const { return !!_real_list; }
    //!< Reports whether the construction succeeded and allocated memory.

  void reset(int number = 0);
    //!< Resizes this array to the "number" of elements specified.

  deep_array &operator = (const deep_array<contents> &copy_from);
    //!< Copies the deep_array in "copy_from" into this.

  int length() const;
    //!< Returns the current reported length of the allocated C array.

  int last() const;
    //!< Returns the last valid element in the deep_array.

  const contents &get(int index) const;
    //!< Accesses individual objects stored in "this" at the "index" position.
    /*!< If the index is out of range, then a bogus reference (to internally
    held garbage) is returned. */
  contents &use(int index);
    //!< A non-constant version of get(); the returned object can be modified.

  inline const contents &operator [] (int index) const { return get(index); }
    //!< Synonym for get that provides the expected array indexing syntax.
  inline contents &operator [] (int index) { return use(index); }
    //!< Synonym for use that provides the expected array indexing syntax.

  outcome put(int index, const contents &to_put);
    //!< Stores an object at the index "index" in the deep_array.
    /*!< The outcome is "OUT_OF_RANGE" if the index does not exist. */

  // store, retrieve and acquire are specializations that operate on pointers.

  outcome store(int index, contents *to_store);
    //!< Drops the "to_store" pointer directly into our list at "index".
    /*!< Be careful!  This requires that the deep_array take over ownership
    of the "to_store" object.  The only valid way to get that object back
    is to acquire() it. */
  const contents *restore(int index) const;
    //!< Pulls the object at "index" back out of the array.
    /*!< This leaves the slot at index empty.  The returned object is now
    the caller's responsibility and must eventually be destroyed. */
  contents *acquire(int index);
    //!< Retrieves the object at "index" and removes the slot in the array.
    /*!< This is just like restore() except that the "index" slot will be
    removed from the array.  It's the equivalent of doing restore() followed
    by a zap(). */

  deep_array concatenation(const deep_array &to_concatenate) const;
    //!< Returns the concatenation of "this" and the array "to_concatenate".
  deep_array concatenation(const contents &to_concatenate) const;
    //!< Returns the concatenation of "this" and the object "to_concatenate".

  deep_array &concatenate(const deep_array &to_concatenate);
    //!< Appends the array "to_concatenate" onto "this" and returns "this".
  deep_array &concatenate(const contents &to_concatenate);
    //!< Appends the object "to_concatenate" onto "this" and returns "this".

  inline deep_array operator + (const deep_array &to_cat) const
          { return concatenation(to_cat); }
    //!< Synonym for concatenation.
  inline deep_array operator + (const contents &to_concatenate) const
          { return concatenation(to_concatenate); }
    //!< Synonym for concatenation.
  inline deep_array &operator += (const deep_array &to_concatenate)
          { return concatenate(to_concatenate); }
    //!< Synonym for concatenate that modifies "this".
  inline deep_array &operator += (const contents &to_concatenate)
          { return concatenate(to_concatenate); }
    //!< Synonym for concatenate that modifies "this".

  deep_array subarray(int start, int end) const;
    //!< Returns the array segment between the indices "start" and "end".
    /*!< This is all characters from "start" to "end" inclusively, as long as
    those values are valid for this array.   Even then, an intelligent default
    is usually assumed if the indices are out of range. */

  outcome insert(int index, int new_indices);
    //!< Adds "new_indices" new positions for objects into the array at "index".

  outcome overwrite(int index, const deep_array &write_with, int count = -1);
    //!< Stores the array "write_with" into the current array at the "index".
    /*!< The current contents are overwritten with "write_with".  If the
    index is invalid, then OUT_OF_RANGE is returned.  If the "write_with"
    array cannot fit due to the boundaries of "this" array, then only the
    portion that can fit is used.  If "count" is negative, the whole
    "write_with" array is used; otherwise, only "count" elements are used. */

  outcome resize(int new_size, common::how_to_copy way = common::NEW_AT_END);
    //!< Changes the size of the array to "new_size".
    /*!< If "way" is NEW_AT_END and the array grows, then new space is added
    at the end and the prefix of the array is the same as the old array.  If
    the "way" is NEW_AT_END, but the array shrinks, then the new array is a
    prefix of the old array.  If "way" is NEW_AT_BEGINNING and the array
    grows, then the suffix of the array is the same as the old one and the
    space is added at the beginning.  if the "way" is NEW_AT_BEGINNING but the
    array shrinks, then the new array is a suffix of the old array.  if "way"
    is DONT_COPY, then the old contents are not copied. */

  outcome zap(int start, int end);
    //!< Deletes from "this" the objects inclusively between "start" and "end".
    /*!< C-array conventions are used (0 through length()-1 are valid if
    length() > 0).  If either index is out of range, then a default is
    assumed. */

  outcome shrink();
    //!< Cuts loose any allocated space that is beyond the real length.

  void swap_contents(deep_array<contents> &other);
    //!< Exchanges the contents of "this" and "other".
    /*!< No validation is performed but this should always succeed given
    deep_arrays that were constructed properly. */

private:
  amorph<contents> *_real_list;  //!< stores the elements as pointers.
};

#endif

