#ifndef GUARDED_VALUE_CLASS
#define GUARDED_VALUE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : guarded_value                                                     *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Remembers whether a value has actually been initialized.
/*!
  Designed to act identically to the contained class except that it
  tracks whether the object held is initialized yet or not.  The templated
  class must support a copy constructor.
*/

template <class contents>
class guarded_value
{
public:
  guarded_value() : _initialized(false) {}
    //!< without an initial value, the guarded_value is considered invalid.
  guarded_value(const contents &def) : _value(def), _initialized(true) {}
    //!< the guarded_value with an initial value is considered valid.
  guarded_value(const guarded_value &to_copy)
      : _value(to_copy._value), _initialized(to_copy._initialized) {}
    //!< Copy constructor makes "this" into a copy of "to_copy".
    /*!< copying a guarded_value keeps all its characteristics and relies on
    a properly implemented copy constructor for the <contents>. */

  bool initialized() const { return _initialized; }
    //!< if the object was initialized and not reset, then this returns true.
    /*!< if this returns false, it is a clear sign that the object held
    cannot be used for any informational purposes. */

  void reset() { _initialized = false; }
    //!< marks the value held by this class as invalid.
    /*!< if the object had any special destruction requirements (for example,
    held pointers...), then those need to be taken care of before the reset. */

  guarded_value &operator =(const guarded_value &t_c)
      { if (this == &t_c) return *this; else return *this = t_c._value; }
    //!< assignment operator that takes another guarded value.
  guarded_value &operator =(const contents &t_c)
      { _value = t_c; _initialized = true; return *this; } 
    //!< assignment operator that takes one of the templated objects.

  operator contents() const { return _value; }
    //!< Operator that returns the held object.
    /*!< This conversion (and access()) yields the type of object held here,
    which re-enables the guarded_value to participate in the normal uses of
    the templated type. */
  contents &access() { return _value; }
    //!< A synonym for the contents() operator.

private:
  bool _initialized;  //!< flag that tracks whether this object is valid or not.
  contents _value;  //!< the actual contents.
};

#endif

