#ifndef MANAGED_OBJECT_CLASS
#define MANAGED_OBJECT_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : managed_object                                                    *
*  Author : Aaron Buchanan                                                    *
*                                                                             *
*******************************************************************************
* Copyright (c) 2001-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/definitions.h>

// forward.
template <class c> class object_storer;

//! Reference counts objects and destroys them automatically.
/*!
  Provides an object management system that allows for references to an
  original stored object to be obtained, and the original object is not
  deleted until all references are gone.
*/

template <class contents>
class managed_object
{
public:
    managed_object();
        /*!< Having a default constructor for a managed object does not make a
        lot of sense, since the whole point is you are managing another
        object.  However, there are circumstances that make it useful. */

    managed_object(contents *&_object);
        //!< Pass a pointer of the object to be stored.
        /*!< The managed_object class takes control of the pointer and sets
        the reference null before returning.  The original object is wrapped
        in the ObjectStorer class so that the number of references can be
        counted. */

    managed_object(const managed_object &_rhs);
        //!< Moves the contained object in _rhs into a new managed_object.
    managed_object &operator =(const managed_object &rhs);
        //!< Moves the contents of _rhs into this object.
        /*!< Causes this object storer to have its reference count decremented,
        then the object storer pointer from the rhs is copied and its reference
        count is incremented. */

    virtual ~managed_object();
        //!< Releases the object contained here.
        /*!< Doesn't delete the actual object, just tells the storer to
        decrement its reference count.  When all references are gone the
        original object will be deleted. */

    bool operator ==(const managed_object &_compare) const;
        /*!< The equality operator for the managed_object invokes the equality 
        operator on its stored pointer. */

    managed_object<contents> *new_reference();
        //!< Return a new reference to the managed object class.
        /*!< This uses the private constructor that simply copies the object
        storer pointer.  We don't want to copy the actual stored object.
        The returned pointer must be deleted when it is finished being used.
        This can be achieved using the copy constructor also. */

    contents *access();
        //!< access the originally stored object

    const contents *observe() const;
        //!< look at the originally stored object

private:
    object_storer<contents> *storer;
};

#endif

