#ifndef MATRIX_CLASS
#define MATRIX_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : matrix                                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/array.h>

//! Represents a two-dimensional array of objects.
/*!
  The indices range from zero to (maximum_value - 1) for rows and columns.
*/

template <class contents>
class matrix : protected array<contents>
{
public:
  matrix(int rows = 0, int cols = 0, contents *data = NIL);
    //!< the "data" array must have at least "rows" * "cols" contents in it.
  matrix(const matrix &to_copy);

  inline ~matrix() {}

  inline int rows() const { return _rows; }
  inline int columns() const { return _cols; }

  matrix &operator = (const matrix &to_copy);

  contents &get(int row, int column);
  const contents &get(int row, int column) const;
    //!< retrieves the contents from the specified "row" and "column".

  bool put(int row, int column, const contents &to_put);
    //!< stores "to_put" into the matrix at "row" and "column".

  contents *operator[] (int row);
    //!< dangerous: indexes by "row" into the underlying contents.
    /*!< the result can be used as a pointer to the whole row of data, or it
    can be indexed into by column to find a row/column position.  this is
    dangerous if one is not careful about ensuring that the column used is
    within bounds.  if the "row" parameter is out of bounds, then NIL is
    returned. */
  const contents *operator[] (int row) const;
    //!< dangerous: constant peek into data for "row" in underlying contents.

  matrix submatrix(int row, int column, int rows, int columns) const;
    //!< returns a submatrix of this one starting at "row" and "column".
    /*!< The returned matrix should contain "rows" rows and "columns" columns.
    if the size or position are out of bounds, an empty matrix is returned. */
  void submatrix(matrix &sub, int row, int column, int rows, int cols) const;
    //!< like submatrix() above, but stores into the "sub" parameter.

  matrix transpose() const;
    //!< provides the transposed form of this matrix.
  void transpose(matrix &resultant) const;
    //!< stores the transpose of this matrix in "resultant".

  array<contents> get_row(int row);
    //!< return the vector at the "row", or an empty array if "row" is invalid.
  array<contents> get_column(int column);
    //!< return the vector at the "column", or an empty array if it's invalid.

  bool insert_row(int before_row);
    //!< inserts a row before the "before_" parameter.
    /*!< all existing data is preserved in the matrix, although it may get
    moved over depending on where it's located with respect to "before_row". */
  bool insert_column(int before_column);
    //!< inserts a column before the "before_" parameter.

  void reset(int rows = 0, int columns = 0);
    //!< empties the matrix and changes its size.

  void redimension(int new_rows, int new_columns);
    //!< changes the size to contain "new_rows" by "new_columns" elements.
    /*!< data that was held previously stays in the array as long as its row
    and column indices are still valid. */

  bool zap_row(int row_to_zap);
  bool zap_column(int column_to_zap);
    //!< removes a row or column from the matrix.
    /*!< the matrix becomes one row or column less than before and all data
    from the deleted vector is lost. */

private:
  int _rows;  //!< number of rows in the matrix.
  int _cols;  //!< number of columns in the matrix.

  int compute_index(int row, int column) const;
    //!< returns the flat index that corresponds to the "row" and "column".
};

#endif

