#ifndef FLOAT_PLUS_CLASS
#define FLOAT_PLUS_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : floating point extension class                                    *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/definitions.h>

//! An extension to floating point primitives providing approximate equality.
/*!
  Allows a programmer to ignore issues of rounding errors on floating point
  numbers by specifying that two floating point numbers are equivalent if
  they are equal within a small number "delta".  This can help to eliminate
  errors in floating point logic.
*/

template <class contents>
class float_plus
{
public:
  #define DEFAULT_DELTA 0.0001
    /*!< the delta is the acceptable amount of difference between two floating
    point numbers that are considered equivalent by this class.  if they
    differ by more than that, they are considered non-equivalent (and
    hence must be greater than or less than each other). */

  //! initializes using "init" as the initial value and equality within "delta".
  float_plus(contents init = 0.0, contents delta = DEFAULT_DELTA)
              : _value(init), _delta(delta) {}

  //! initializes this from "to_copy".
  float_plus(const float_plus &to_copy)
              : _value(to_copy._value), _delta(to_copy._delta) {}

  //! standard assignment operator.
  float_plus &operator = (const float_plus &cp)
              { _value = cp._value; _delta = cp._delta; return *this; }

  inline contents value() const { return _value; }
    //!< observes the value held in this.
  inline operator contents () const { return _value; }
    //!< observes the value held in this.

  inline contents delta() const { return _delta; }
    //!< observes the precision for equality comparisons.
  inline void delta(contents new_delta) { _delta = new_delta; }
    //!< modifies the precision for equality comparisons.

  bool operator == (const float_plus &to_compare) { return d_eq(to_compare); }
    //!< returns true if this equals "to_compare" within the "delta" precision.
  bool operator != (const float_plus &to_compare) { return !d_eq(to_compare); }
    //!< returns true if this is not equal to "to_compare" within "delta".
  bool operator < (const float_plus &to_compare)
              { return !d_eq(to_compare) && (_value < to_compare); }
    //!< returns true if this is less than "to_compare".
  bool operator > (const float_plus &to_compare)
              { return !d_eq(to_compare) && (_value > to_compare); }
    //!< returns true if this is greater than "to_compare".
  bool operator <= (const float_plus &to_compare)
              { return _value < to_compare || d_eq(to_compare); }
    //!< true if this is less than or equal to "to_compare" within "delta".
  bool operator >= (const float_plus &to_compare)
              { return _value > to_compare || d_eq(to_compare); }
    //!< true if this is greater than or equal to "to_compare" within "delta".

private:
  contents _value;  //!< the contained floating point value.
  contents _delta;  //!< the offset within which equality is still granted.

  //! returns true if "to_compare" is within the delta of this.
  bool d_eq(contents to_compare) {
    contents diff = absolute_value(_value - to_compare);
    return diff < absolute_value(_delta); 
  }
};

#endif

