/*****************************************************************************\
*                                                                             *
*  Name   : time_stamp                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1995-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "time_stamp.h"

#include <basis/function.h>
#include <basis/istring.h>
#include <basis/mutex.h>
#include <basis/portable.h>
#include <basis/utility.h>
#include <data_struct/static_memory_gremlin.h>

time_stamp::time_stamp() : _stamp(0) { fill_in_time(); }

time_stamp::time_stamp(time_representation offset)
: _stamp(0) { reset(offset); }

void time_stamp::reset() { fill_in_time(); }

istring time_stamp::text_form(stamp_display_style style) const
{
  time_representation stump = _stamp;
  bool past = false;
  if (style == STAMP_RELATIVE) {
    // adjust the returned time by subtracting the current time.
    stump -= get_time_now();
    if (negative(stump)) {
      // if we're negative, just note that the stamp is in the past.
      past = true;
      stump = absolute_value(stump);
    }
  }
  time_representation divisor = 3600 * SECOND_ms;
  u_int hours = u_int(stump / divisor);
  stump -= divisor * time_representation(hours);
  divisor /= 60;
  u_int minutes = u_int(stump / divisor);
  stump -= divisor * time_representation(minutes);
  divisor /= 60;
  u_int seconds = u_int(stump / divisor);
  stump -= divisor * time_representation(seconds);
  u_int milliseconds = u_int(stump);

  istring to_return;
  bool did_hours = false;
  if (hours) {
    to_return += istring(istring::SPRINTF, "%uh:", hours);
    did_hours = true;
  }
  if (minutes || did_hours)
    to_return += istring(istring::SPRINTF, "%02um:", minutes);
  to_return += istring(istring::SPRINTF, "%02us.%03u", seconds, milliseconds);
  if (style == STAMP_RELATIVE) {
    if (past) to_return += " ago";
    else to_return += " from now";
  }
  return to_return;
}

void time_stamp::fill_in_time()
{
  time_representation current = get_time_now();
  _stamp = current;  // reset our own time now.
}

void time_stamp::reset(time_representation offset)
{
  fill_in_time();
  _stamp += offset;
}

#ifdef __WIN32__
  #include <mmsystem.h>
#else
  // unix timing information assumed for all others.
  #include <sys/times.h>
#endif

class rollover_record : public virtual object_base
{
public:
  int _rollovers;  // counts number of rollovers in cruddy uptime.
  u_int _last_clock;  // tracks last clock we issued.
  time_stamp::time_representation _MAX;
    // the largest value the time rep can hold.
  mutex _stamp_lock;  // protects access.

  IMPLEMENT_CLASS_NAME("rollover_record");

  rollover_record()
  : _rollovers(0),
    _last_clock(0), 
#ifdef __WIN32__
    _MAX(time_stamp::time_representation(MAXINT) + time_stamp::time_representation(MINLONG)),
      // calculate wrap point for u_int.
#else
    _MAX(time_stamp::time_representation(MAXINT)),
      // calculate wrap point for signed int.
#endif
    _stamp_lock()
  {
  }

};

SAFE_STATIC(rollover_record, time_stamp::rollover_rover, );
  // protect our rollover record from concurrent construction.

time_stamp::time_representation time_stamp::get_time_now()
{
  rollover_record &rover = rollover_rover();
  auto_synchronizer l(rover._stamp_lock);
    // protect the changes we're making below.  NOTE: this must reside where it
    // is because it also protects the check on system time; if two threads
    // come in with different times and then check against the rollover record,
    // the can get totally confused about whether there was a rollover.  the
    // threads' concurrent access must be prevented before they get the system
    // time.

  u_int current_time = portable::system_uptime();
    // get the time since system startup.

  // update the global record of rollovers if necessary.
  if (current_time < rover._last_clock) {
    // detected a roll-over.
    rover._rollovers++;
  }
  // record this call's time for later.
  rover._last_clock = current_time;
  // count current number of rollovers and add "now".
  return time_representation(rover._rollovers)
      * rover._MAX + current_time;
}

