#ifndef TIME_STAMP_CLASS
#define TIME_STAMP_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : time_stamp                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1995-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "mechanisms_dll.h"

// forward.
class rollover_record;

//! Represents a point in time relative to the operating system startup time.
/*!
  This duration is measured in milliseconds.  This class provides a handy way
  of measuring relative durations at the millisecond time scale.
  Unfortunately, operating systems that reckon their millisecond uptime in
  32 bit integers will suffer from a rollover problem every 49 days or so,
  but this class corrects this issue.
*/

class MECHANISMS_CLASS_STYLE time_stamp
{
public:
  typedef double time_representation;
    //!< the representation of time for this universe, measured in milliseconds.

  time_stamp();
    //!< creates a time_stamp containing the current time.

  time_stamp(time_representation offset);
    //!< creates a stamp after the current time by "offset" milliseconds.
    /*!< negative offsets are allowed, in which case the stamp will be
    prior to the current time. */

  void reset();
    //!< sets the stamp time back to now.
  void reset(time_representation offset);
    //!< sets the stamp forward by "offset" similar to the second constructor.

  time_representation value() const { return _stamp; }
    //!< returns the time_stamp in terms of the lower level type.

  enum stamp_display_style { STAMP_RELATIVE, STAMP_ABSOLUTE };

  istring text_form(stamp_display_style style = STAMP_RELATIVE) const;
    //!< returns a simple textual representation of the time_stamp.
    /*!< if the "style" is ABSOLUTE, then the stamp is shown in H:M:S.ms
    form based on the system uptime.  if the "style" is RELATIVE, then the
    stamp is shown as an offset from now. */

  // standard operators: keep in mind that time is represented by an ever
  // increasing number.  so, if a value A is less than a value B, that means
  // that A is older than B, since it occurred at an earlier time.
  inline bool operator < (const time_stamp &that) const
          { return _stamp < that._stamp; }
  inline bool operator > (const time_stamp &that) const
          { return _stamp > that._stamp; }
  inline bool operator <= (const time_stamp &that) const
          { return _stamp <= that._stamp; }
  inline bool operator >= (const time_stamp &that) const
          { return _stamp >= that._stamp; }
  inline bool operator != (const time_stamp &that) const
          { return _stamp != that._stamp; }
  inline bool operator == (const time_stamp &that) const
          { return _stamp == that._stamp; }

private:
  time_representation _stamp;  //!< the low-level time stamp is held here.

  void fill_in_time();  //!< stuffs the current time into the held value.

  static time_representation get_time_now();
    //!< platform specific function that gets uptime.

  static rollover_record &rollover_rover();
};

#endif

