#ifndef APPLICATION_SHELL_CLASS
#define APPLICATION_SHELL_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : application_shell                                                 *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! The application_shell is a base object for console programs.
/*!
  It is generally used in that context (console mode), but can be employed
  as the root of windowed programs also.  The application_shell provides a
  few features, such as logging functionality and INI file editing, that make
  it a bit easier than starting up a program from scratch every time.
*/

#include "application_base.h"

// forward.
class ini_configurator;

class OPSYSTEM_CLASS_STYLE application_shell : public application_base
{
public:  
  enum limits { LOG_FILE_SIZE = 1244688 };
    //!< this just defines the default for the log file size.
//hmmm: move the limit lower since file_logger also has this.

  application_shell(const istring &filename_root, int max_log = LOG_FILE_SIZE);
    //!< constructs an application_shell object given the program name.
    /*!< "filename_root" is used as the base part of the logging filename.
    the "max_log" parameter specifies how large the output log is allowed to
    grow; if it's zero, then no log will be created. */

  virtual ~application_shell();

  virtual int execute() = 0;
    //!< forwards application_base responsibility upwards to derived objects.

  chaos &randomizer() const { return *_rando; }
    //!< provides access to the random number generator owned by this app.

  istring filename_root() const;
    //!< returns the name of this object.
    /*!< the name is passed to the constructor and cannot be changed during
    execution of the application_shell.  the name should not contain any
    characters that are invalid for filenames, because it is also used for
    the base-name of the log file. */

  ini_configurator &ini() const;
    //!< allows access to the ini_configurator object.

  istring ini_name() const;
    //!< returns the filename of the INI file.

  istring log_name() const;
    //!< returns the filename of the log file.

  virtual outcome log(const istring &to_print);
    //!< logs a debugging line "to_print".
    /*!< this mirrors its output to the console when it logs to the file. */
  virtual outcome log(const istring &to_print, int filter);
    //!< logs a debugging line "to_print" using the "filter".
    /*!< the second version uses the filter value to assess whether to print
    the string or not.  the string will not print if that filter is not
    enabled for the program wide logger. */

  log_base &logger() const;
    //!< provides access to the combo_logger object that is used for debugging.
    /*!< this allows the user to manipulate filter values and so forth.  note
    that if you print directly to the logger, then you will not see this
    information on the console. */

  log_base *swap_logger(log_base *new_logger);
    //!< replaces the default log_base object with "new_logger".
    /*!< this allows a derived application_shell to use a different type of
    logging.  this also retasks the program wide logger to use the
    "new_logger".  the returned log_base object is the previously hooked up
    logger and it must be dealt with appropriately. */

  friend application_shell &operator << (application_shell &out,
          const istring &to_print);
    //!< mimics the iostream operation, but is a synonym for log().

private:
  chaos *_rando;  //!< random number generator.
  istring *_root_name;  //!< given to us as the program's name.
  ini_configurator *_ini;  //!< the ini editing support.

  static istring LOG_NAME(const istring &root);
    //!< helper function for creating right log file based on "root".

  // not applicable.
  application_shell(const application_shell &);
  application_shell &operator =(const application_shell &);
};

#endif // outer guard.

