/*****************************************************************************\
*                                                                             *
*  Name   : byte_filer                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "byte_filer.h"

#include <basis/common_templates.h>
#include <basis/istring.h>
#include <basis/portable.h>
#include <basis/utility.h>

#include <stdio.h>
#include <string.h>
#ifdef __UNIX__
  #include <unistd.h>
#endif
#ifdef __WIN32__
  #include <io.h>
#endif

class file_hider
{
public:
  FILE *fp;  // the real file pointer.

  file_hider() : fp(NIL) {}
};

////////////////////////////////////////////////////////////////////////////

byte_filer::byte_filer()
: _handle(new file_hider),
  _filename(new istring)
{}

byte_filer::byte_filer(const istring &filename, const istring &perms)
: _handle(new file_hider),
  _filename(new istring)
{ open(filename, perms); }

byte_filer::~byte_filer() { close(); WHACK(_handle); WHACK(_filename); }

istring byte_filer::filename() const { return *_filename; }

bool byte_filer::open(const istring &filename, const istring &perms)
{
  close();
  *_filename = filename;
  _handle->fp = filename.t()? fopen(filename.s(), perms.s()) : NIL;
  return good();
}

void byte_filer::close()
{
  *_filename = "";
  if (_handle->fp) fclose(_handle->fp);
  _handle->fp = NIL;
}

bool byte_filer::good() { return !!_handle->fp; }

int byte_filer::tell() { return !_handle->fp ? 0 : ::ftell(_handle->fp); }

void *byte_filer::file_handle() { return _handle->fp; }

bool byte_filer::eof() { return !_handle->fp ? true : !!feof(_handle->fp); }

int byte_filer::read(byte *buff, int size)
{ return !_handle->fp ? 0 : ::fread((char *)buff, 1, size, _handle->fp); }

int byte_filer::write(const byte *buff, int size)
{ return !_handle->fp ? 0 : ::fwrite((char *)buff, 1, size, _handle->fp); }

int byte_filer::read(byte_array &buff, int desired_size)
{
  buff.reset(desired_size);
  int to_return = read(buff.access(), desired_size); 
  buff.reset(to_return);
  return to_return;
}

int byte_filer::write(const byte_array &buff)
{ return write(buff.observe(), buff.length()); }

int byte_filer::length()
{
  int current_posn = tell();
  seek(0, FROM_END);  // jump to end of file.
  int file_size = tell();  // get position.
  seek(current_posn, FROM_START);  // jump back to previous place.
  return file_size;
}

int byte_filer::read(istring &s, int desired_size)
{
  s.pad(desired_size + 2);
  int found = read((byte *)s.observe(), desired_size);
  if (non_negative(found)) s[found] = '\0';
  s.shrink();
  return found;
}

int byte_filer::write(const istring &s, bool add_null)
{
  int len = s.length();
  if (add_null) len++;
  return write((byte *)s.observe(), len);
}

void byte_filer::flush()
{
  if (!_handle->fp) return;
  ::fflush(_handle->fp);
}

bool byte_filer::truncate()
{
  flush();
  int fnum = fileno(_handle->fp);
#ifdef __WIN32__
  return SetEndOfFile((HANDLE)_get_osfhandle(fnum));
#else
  return !ftruncate(fnum, tell());
#endif
}

bool byte_filer::seek(int where, origins origin)
{
  if (!_handle->fp) return false;
  int real_origin;
  if (origin == FROM_START) real_origin = SEEK_SET;
  else if (origin == FROM_END) real_origin = SEEK_END;
  else if (origin == FROM_CURRENT) real_origin = SEEK_CUR;
  else return false;  // not a valid choice.
  int ret = ::fseek(_handle->fp, where, real_origin);
  return !ret;
}

int byte_filer::getline(byte *buff, int desired_size)
{
  if (!_handle->fp) return 0;
  char *ret = ::fgets((char *)buff, desired_size, _handle->fp);
  return !ret? 0 : strlen((char *)buff) + 1;
}

int byte_filer::getline(byte_array &buff, int desired_size)
{
  buff.reset(desired_size + 1);
  return getline(buff.access(), desired_size);
}

int byte_filer::getline(istring &buff, int desired_size)
{
  buff.pad(desired_size + 1);
  int to_return = getline((byte *)buff.access(), desired_size);
  if (non_negative(to_return)) buff[to_return] = '\0';
  buff.shrink();
  return to_return;
}

