#ifndef DIRECTORY_CLASS
#define DIRECTORY_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : directory                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2001-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Implements a scanner that finds all filenames in the directory specified.

#include "opsysdll.h"

class OPSYSTEM_CLASS_STYLE directory
{
public:
  directory(const istring &path, const char *pattern = "*");
    //!< opens up the "path" specified and scans for files and subdirectories.
    /*!< if the location was accessible, then the good() method returns true.
    note that the "path" should just be a bare directory without any
    wildcards attached.  the "pattern" can be specified if you wish to
    strain out just a subset of the files in the directory.  it must meet
    the same requirements that the operating system places on wildcard
    patterns. */

  ~directory();

  bool good() const { return _scanned_okay; }
    //!< returns true if the directory existed and we were able to read its contents.

  const istring &path() const;
    //!< returns the directory that we manage.

  static istring absolute_path(const istring &relative_path);
    //!< returns the absolute path to a file with "relative_path".
    /*!< an empty string is returned on failure. */

  bool reset(const istring &path, const char *pattern = "*");
    //!< gets rid of any current files and rescans the directory at "path".
    /*!< a new "pattern" can be specified at this time also. */

  bool rescan();
    //!< reads our current directory's contents over again.

  const string_array &files() const;
    //!< returns the list of files that we found in this directory.
    /*!< these are all assumed to be located in our given path.  to find out
    more information about the files themselves, construct a filename object
    with the path() and the file of interest. */
    
  const string_array &directories() const;
    //!< these are the directory names from the folder.
    /*!< they can also be examined using the filename object.  note that
    this does not include the entry for the current directory (.) or the
    parent (..). */

  // static methods of general directory-related interest.

  static istring current();
    //!< returns the current directory, as reported by the operating system.

  static bool make_directory(const istring &path);
    //!< returns true if the directory "path" could be created.

  static bool remove_directory(const istring &path);
    //!< returns true if the directory "path" could be removed.

private:
  bool _scanned_okay;  //!< did this directory work out?
  istring *_path;  //!< the directory we're looking at.
  string_array *_files;  //!< the list of files.
  string_array *_folders;  //!< the list of directories.
  istring *_pattern;  //!< the pattern used to find the files.
};

#endif

