#ifndef FILE_LOGGER_CLASS
#define FILE_LOGGER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : file_logger                                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Enables the printing of information to a log file.
/*!
  The information can be conditionally printed using the filter support.
  The log file will automatically be truncated when it passes the size limit.
*/

#include "opsysdll.h"

#include <basis/log_base.h>

// forward.
class byte_filer;
class console_logger;

class OPSYSTEM_CLASS_STYLE file_logger : public log_base
{
public:
  file_logger();
    //!< creates a logger without a log file and with the default size limit.
    /*!< the log file name can be changed using filename(). */

  file_logger(const istring &filename, int limit = LOG_FILE_SIZE);
    //!< constructs a logger using the "filename" for output.
    /*!< there will be no logging if the "filename" is empty.  the "limit"
    specifies how large the log file can be (in bytes). */

  virtual ~file_logger();

  IMPLEMENT_CLASS_NAME("file_logger");

  enum limits {
    //! this just defines the default for the log file size.
#ifndef EMBEDDED_BUILD
    LOG_FILE_SIZE = 0x10F00D
#else
    LOG_FILE_SIZE = 0x10FAB
#endif
  };

  bool good() const;
    //!< returns true if the logger appears correctly hooked up to a file.
    /*!< note that we don't open the file when file_logger is constructed;
    it is only opened once the first logging is attempted. */

  bool reopen();
    //!< closes the current file and attempts to reopen it.
    /*!< this is handy if the original opening of the file failed. */

  outcome log(const istring &info, int filter = ALWAYS_PRINT);
    //!< writes information to the log file (if the filename is valid).
    /*!< the "filter" value is checked to see if it is in the current set
    of allowed filters.  a value of zero is always printed.  if the filename()
    has not been set, then the information is lost. */

  outcome log_bytes(const byte_array &to_log, int filter = ALWAYS_PRINT);
    //!< sends a stream of bytes "to_log" without interpretation into the log.
    /*!< if the "filter" is not enabled, then the info is just tossed out. */

  outcome format_bytes(const byte_array &to_log, int filter = ALWAYS_PRINT);
    //!< fancifully formats a stream of bytes "to_log" and sends them into log.

  istring name() const;
    //!< observes the filename where logged information is written.
  void name(const istring &new_name);
    //!< modifies the filename where logged information will be written.
    /*!< if "new_name" is blank, then the logged information will not
    be saved. */

  int limit() const { return _file_limit; }
    //!< observes the allowable size of the log file.
  void limit(int new_limit) { _file_limit = new_limit; }
    //!< modifies the allowable size of the log file.

  void flush();
    //!< causes any pending writes to be sent to the output file.

  void truncate(int new_size);
    //!< chops the file to ensure it doesn't go much over the file size limit.
    /*!< this can be used externally also, but be careful with it. */

private:
  istring *_filename;  //!< debugging output file.
  int _file_limit;  //!< maximum length of file before truncation.
  byte_filer *_outfile;  //!< the object that points at our output file.
  mutex *_flock;  //!< protects the file and other parameters.

  int size_reduction() const;
    //!< returns the size of the chunk to truncate from the file.
    /*!< this is a percentage of the maximum size allowed. */

  bool open_file();
    //!< if the opening of the file is successful, then true is returned.
    /*!< also, the _outfile member variable is non-zero. */

  void close_file();
    //!< shuts down the file, if any, we had opened for logging.

  // unavailable.
  file_logger(const file_logger &);
  file_logger &operator =(const file_logger &);
};

////////////////////////////////////////////////////////////////////////////

//!< returns a log file name for file_logger based on the program name.
/*!< for a program named myapp.exe, this will be in the form:
  {logging_dir}/myapp.log
*/
istring OPSYSTEM_FUNCTION_STYLE app_name_for_file_logger();

//!< retasks the program wide logger to use a new file_logger.
/*!< the original log_base object is returned and must be dealt with
appropriately. */
log_base OPSYSTEM_FUNCTION_STYLE *set_PW_logger_for_file
    (const istring &file_name, int limit = file_logger::LOG_FILE_SIZE);

//!< a macro that retasks the program-wide logger as a file_logger.
#ifndef OMIT_PROGRAM_WIDE_LOGGER
  #define SET_DEFAULT_FILE_LOGGER { \
    log_base *old_log = set_PW_logger_for_file(app_name_for_file_logger()); \
    WHACK(old_log); \
  }
#else
  #define SET_DEFAULT_FILE_LOGGER
#endif

////////////////////////////////////////////////////////////////////////////

//!< combines a file_logger with a console logger (see console_logger.h).
/*!< this will output both to the file and to the console. */

class OPSYSTEM_CLASS_STYLE combo_logger : public file_logger
{
public:
  combo_logger(const istring &filename, int limit = LOG_FILE_SIZE);
  virtual ~combo_logger();

  virtual outcome log(const istring &info, int filter = ALWAYS_PRINT);

  // overrides that enforce properties for both loggers.
  virtual void add_filter(int new_filter);
  virtual void remove_filter(int old_filter);
  virtual void clear_filters();
  virtual void eol(line_ending to_set);

  // note: currently the byte formatting routines are not supported.

private:
  console_logger *_out;  //!< mirrored output.
};

////////////////////////////////////////////////////////////////////////////

//!< retasks the program wide logger to use a new combo_logger.
/*!< the original log_base object is returned and must be dealt with
appropriately. */
log_base OPSYSTEM_FUNCTION_STYLE *set_PW_logger_for_combo
    (const istring &file_name, int limit = file_logger::LOG_FILE_SIZE);

//!< a macro that retasks the program-wide logger as a combo_logger.
#ifndef OMIT_PROGRAM_WIDE_LOGGER
  #define SET_DEFAULT_COMBO_LOGGER { \
    log_base *old_log = set_PW_logger_for_combo(app_name_for_file_logger()); \
    WHACK(old_log); \
  }
#else
  #define SET_DEFAULT_COMBO_LOGGER
#endif

#endif

