/*****************************************************************************\
*                                                                             *
*  Name   : filename                                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

// implementation note: the filename is kept canonicalized.  any constructor
// or assignment operator should ensure this (except the blank constructor).

#include "byte_filer.h"
#include "filename.h"

#include <basis/byte_array.h>
#include <basis/common_templates.h>
#include <basis/function.h>
#include <basis/istring.h>
#include <basis/log_base.h>
#include <basis/portable.h>
#include <basis/utility.h>

#include <stdio.h>

#include <sys/stat.h>

using namespace basis;

#if defined(__WIN32__) || defined(__VMS__)
  const char DEFAULT_SEPARATOR = '\\';
#elif defined(__UNIX__)
  const char DEFAULT_SEPARATOR = '/';
#else
  #error "We have no idea what the default path separator is."
#endif

const char *NO_PARENT_DEFAULT = ".";
  // used when no directory name can be popped off.

filename::filename()
: _had_directory(false),
  _contents(new istring)
{}

filename::filename(const istring &name)
: _had_directory(true),
  _contents(new istring(name))
{ canonicalize(); }

filename::filename(const istring &directory, const istring &name_of_file)
: _had_directory(true),
  _contents(new istring(directory))
{
  // if the directory is empty, use the current directory.
  if (!directory) {
    *_contents = istring(NO_PARENT_DEFAULT);
    _had_directory = false;
  }
  // check for a slash on the end of the directory.  add one if there is none
  // currently.
  bool add_slash = false;
  if ( (directory[directory.end()] != '\\')
       && (directory[directory.end()] != '/') ) add_slash = true;
  if (add_slash) *_contents += DEFAULT_SEPARATOR;
  *_contents += name_of_file;
  canonicalize();
}

filename::filename(const filename &to_copy)
: _had_directory(to_copy._had_directory),
  _contents(new istring(*to_copy._contents))
{ canonicalize(); }

filename::~filename() { WHACK(_contents); }

istring filename::default_separator() { return istring(DEFAULT_SEPARATOR, 1); }

filename::operator const istring &() const { return *_contents; }

istring &filename::raw() { return *_contents; }

const istring &filename::raw() const { return *_contents; }

bool filename::good() const { return exists(); }

bool filename::unlink() const { return ::unlink(_contents->s()) == 0; }

bool filename::separator(char is_it)
{ return (is_it == pc_separator) || (is_it == unix_separator); }

filename &filename::operator = (const filename &to_copy)
{
  if (this == &to_copy) return *this;
  *_contents = *to_copy._contents;
  _had_directory = to_copy._had_directory;
  return *this;
}

filename &filename::operator = (const istring &to_copy)
{
  _had_directory = true;
  *_contents = to_copy;
  canonicalize();
  return *this;
}

bool filename::operator == (const filename &that) const
{ return *_contents == *that._contents; }

istring filename::pop()
{
  istring to_return = basename();
  filename parent_dir = parent();
  if (parent_dir.raw() == NO_PARENT_DEFAULT) {
    // we haven't gone anywhere.
    return "";  // signal that nothing was removed.
  }
  *this = parent_dir;
  return to_return;
}

filename filename::parent() const { return dirname(); }

void filename::push(const istring &to_push)
{
  *this = filename(*this, to_push);
}

void filename::canonicalize()
{
  // turn all the non-default separators into the default.
  bool found_sep = false;
  for (int j = 0; j < _contents->length(); j++) {
    if (separator(_contents->get(j))) {
      found_sep = true;
      _contents->put(j, DEFAULT_SEPARATOR);
#ifdef __WIN32__
    } else if ( (_contents->get(j) == '"') || (_contents->get(j) == '\'') ) {
      // why are we eating these?  it is probably because dos/windoze is
      // severely underpowered at the command line.  if a path has quotes
      // around it on the command line, it usually comes into the program that
      // way as well.  it actually makes sense to remove these, since they
      // are not really intended as part of the path...
      _contents->zap(j, j);
      j--;  // skip back for character we chowed.
#endif
    }
  }

  // if there wasn't a single directory separator, then they must not have
  // specified any directory name for this filename (although it could itself
  // be a directory).
  if (!found_sep) _had_directory = false;

  // remove all occurrences of double separators except for the first
  // double set, which could be a UNC filename.
  bool saw_sep = false;
  for (int i = 1; i < _contents->length(); i++) {
    if (separator(_contents->get(i))) {
      if (saw_sep) {
        _contents->zap(i, i);
          // two in a row is no good, except for the first two.
        i--;  // skip back one and try again.
        continue;
      }
      saw_sep = true;
    } else saw_sep = false;
  }

#ifdef __WIN32__
  // special processing for cygwin.  we turn the cygdrive format for drive
  // letters back into a normal X: so that the rest of the windows system
  // can handle it.  this is necessary when we've got things running under
  // bash and we're getting those wacky paths for $TEMP and such.
  //    path /cygdrive/c....
  //    posn 0123456789111
  //                   012
  if (separator(_contents->get(0)) && (_contents->length() >= 11) ) {
    // the first requirement was met; it starts with a slash and has enough
    // length for us to do this.
    const char *CYGDRIVE_PORTION = "cygdrive";
//hmmm: make this more efficient.
    if (_contents->substring(1, 8).iequals(CYGDRIVE_PORTION)
        && separator(_contents->get(9)) ) {
      // found the dreaded string in the path; the OS will not know this.
      _contents->zap(0, 9);  // whack cygdrive and first slash.
      _contents->insert(1, ":");
    }
  }
#endif

  // we don't crop the last separator if the name's too small.  for msdos
  // names, that would be chopping a slash off the c:\ style name.
  if (_contents->length() > 3) {
    // zap any separators that are hiding on the end.
    const int last = _contents->end();
    if (separator(_contents->get(last))) _contents->zap(last, last);
  } else if ( (_contents->length() == 2) 
      && (_contents->get(1) == ':') ) {
    // special case for dos drive names.  we turn it back into a valid
    // directory rather than leaving it as just "X:".  that form of the name
    // means something else under dos/windows.
    *_contents += istring(DEFAULT_SEPARATOR, 1);
  }
}

char filename::drive(bool interact_with_fs) const
{
  // first guess: if second letter's a colon, first letter's the drive.
  if (_contents->length() < 2) return '\0';
  if (_contents->get(1) == ':') return _contents->get(0);
  if (!interact_with_fs) return '\0';

  // otherwise, retrieve the file system's record for the file.
  status_info fill;
  if (!get_info(fill)) return '\0';
  return char('A' + fill.st_dev);
}

istring filename::extension() const
{
  istring base(basename().raw());
  int posn = base.find('.', base.end(), true);
  if (negative(posn)) return "";
  return base.substring(posn + 1, base.length() - 1);
}

istring filename::rootname() const
{
  istring base(basename().raw());
  int posn = base.find('.', base.end(), true);
  if (negative(posn)) return base;
  return base.substring(0, posn - 1);
}

bool filename::get_info(status_info &to_fill) const
{
  int ret = stat(_contents->observe(), &to_fill);
  if (ret) return false;
  return true;
}

bool filename::is_directory() const
{
  status_info fill;
  if (!get_info(fill)) return false;
  return !!(fill.st_mode & S_IFDIR);
}

bool filename::is_writable() const
{
  status_info fill;
  if (!get_info(fill)) return false;
  return !!(fill.st_mode & S_IWRITE);
}

bool filename::is_readable() const
{
  status_info fill;
  if (!get_info(fill)) return false;
  return !!(fill.st_mode & S_IREAD);
}

bool filename::is_executable() const
{
  status_info fill;
  if (!get_info(fill)) return false;
  return !!(fill.st_mode & S_IEXEC);
}

int filename::find_last_separator(const istring &look_at) const
{
  int last_sep = -1;
  int sep = 0;
  while (sep >= 0) {
    sep = look_at.find(DEFAULT_SEPARATOR, last_sep + 1);
    if (sep >= 0) last_sep = sep;
  }
  return last_sep;
}

filename filename::basename() const
{
  istring basename = *_contents;
  int last_sep = find_last_separator(basename);
  if (last_sep >= 0) basename.zap(0, last_sep);
  return basename;
}

filename filename::dirname() const
{
  istring dirname = *_contents;
  int last_sep = find_last_separator(dirname);
  // we don't accept ripping off the first slash.
  if (last_sep >= 1) {
    // we can rip the slash and suffix off to get the directory name.  however,
    // this might be in the form X: on windows.  if they want the slash to
    // remain, they can use the dirname that appends it.
    dirname.zap(last_sep, dirname.end());
  } else {
    if (_contents->get(0) == DEFAULT_SEPARATOR) {
      // handle when we're up at the top of the filesystem.  on unix, once
      // you hit the root, you can keep going up but you still remain at
      // the root.  similarly on windoze, if there's no drive name in there.
      dirname = istring(DEFAULT_SEPARATOR, 1);
    } else {
      // there's no slash at all in the filename any more.  we assume that
      // the directory is the current one, if no other information is
      // available.  this default is already used by some code.
      dirname = NO_PARENT_DEFAULT;
    }
  }
  return dirname;
}

istring filename::dirname(bool add_slash) const
{
  istring tempname = dirname().raw();
  if (add_slash) tempname += DEFAULT_SEPARATOR;
  return tempname;
}

bool filename::exists() const
{
  if (is_directory()) return true;
  if (!_contents->length()) return false;
  byte_filer opened(_contents->observe(), "r");
  return opened.good();
}

bool filename::legal_character(char to_check)
{
  switch (to_check) {
    case ':': case ';':
    case '\\': case '/':
    case '*': case '?': case '$': case '&': case '|':
    case '\'': case '"': case '`':
    case '(': case ')':
    case '[': case ']':
    case '<': case '>':
    case '{': case '}':
      return false;
    default: return true;
  }
}

void filename::detooth_filename(istring &to_clean, char replacement)
{
  for (int i = 0; i < to_clean.length(); i++) {
    if (!legal_character(to_clean[i]))
      to_clean[i] = replacement;
  }
}

void filename::pack(byte_array &packed_form) const
{
  attach(packed_form, int(_had_directory));
  _contents->pack(packed_form);
}

bool filename::unpack(byte_array &packed_form)
{
  int temp;
  if (!detach(packed_form, temp)) return false;
  _had_directory = temp;
  if (!_contents->unpack(packed_form)) return false;
  return true;
}

void filename::separate(string_array &pieces) const
{
  pieces.reset();
  const istring &raw_form = raw();
  istring accumulator;  // holds the names we find.
  for (int i = 0; i < raw_form.length(); i++) {
    if (separator(raw_form[i])) {
      // this is a separator character, so eat it and add the accumulated
      // string to the list.
      if (!i || accumulator.length()) pieces += accumulator;
      // now reset our accumulated text.
      accumulator = istring::empty_string();
    } else {
      // not a separator, so just accumulate it.
      accumulator += raw_form[i];
    }
  }
  if (accumulator.length()) pieces += accumulator;
}

void filename::join(const string_array &pieces)
{
  istring constructed_name;  // we'll make a filename here.
  for (int i = 0; i < pieces.length(); i++) {
    constructed_name += pieces[i];
    if (!i || (i != pieces.length() - 1))
      constructed_name += DEFAULT_SEPARATOR;
  }
  *this = constructed_name;
}

bool filename::base_compare_prefix(const filename &to_compare,
    string_array &first, string_array &second)
{
  separate(first);
  to_compare.separate(second);
  if (first.length() > second.length()) return false;
    // that case should never be allowed, since there are some bits missing
    // in the name to be compared.

  // compare each of the pieces.
  for (int i = 0; i < first.length(); i++) {
#if defined(__WIN32__) || defined(__VMS__)
    // case-insensitive compare.
    if (!first[i].iequals(second[i])) return false;
#else
    // case-sensitive compare.
    if (first[i] != second[i]) return false;
#endif
  }
  return true;
}

bool filename::compare_prefix(const filename &to_compare, istring &sequel)
{
  sequel = istring::empty_string();  // clean our output parameter.
  string_array first;
  string_array second;
  if (!base_compare_prefix(to_compare, first, second)) return false;

  // create the sequel string.
  int extra_strings = second.length() - first.length();
  for (int i = second.length() - extra_strings; i < second.length(); i++) {
    sequel += second[i];
    if (i != second.length() - 1) sequel += DEFAULT_SEPARATOR;
  }

  return true;
}

bool filename::compare_prefix(const filename &to_compare)
{
  string_array first;
  string_array second;
  return base_compare_prefix(to_compare, first, second);
}

bool filename::base_compare_suffix(const filename &to_compare,
    string_array &first, string_array &second)
{
  separate(first);
  to_compare.separate(second);
  if (first.length() > second.length()) return false;
    // that case should never be allowed, since there are some bits missing
    // in the name to be compared.

  // compare each of the pieces.
  for (int i = first.length() - 1; i >= 0; i--) {
//clean up this computation; the difference in lengths is constant--use that.
    int distance_from_end = first.length() - 1 - i;
    int j = second.length() - 1 - distance_from_end;
#if defined(__WIN32__) || defined(__VMS__)
    // case-insensitive compare.
    if (!first[i].iequals(second[j])) return false;
#else
    // case-sensitive compare.
    if (first[i] != second[j]) return false;
#endif
  }
  return true;
}

bool filename::compare_suffix(const filename &to_compare, istring &prequel)
{
  prequel = istring::empty_string();  // clean our output parameter.
  string_array first;
  string_array second;
  if (!base_compare_suffix(to_compare, first, second)) return false;

  // create the prequel string.
  int extra_strings = second.length() - first.length();
  for (int i = 0; i < extra_strings; i++) {
    prequel += second[i];
    if (i != second.length() - 1) prequel += DEFAULT_SEPARATOR;
  }
  return true;
}

bool filename::compare_suffix(const filename &to_compare)
{
  string_array first;
  string_array second;
  return base_compare_suffix(to_compare, first, second);
}

////////////////////////////////////////////////////////////////////////////

file_info::file_info(const filename &to_copy, int file_size, int checksum)
: filename(to_copy),
  _file_size(file_size),
  _checksum(checksum),
  _secondary(new istring),
  _attachment(new byte_array)
{}

file_info::file_info(const file_info &to_copy)
: filename(to_copy),
  _file_size(to_copy._file_size),
  _checksum(to_copy._checksum),
  _secondary(new istring(*to_copy._secondary)),
  _attachment(new byte_array(*to_copy._attachment))
{
}

file_info::~file_info()
{
  WHACK(_secondary);
  WHACK(_attachment);
}

const byte_array &file_info::attachment() const { return *_attachment; }

void file_info::attachment(const byte_array &new_attachment)
{ *_attachment = new_attachment; }

const istring &file_info::secondary() const { return *_secondary; }

void file_info::secondary(const istring &new_sec) const
{ *_secondary = new_sec; }

istring file_info::text_form() const
{
  istring to_return = raw()
      + isprintf(", size=%d, chksum=%d", _file_size, _checksum);
  if (_secondary->t()) to_return += istring(", 2ndary=") + *_secondary;
  return to_return;
}

bool file_info::calculate(const istring &prefix, bool just_size,
    int chunking_factor)
{
  filename full;
  if (prefix.t()) full = prefix + "/" + *this;
  else full = *this;
  if (!full.exists()) return false;
  // open the file for reading.
  byte_filer to_read(full.raw().s(), "rb");
  if (!to_read.good()) return false;  // why did that happen?
  // set the size appropriately.
  _file_size = to_read.length();
  if (just_size) return true;  // done for that case.

  // now read the file and compute a checksum.
  uint16 curr_sum = 0;  // the current checksum being computed.
  byte_array chunk;  // temporary chunk of data from file.

//hmmm: make this optimization (hack) optional!

  // this algorithm takes a chunk on each end of the file for checksums.
  // this saves us from reading a huge amount of data, although it will be
  // fooled if a huge binary file is changed only in the middle and has the
  // same size as before.  for most purposes, this is not a problem, although
  // databases that are fixed size might fool us.  if records are written in
  // the middle without updating the head or tail sections, then we're hosed.

  bool skip_tail = false;  // true if we don't need the tail piece.
  int head_start = 0, head_end = 0, tail_start = 0, tail_end = _file_size - 1;
////  int chunking_factor = FILENAME_CHUNKING_FACTOR;
////    // arbitrary decision that we read two chunks instead of one; this adds
////    // up to a total of one megabyte maximum checksum for all files.
  if (_file_size <= 2 * chunking_factor) {
    // we're applying a rule for when the file is too small compared to
    // the chunk factor doubled; we'll just read the whole file.
    head_end = _file_size - 1;
    skip_tail = true;
  } else {
    // here we compute the ending of the head piece and the beginning of
    // the tail piece.  each will be about chunk_factor in size.
    head_end = minimum(_file_size / 2, chunking_factor) - 1;
    tail_start = _file_size - minimum(_file_size / 2, chunking_factor);
  }

  // read the head end of the file.
  to_read.read(chunk, head_end - head_start + 1);
  curr_sum = utility::rolling_fletcher_checksum(curr_sum, chunk.observe(),
      chunk.length());

  // read the tail end of the file.
  if (!skip_tail) {
    to_read.seek(tail_start, byte_filer::FROM_START);
    
    to_read.read(chunk, tail_end - tail_start + 1);
    curr_sum = utility::rolling_fletcher_checksum(curr_sum, chunk.observe(),
        chunk.length());
  }

  _checksum = curr_sum;
  return true;
}

void file_info::pack(byte_array &packed_form) const
{
  filename::pack(packed_form);
  attach(packed_form, _file_size);
  attach(packed_form, _checksum);
  _secondary->pack(packed_form);
  attach(packed_form, *_attachment);
}

bool file_info::unpack(byte_array &packed_form)
{
  if (!filename::unpack(packed_form)) return false;
  if (!detach(packed_form, _file_size)) return false;
  if (!detach(packed_form, _checksum)) return false;
  if (!_secondary->unpack(packed_form)) return false;
  if (!detach(packed_form, *_attachment)) return false;
  return true;
}

file_info &file_info::operator = (const file_info &to_copy)
{
  if (this == &to_copy) return *this;
  (filename &)(*this) = (filename &)to_copy;
  *_attachment = *to_copy._attachment;
  _file_size = to_copy._file_size;
  *_secondary = *to_copy._secondary;
  _checksum = to_copy._checksum;
  return *this;
}

