#ifndef FILENAME_CLASS
#define FILENAME_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : filename & file_info                                              *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Provides operations commonly needed on file names.

#include "opsysdll.h"

#include <basis/packable.h>

// forward.
class string_array;

class OPSYSTEM_CLASS_STYLE filename : public packable
{
public:
  filename();  //!< blank constructor.
  filename(const istring &name);
    //!< creates a filename from any part of a full pathname, if possible.
    /*!< if the name contains quotes, they are stripped out. */
  filename(const istring &directory, const istring &name_of_file);
    //!< constructs a filename from a "directory" and the "name_of_file".
    /*!< the "name_of_file" can itself be a directory. */
  filename(const filename &to_copy);  //!< copy constructor.

  virtual ~filename();

  bool good() const;
    //!< returns true if the filename seems to be valid.
    /*!< this means that not only was the pathname parsed and found valid,
    but the file actually exists. */

  operator const istring &() const;
    //!< observes the istring that we're holding onto for the path.
  const istring &raw() const;
    //!< returns the istring that we're holding onto for the path.
  istring &raw();
    //!< accesses the istring that we're holding onto for the path.
    /*!< important note: if you change the string with this non-const raw()
    method, you MUST call canonicalize() on it again afterwards. */

  filename &operator = (const filename &to_copy);
    //!< provides assignment for this object, plus a simple string.
  filename &operator = (const istring &to_copy);
    //!< provides assignment for this object, plus a simple string.
    /*!< the latter version invokes canonicalize to clean the string up. */

  // standard comparisons.
  bool operator == (const filename &that) const;
  inline bool operator != (const filename &that) const
          { return !(*this == that); }

  void canonicalize();
    //!< cleans up the filename as needed for the current operating system.
    /*!< reforms the name by replacing any alternate directory separators with
    the operating system's preferred character. */

  bool exists() const;
    //!< returns true if the file exists.

  bool unlink() const;
    //!< actually removes the file, if possible.
    /*!< if the file was successfully deleted, then true is returned. */

  filename parent() const;
    //!< returns the parent filename for this one.

  istring pop();
    //!< removes the deepest component of the pathname.
    /*!< the component might be a file or directory name, but popping beyond
    the top-level directory will not succeed.  the returned string contains
    the component that was removed.  it will be a blank string if nothing
    could be popped. */

  void push(const istring &to_push);
    //!< pushes a new filename onto the current pathname.
    /*!< this only makes sense as a real pathname if this is currently a
    directory name and the component "to_push" is a child of that directory
    (or one intends to create that component as a child).  this is the
    opposite of pop. */

  filename basename() const;
    //!< returns the base of the filename; no directory.
  filename dirname() const;
    //!< returns the directory for the filename.
    /*!< if no directory name can be found in the filename, then "." is
    returned. */
  istring dirname(bool add_slash) const;
    //!< returns the directory for the filename and optionally adds a slash.
    /*!< if "add_slash" is true, then the default directory separator will be
    present on the end of the string. */
  bool had_directory() const { return _had_directory; }
    //!< returns true if the name that we were given had a non-empty directory.
    /*!< this allows one to distinguish between a file with the current
    directory (.) attached and a file with no directory specified. */

  char drive(bool interact_with_fs = false) const;
    //!< returns the drive letter for the file, without the colon.
    /*!< this only makes sense for a fully qualified MS-DOS style name.  if no
    drive letter is found, then '\0' is returned.  if "interact_with_fs" is
    true, then the file system will be checked for the actual drive if no
    drive letter was found in the contents. */

  istring extension() const;
    //!< returns the extension for the file, if one is present.

  istring rootname() const;
    //!< returns the root part of the basename without an extension.

  // status functions return true if the characteristic embodied in
  // the name is also true.

  bool is_directory() const;
  bool is_writable() const;
  bool is_readable() const;
  bool is_executable() const;

  //! the default separator for directories per operating system.
  /*! the PC uses the backward slash to separate file and directory names from
  each other, while Unix uses the forward slash. */
  enum directory_separator { pc_separator = '\\', unix_separator = '/' };

  static bool separator(char is_it);
    //!< returns true if the character "is_it" in question is a separator.

  static istring default_separator();
    //!< returns the default separator character for this OS.

  static bool legal_character(char to_check);
    //!< returns true if "to_check" is a valid character in a filename.
    /*!< this does not consider separator characters; it only looks at the
    the name components.  also, it is appropriate for the union of the
    operating systems we support. */

  static void detooth_filename(istring &to_clean, char replacement = '_');
    //!< takes any known illegal file system characters out of "to_clean".
    /*!< this prepares "to_clean" for use as a component in a larger filename
    by ensuring that the file system will not reject the name (as long as a
    suitable directory path is prepended to the name and permissions allow
    the file to be created or accessed).  the "replacement" is used as the
    character that is substituted instead of illegal characters. */

  void separate(string_array &pieces) const;
    //!< breaks the filename into its component directories.
    /*!< this returns an array containing the component names.  the last
    component, unless the filename held is actually a directory, should be the
    name of the file.  if the first character is a directory, then the first
    component will be empty. */

  void join(const string_array &pieces);
    //!< undoes a separate() operation to get the filename back.
    /*!< "this" is set to a filename made from each of the "pieces".  if there
    are any directory separators inside the pieces, then they will be removed
    by canonicalize(). */

  // these implement the packing functionality.
  virtual void pack(byte_array &packed_form) const;
  virtual bool unpack(byte_array &packed_form);

  bool compare_prefix(const filename &to_compare, istring &sequel);
    //!< compares "this" filename against "to_compare".
    /*!< this returns true if all of "this" is the same as the first portion
    of "to_compare".  that is, if "this" is a prefix of "to_compare", then
    true is returned.  this will always fail if there are fewer components in
    "to_compare".  it will always succeed if the two filenames are identical.
    on success, the "sequel" is set to the portion of "to_compare" that's
    not included in this filename. */

  bool compare_prefix(const filename &to_compare);
    //!< this simpler form doesn't bother with computing the sequel.

  bool compare_suffix(const filename &to_compare, istring &prequel);
    //!< compares the back end of a filename to this.
    /*!< this is similar to compare_prefix() but it checks to see if the
    back end of "this" filename is the same as "to_compare".  if "this" is
    longer than "to_compare", then failure occurs.  only if all of the bits
    in "this" are seen in the back of "to_compare" is true returned. */

  bool compare_suffix(const filename &to_compare);

private:
  bool _had_directory;  //!< true if _some_ directory was specified on init.
  istring *_contents;  //!< the full path is held here.

  int find_last_separator(const istring &look_at) const;
    //!< locates the last separator character in the filename.

  typedef struct stat status_info;
  bool get_info(status_info &to_fill) const;
    //!< returns information for the filename.

  // helper functions do the real work for comparing.
  bool base_compare_prefix(const filename &to_compare, string_array &first,
          string_array &second);
  bool base_compare_suffix(const filename &to_compare, string_array &first,
          string_array &second);
};

////////////////////////////////////////////////////////////////////////////

class OPSYSTEM_CLASS_STYLE file_info : public filename 
{
public:
  int _file_size;  //!< the size of the file.
  int _checksum;  //!< the checksum for the file.

  file_info(const filename &to_copy = filename(), int file_size = 0,
          int checksum = 0);
    //!< to get the real file size and checksum, invoke the calculate method.

  file_info(const file_info &to_copy);

  virtual ~file_info();

  file_info &operator = (const file_info &to_copy);

  istring text_form() const;

  bool calculate(const istring &prefix, bool just_size = false,
        int chunk_factor = 128 * KILOBYTE);
    //!< fills in the correct file size and checksum information for this file.
    /*!< note that the file must exist for this to work.  if "just_size"
    is true, then the checksum is not calculated.  if the "prefix" is not
    empty, then it is used as a directory path that needs to be added to
    the filename to make it completely valid.  this is common if the
    filename is stored relative to a path.  the "chunk_factor" is used for
    checksum calculations; only 2 * chunk_factor bytes will be factored in,
    half on each end of the file. */

  const istring &secondary() const;
    //!< observes the alternate form of the name.
  void secondary(const istring &new_sec) const;
    //!< accesses the alternate form of the name.

  const byte_array &attachment() const;
    //!< returns the chunk of data optionally attached to the file's info.
    /*!< this supports extending the file's record with extra data that might
    be needed during processing. */
  void attachment(const byte_array &new_attachment);
    //!< sets the optional chunk of data hooked up to the file's info.

  // standard streaming operations.
  virtual void pack(byte_array &packed_form) const;
  virtual bool unpack(byte_array &packed_form);

private:
  istring *_secondary;  //!< alternate filename for the main one.
  byte_array *_attachment;  //!< extra information, if needed.
};

#endif

