#ifndef FILENAME_LIST_CLASS
#define FILENAME_LIST_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : filename_list                                                     *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2005-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Implements a list of filenames.
/*!
  This is based on an amorph so that adding to the list is efficient.
  The underlying type held is actually a file_info rather than a filename.
  This should not impose much extra overhead.
  Note: this is a heavyweight header; it shouldn't be used in other headers.
*/

#include "filename.h"

#include <basis/common_templates.h>
#include <basis/log_base.h>
#include <basis/packable.h>
#include <data_struct/amorph.cpp>

AMORPH_PACKING;  // get the two methods for packing and unpacking amorphs.

//hmmm: submerge the implementation into the cpp file.

class OPSYSTEM_CLASS_STYLE filename_list
: public amorph<file_info>, public packable
{
public:
  filename_list() : amorph<file_info>() {}

  filename_list &operator =(const filename_list &to_copy) {
    if (this == &to_copy) return *this;
    reset();
    for (int i = 0; i < to_copy.elements(); i++) {
      append(new file_info(*to_copy.get(i)));
    }
    return *this;
  }

  int total_files() const { return elements(); }

  int total_size() const {
    int to_return = 0;
    for (int i = 0; i < elements(); i++)
      to_return += get(i)->_file_size;
    return to_return;
  }

  bool calculate_progress(const filename &file, int current_offset,
      int &current_file, int &current_size) {
    current_file = 0;
    current_size = 0;
    int posn = locate(file);
    if (negative(posn)) {
      if (file.raw().t()) return false;  // not a member.
      // they're at the start of the run.
      current_file = 1;
      return true;
    }
    current_file = posn + 1;  // position zero in array means file number 1.
    int size_finished = 0;
    // iterate on all files before the current one.
    for (int i = 0; i < posn; i++) {
      size_finished += get(i)->_file_size;
    }
    current_size = size_finished + current_offset;
    return true;
  }

  filename_list &operator = (const string_array &to_copy) {
    reset();
    for (int i = 0; i < to_copy.length(); i++) {
      append(new file_info(to_copy[i]));
    }
    return *this;
  }

  void fill(string_array &to_fill) {
    to_fill.reset();
    for (int i = 0; i < elements(); i++) {
      to_fill += get(i)->raw();
    }
  }

  const file_info *find(const filename &to_check) const {
    for (int i = 0; i < elements(); i++) {
#if defined(__WIN32__) || defined(__VMS__)
      if (to_check.raw().iequals(get(i)->raw())) return get(i);
#else
      if (to_check.raw() == get(i)->raw()) return get(i);
#endif
    }
    return NIL;
  }

  //! finds the index of the record for "to_find" or returns a negative number.
  int locate(const filename &to_find) const {
    for (int i = 0; i < elements(); i++) {
#if defined(__WIN32__) || defined(__VMS__)
      if (to_find.raw().iequals(get(i)->raw())) return i;
#else
      if (to_find.raw() == get(i)->raw()) return i;
#endif
    }
    return common::NOT_FOUND;
  }

  bool member(const filename &to_check) const {
    for (int i = 0; i < elements(); i++) {
#if defined(__WIN32__) || defined(__VMS__)
      if (to_check.raw().iequals(get(i)->raw())) return true;
#else
      if (to_check.raw() == get(i)->raw()) return true;
#endif
    }
    return false;
  }

  bool member(const file_info &to_check, bool check_size = true,
          bool check_checksum = true) const {
    for (int i = 0; i < elements(); i++) {
#if defined(__WIN32__) || defined(__VMS__)
      if (to_check.raw().iequals(get(i)->raw())) {
#else
      if (to_check.raw() == get(i)->raw()) {
#endif
        // once we have matched a name, the other checks will cause us to
        // reject any other potential matches after this one if the requested
        // check fails.
        if (check_size && (to_check._file_size != get(i)->_file_size) )
          return false;
        if (check_checksum && (to_check._checksum != get(i)->_checksum) )
          return false;
        return true;
      }
    }
    return false;
  }

  istring text_form() const {
    istring to_return;
//hmmm: a length limit might be nice?
    for (int i = 0; i < elements(); i++) {
      to_return += isprintf("%d. ", i + 1);
      if (get(i))
        to_return += get(i)->text_form();
      if (i != elements() - 1)
        to_return += log_base::platform_ending();
    }
    return to_return;
  }

  virtual void pack(byte_array &packed_form) const {
    amorph_pack(packed_form, *this);
  }

  virtual bool unpack(byte_array &packed_form) {
    return amorph_unpack(packed_form, *this);
  }
};

#endif

