/*****************************************************************************\
*                                                                             *
*  Name   : ini_configurator                                                  *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#ifdef __UNIX__
  #include "byte_filer.h"
#endif
#include "command_line.h"
#include "filename.h"
#include "ini_config.h"
#include "path_configuration.h"

#include <basis/common_templates.h>
#include <basis/convert_utf.h>
#include <basis/istring.h>
#include <basis/portable.h>
#include <basis/utility.h>
#include <data_struct/common_templates.h>
#include <data_struct/static_memory_gremlin.h>
#include <data_struct/symbol_table.h>
#include <textual/tokenizer.h>

#undef LOG
#define LOG(s) CLASS_EMERGENCY_LOG(program_wide_logger(), s)

//#define DEBUG_INI_CONFIGURATOR
  // uncomment for noisy version.

const int MAXIMUM_LINE_SIZE = 16384;

// a default we hope never to see in an ini file.
SAFE_STATIC_CONST(istring_object, ini_configurator::ini_str_fake_default,
    ("NoTomatoesNorPotatoesNorQuayle"));

ini_configurator::ini_configurator(const istring &ini_filename,
      treatment_of_defaults behavior, file_location_default where)
: configurator(behavior),
  _ini_name(new filename),
#ifdef __UNIX__
  _parser(new ini_parser("", behavior)),
#endif
  _where(where)
{
  name(ini_filename);  // set name properly.
}

ini_configurator::~ini_configurator()
{
  WHACK(_ini_name);
#ifdef __UNIX__
  WHACK(_parser);
#endif
}

istring ini_configurator::name() const { return _ini_name->raw(); }

void ini_configurator::name(const istring &name)
{
  *_ini_name = name;

  bool use_appdir = false;
    // true if we should put files where programs start for those filenames
    // that don't include a directory name.
  if (_where != OS_DIRECTORY) use_appdir = true;
#ifndef __WIN32__
  use_appdir = true;
#endif
  // we must seek the file, since they specified no directory at all.
  if (use_appdir && !_ini_name->had_directory()) {
    // first try without environment.  this is needed in case there is an ini 
    // right with the file; our standard is to check there first.
    bool consult_env = false;
    *_ini_name = filename(path_configuration::application_directory
        (consult_env), _ini_name->basename());
    if (!_ini_name->exists()) {
      // not in the application dir, so check in the environment if allowed.
      if (_where == ENV_APPLICATION_DIR) consult_env = true;
      // we find the program directory, including checking environment, if
      // the user specified that.  no point in bothering to try again unless
      // they do like to consult the environment variable.
      if (consult_env)
        *_ini_name = filename(path_configuration::application_directory
            (consult_env), _ini_name->basename());
      // this is the ini file of last resort; if it doesn't exist, then
      // we're hosed.
    }
  }
#ifdef __UNIX__
  // read in the file's contents.
  read_ini_file();
#endif
}

bool ini_configurator::section_exists(const istring &section)
{
#ifdef __WIN32__
  string_table infos;
  // heavy-weight call here...
  return get_section(section, infos);
#else
  return _parser->section_exists(section);
#endif
}

#ifdef __UNIX__
void ini_configurator::read_ini_file()
{
  #define func "read_ini_file"
  _parser->reset("");  // clear out our current contents.
  byte_filer ini_file;
  bool open_ret = ini_file.open(*_ini_name, "r+");
    // open the file for binary read/write while keeping previous contents.
#ifdef DEBUG_INI_CONFIGURATOR
  if (!open_ret) LOG(istring("failed to open ini file: ") + *_ini_name);
  if (!ini_file.good()) LOG(istring("ini file not good: ") + *_ini_name);
#endif
  if (!open_ret || !ini_file.good()) {
    return;  // failure.
  }
  int file_size = ini_file.length();  // get the file length.
  // read the file.
  istring contents(' ', file_size + 3);
  int bytes_read = ini_file.read((byte *)contents.observe(), file_size);
  contents.zap(bytes_read + 1, contents.end());
  _parser->reset(contents);
  #undef func
}

void ini_configurator::write_ini_file()
{
  #define func "write_ini_file"
//eventually: just set dirty flag, schedule thread to write.

  // open filer with new mode for cleaning.
  byte_filer ini_file;
  bool open_ret = ini_file.open(*_ini_name, "w");
    // open the file for binary read/write and drop previous contents.
#ifdef DEBUG_INI_CONFIGURATOR
  if (!open_ret) LOG(istring("failed to open ini file: ") + *_ini_name);
  if (!ini_file.good()) LOG(istring("ini file not good: ") + *_ini_name);
#endif
  if (!open_ret || !ini_file.good()) return;  // failure.

  // output table's contents to text.
  istring text;
  _parser->restate(text);
  ini_file.write((byte *)text.observe(), text.length());
  #undef func
}
#endif //UNIX

bool ini_configurator::delete_section(const istring &section)
{
#ifdef __WIN32__
  return put_profile_string(section, "", ""); 
#else
  // zap the section.
  bool to_return = _parser->delete_section(section);
  // schedule the file to write.
  write_ini_file();
  return to_return;
#endif
}

bool ini_configurator::delete_entry(const istring &section, const istring &ent)
{
#ifdef __WIN32__
  return put_profile_string(section, ent, "");
#else
  // zap the entry.
  bool to_return = _parser->delete_entry(section, ent);
  // schedule the file to write.
  write_ini_file();
  return to_return;
#endif
}

bool ini_configurator::put(const istring &section, const istring &entry,
    const istring &to_store)
{
  #define func "put"
  if (!to_store.length()) return delete_entry(section, entry);
  else if (!entry.length()) return delete_section(section);
  else if (!section.length()) return false;
#ifdef __WIN32__
  return put_profile_string(section, entry, to_store);
#else
  // write the entry.
  bool to_return = _parser->put(section, entry, to_store);
  // schedule file write.
  write_ini_file();
  return to_return;
#endif
  #undef func
}

bool ini_configurator::get(const istring &section, const istring &entry,
    istring &found)
{
#ifndef __WIN32__
  return _parser->get(section, entry, found);
#else
  flexichar temp_buffer[MAXIMUM_LINE_SIZE];
  temp_buffer[0] = 0;
  get_profile_string(section, entry, ini_str_fake_default(),
      temp_buffer, MAXIMUM_LINE_SIZE - 1);
  found = from_unicode_temp(temp_buffer);
  return !(ini_str_fake_default() == found);
#endif
}

bool ini_configurator::get_section(const istring &section, string_table &info)
{
  #define func "get_section"
#ifndef __WIN32__
  return _parser->get_section(section, info);
#else
  info.reset();
  const int buffer_size = 200000;

  flexichar low_buff[buffer_size + 3];
  int read_len = GetPrivateProfileSection(to_unicode_temp(section.observe()),
      low_buff, buffer_size - 1, to_unicode_temp(name()));
  if (!read_len) return false;  // assume the API means there was no section.

  low_buff[read_len] = '\1';  // signal beyond the end of the stuff.
  low_buff[read_len + 1] = '\0';  // make sure we're still zero terminated.

  bool last_was_nil = false;
  // this loop replaces all the embedded nils with separators to allow the
  // tokenizer to retrieve all the strings from the section.
  for (int i = 0; i < read_len; i++) {
    if (!low_buff[i] && last_was_nil) {
      // termination condition; we got two nils in a row.
      // this is just paranoia; the length should tell us.
      break;
    } else if (!low_buff[i]) {
      low_buff[i] = '\1';  // replace with a separator.
      last_was_nil = true;
    } else last_was_nil = false;  // reset the nil flag.
  }

  // now convert to a simple istring.
  istring buff = from_unicode_temp(low_buff);
  int length = buff.length();
  buff.shrink();
  tokenizer parser("\1", "=");
  parser.parse(buff);
  info = parser.table();
  return true;
#endif
  #undef func
}

bool ini_configurator::put_section(const istring &section,
    const string_table &info)
{
#ifdef __WIN32__
  tokenizer parser("\1", "=");
  parser.table() = info;
  istring flat = parser.text_form();
  flat += "\1\1";  // add terminating guard.
  int len = flat.length();
  for (int i = 0; i < len; i++) {
    if (flat[i] == '\1') {
      flat[i] = '\0';
      if (flat[i+1] == ' ') {
        // if the space character is next, shift it before the nil to avoid
        // keys with a preceding space.
        flat[i] = ' ';
        flat[i + 1] = '\0';
      }
    }
  }
  return WritePrivateProfileSection(to_unicode_temp(section),
      to_unicode_temp(flat), to_unicode_temp(name()));
#else
  // write the section.
  bool to_return = _parser->put_section(section, info);
  // schedule file write.
  write_ini_file();
  return to_return;
#endif
}

#ifdef __WIN32__
bool ini_configurator::put_profile_string(const istring &section,
    const istring &entry, const istring &to_store)
{
  return bool(WritePrivateProfileString(to_unicode_temp(section),
      entry.length() ? (flexichar *)to_unicode_temp(entry) : NIL,
      to_store.length() ? (flexichar *)to_unicode_temp(to_store) : NIL,
      to_unicode_temp(name())));
}

void ini_configurator::get_profile_string(const istring &section,
    const istring &entry, const istring &default_value,
    flexichar *return_buffer, int buffer_size)
{
  GetPrivateProfileString(section.length() ?
      (flexichar *)to_unicode_temp(section) : NIL,
      entry.length() ? (flexichar *)to_unicode_temp(entry) : NIL,
      to_unicode_temp(default_value),
      return_buffer, buffer_size, to_unicode_temp(name()));
}
#endif

