#ifndef INI_CONFIGURATOR_CLASS
#define INI_CONFIGURATOR_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : ini_configurator                                                  *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Supports the configurator interface on INI files, commonly used in MS-   *
*  windows for storing configuration parameters.                              *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/object_base.h>
#ifdef __WIN32__
  #include <data_struct/configurator.h>
#else
  #include "ini_parser.h"
#endif

// forward.
class byte_filer;
class filename;

class OPSYSTEM_CLASS_STYLE ini_configurator : public configurator
{
public:
  // the ini file being manipulated will be stored in either the same
  // directory that the program is being executed from (APPLICATION_DIRECTORY)
  // or in the directory where the operating system resides.  if the flag is
  // ENV_APPLICATION_DIR, then the environment variable HOOPLE_APPS is used
  // as the storage directory (see path_configuration.h).
  enum file_location_default {
    APPLICATION_DIRECTORY,
    OS_DIRECTORY,
    ENV_APPLICATION_DIR
  };

  ini_configurator(const istring &ini_filename, treatment_of_defaults behavior,
          file_location_default where = ENV_APPLICATION_DIR);
    // creates an ini_configurator that stores entries into the "ini_filename"
    // and that has the "behavior".  "where" dictates the file's location if
    // no path is specified as part of "ini_filename".

  virtual ~ini_configurator();

  IMPLEMENT_CLASS_NAME("ini_configurator");

  istring name() const;
  void name(const istring &name);
    // observes or modifies the name of the file used for ini entries.

  virtual bool get(const istring &section, const istring &entry,
          istring &found);
    // implements the configurator retrieval function.  this returns true if
    // the entry was present and stores it in "found".

  virtual bool section_exists(const istring &section);
    // returns true if the "section" was found in the file.
    // NOTE: for an INI file, this is quite a heavy-weight call.

  virtual bool put(const istring &section, const istring &entry,
          const istring &to_store);
    // implements the configurator storage function.

  virtual bool delete_section(const istring &section);
    // removes the entire "section" specified.

  virtual bool delete_entry(const istring &section, const istring &entry);
    // removes the entry specified by the "section" and "entry" name.

  virtual bool get_section(const istring &section, string_table &info);
    // reads the entire "section" into a table called "info".  on win95, this
    // will fail if the section's data exceeds 32K.

  virtual bool put_section(const istring &section, const string_table &info);
    // writes a table called "info" into the "section" in the INI file.  any
    // existing data for that section is wiped out.  on win95, this will fail
    // if the section's data exceeds 32K.

private:
  filename *_ini_name;  // the file we're manipulating.
#ifdef __UNIX__
  ini_parser *_parser;  // used for real storage and parsing.
#endif
  file_location_default _where;  // where to find and store the file.

#ifdef __WIN32__
  // these two functions make the job of talking to windows a little easier.
  bool put_profile_string(const istring &section, const istring &entry,
          const istring &to_store);
  void get_profile_string(const istring &section, const istring &entry,
          const istring &default_value, flexichar *return_buffer,
          int buffer_size);
#endif
#ifdef __UNIX__
  void read_ini_file();
    // reads the INI file's contents into memory.
  void write_ini_file();
    // store the current contents into the INI file.
#endif

  // not to be called.
  ini_configurator(const ini_configurator &);
  ini_configurator &operator =(const ini_configurator &);

  static const istring_object &ini_str_fake_default();
};

#endif

