#ifndef LOCKED_OBJECT_CLASS
#define LOCKED_OBJECT_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : locked_object                                                     *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Wraps a templated object safely with regard to threads.  Both accessing  *
*  and swapping-out of the object are thread-safe.  This means that the held  *
*  object can safely be used from as many threads as desired (if the object   *
*  itself is thread-safe) and the held object can be switched with another of *
*  the same type without halting the threads operating on it.                 *
*    This is a fairly heavy-weight header, so it would be better to forward   *
*  declare the locked_object template rather than including this in a header. *
*                                                                             *
*******************************************************************************
* Copyright (c) 2002-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#ifdef CATCH_ERRORS
  #include <basis/guards.h>
#endif
#include <basis/mutex.h>
#include <basis/utility.h>

template <class contents>
class locked_object
{
public:
  locked_object(contents *object_held = NIL) { _the_object = object_held; }
    // takes possession of the "object_held" and protects it from multiple
    // accesses by threads.  there is a certain amount of cooperation required;
    // when the object is locked for accessing, it must be unlocked again
    // afterwards.  the object can be locked multiple times only on the same
    // thread and it must be unlocked an equal number of times.

  // destroys the object held before closing up shop.  rescue it with the
  // swap method if that's a problem.
  ~locked_object() {
    contents *to_zap = swap(NIL);
    WHACK(to_zap);
  }

  contents *lock_object() { _locking.lock(); return _the_object; }
    // allows the object to be accessed safely.  the internal synchronization
    // primitive is locked before the object is returned.  if the object held
    // was passed as NIL, then this will return NIL; it is still necessary
    // to unlock this object again even though NIL is returned!

  void unlock_object(contents *to_unlock) {
    // unlocks this locked_object given the object "to_unlock" that was earlier
    // granted by lock_object().  that object pointer MUST match what was given
    // out earlier or grave consequences result.
#ifdef CATCH_ERRORS
    if (to_unlock != _the_object)
      deadly_error("locked_object", "unlock_object",
          "erroneous pointer has been passed in");
#endif
    if (to_unlock != _the_object) return;
    _locking.unlock();
  }

  contents *swap(contents *new_object) {
    // replaces the existing object with the "new_object".  the old object is
    // returned and MUST be deleted eventually.
    _locking.lock();
    contents *to_return = _the_object;
    _the_object = new_object;
    _locking.unlock();
    return to_return;
  }

private:
  contents *_the_object;
  mutex _locking;

  // not permitted.
  locked_object(const locked_object &);
  locked_object &operator =(const locked_object &);
};

#endif

