/*****************************************************************************\
*                                                                             *
*  Name   : path_configuration                                                *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "command_line.h"
#include "filename.h"
#include "ini_config.h"
#include "path_configuration.h"

#include <basis/function.h>
#include <basis/guards.h>
#include <basis/istring.h>
#include <basis/mutex.h>
#include <basis/object_base.h>
#include <basis/portable.h>
#include <basis/utility.h>
#include <data_struct/static_memory_gremlin.h>
#include <textual/parser_bits.h>

#ifdef __WIN32__
  #include <direct.h>
#else
  #include <dirent.h>
#endif
#include <string.h>
#include <sys/stat.h>

namespace path_configuration {

const char *PATH_CONFIGURATION_FILENAME() { return "paths.ini"; }

istring path_configuration_file(bool consult_environment)
{
  filename cfg_file(application_directory(false) + "/"
      + PATH_CONFIGURATION_FILENAME());
  if (cfg_file.exists()) return cfg_file.raw();
  cfg_file = filename(application_directory(true) + "/"
      + PATH_CONFIGURATION_FILENAME());
  // we just return this, since consulting the environment is our last
  // resort anyway.
  return cfg_file.raw();
}

const istring &GLOBAL_SECTION_NAME() { STATIC_STRING("Common"); }

const istring &LOCAL_FOLDER_NAME() { STATIC_STRING("LocalFiles"); }

const istring &LOGGING_FOLDER_NAME() { STATIC_STRING("LogPath"); }

////////////////////////////////////////////////////////////////////////////

const int MAX_LOG_PATH = 200;
  // the maximum length of the entry stored for the log path.

istring installation_root(bool consult_environment)
{
  // get our main paths config file.
  filename ini_name = path_configuration_file(consult_environment);

  // the default uses the directory where the paths file is located.
  ini_configurator ini(ini_name, ini_configurator::RETURN_ONLY);
  return ini.load(GLOBAL_SECTION_NAME(), LOCAL_FOLDER_NAME(),
      ini_name.dirname());
}

istring get_logging_directory()
{
  // get the real config file's name, but using the environment for the lookup.
  filename ini_name = path_configuration_file(true);

  // set up a name that will be the default if the entry doesn't exist already.
  ini_configurator ini(ini_name, ini_configurator::RETURN_ONLY);
  istring def_log = ini.load(GLOBAL_SECTION_NAME(), LOCAL_FOLDER_NAME(),
      ini_name.dirname());
    // default places the folder in the current directory.
  def_log += "/logs";
    // add the subdirectory for logs.

  // now grab the current value for the name, if any.
  istring log_dir = ini.load(GLOBAL_SECTION_NAME(), LOGGING_FOLDER_NAME(), "");
    // get the entry for the logging path.
  if (!log_dir) {
    // if the entry was absent, we set it.
    ini.store(GLOBAL_SECTION_NAME(), LOGGING_FOLDER_NAME(), def_log);
  } else {
    // they gave us something.  let's replace the environment variables
    // in their string so we resolve paths and such.
    log_dir = parser_bits::substitute_env_vars(log_dir);
  }

  // now we make sure the directory exists.
  struct stat to_fill;
  int stat_ret = stat(log_dir.observe(), &to_fill);
  if (stat_ret || !(to_fill.st_mode & S_IFDIR) ) {
    // if it's not anything yet or if it's not a directory, then we need
    // to create it.
//if it's something besides a directory... should it be deleted?
#ifdef __UNIX__
    int mk_ret = mkdir(log_dir.s(), 0777);
#endif
#ifdef __WIN32__
    int mk_ret = mkdir(log_dir.s());
#endif
    if (mk_ret) return "";
//can't have a log file if we can't make the directory successfully???
  }

  return log_dir;
}

istring make_logfile_name(const istring &base_name)
{ return get_logging_directory() + "/" + base_name; }

SAFE_STATIC(istring_object, _hidden_hardcoding_appname, );

void hard_code_application_name(const istring &new_name)
{ _hidden_hardcoding_appname() = new_name; }

istring application_name()
{
  if (_hidden_hardcoding_appname().t()) return _hidden_hardcoding_appname();
  return portable::application_name();
}

SAFE_STATIC(istring_object, _hidden_hardcoding_dir, );

void hard_code_application_dir(const istring &new_dir)
{ _hidden_hardcoding_dir() = new_dir; }

istring application_directory(bool consult_environment)
{
  // first check for the so-called hard-coded directory.
  if (_hidden_hardcoding_dir().t())
    return _hidden_hardcoding_dir();

  // look in environment variables if they want that.
  if (consult_environment) {
    istring envirodir = portable::env_string("HOOPLE_APPS");
      // query using library name first.
    if (!envirodir) envirodir = portable::env_string("INOVA_APPS");
      // if that didn't work, we try old standby.
    if (envirodir.t()) return envirodir;
      // if either yielded a directory, we return that instead of exe dir.
  }

  // last chance; get the directory that our application was started in.
  return filename(application_name()).dirname().raw();
}

istring read_item(const istring &key_name)
{
  filename ini_name = path_configuration_file(true);
  ini_configurator ini(ini_name, ini_configurator::RETURN_ONLY);
  return ini.load(GLOBAL_SECTION_NAME(), key_name, "");
}

} // namespace.

