#ifndef RUNTIME_HISTORY_CLASS
#define RUNTIME_HISTORY_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : runtime_history                                                   *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Keeps a record of the outcomes that have been recorded by objects.       *
*  A record here is generally a sign that these outcomes were suspicious,     *
*  unusual, extraordinary, or horrendously bad.  The history list is allowed  *
*  to hold a certain number of entries before the more ancient entries begin  *
*  dropping out of the list.                                                  *
*                                                                             *
*  Note: this class is thread-safe and can be used globally within a program. *
*                                                                             *
*******************************************************************************
* Copyright (c) 1990-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/packable.h>

const int DEFAULT_RUNTIME_HISTORY = 2000;

// forward.
class mutex;
class occurrence;
class occurrence_list;

class OPSYSTEM_CLASS_STYLE runtime_history : public packable
{
public:
  runtime_history(int maximum_length = DEFAULT_RUNTIME_HISTORY);
  ~runtime_history();

  void add_outcome(const occurrence &happening);
    // adds the occurrence in "happening" to the runtime history.  if the
    // list would exceed its maximum, then an entry is chopped.

  // note that elements() and get() are thread safe, but unless one locks
  // the list, one is not guaranteed that the list of occurrences will not
  // change.  if you do lock the list, you must unlock it again afterwards.

  int elements() const;
    // returns the number of occurrences that are in our list.

  const occurrence &get(int index);
    // returns the occurrences that resides at index in the list.

  void lock();  // fixate the list for scanning.  must unlock when done.
  void unlock();  // allow others to use the list again.

  virtual void pack(byte_array &packed_form) const;
  virtual bool unpack(byte_array &packed_form);
    // supports streaming the history out.

  // observes or modifies the maximum number of occurrences that we'll hold.
  int max_items() const;
  void set_max_items(int new_max);

private:
  int _bound;  // maximum number of occurrences.
  mutex *_lock;  // protects the list.
  occurrence_list *_occurrences;  // the list of occurrences.

  const occurrence &empty_junk_occurrence();  // a blank occurrence.
};

#endif

