#ifndef SYSTEM_VALUES_CLASS
#define SYSTEM_VALUES_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : system_values                                                     *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    This class provides a way to look up values for the code base, including *
*  outcomes and filters.  These items are reported in the build manifest.     *
*  That file can be read as an INI file to provide the information that is    *
*  available through this class.                                              *
*                                                                             *
*******************************************************************************
* Copyright (c) 2004-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/object_base.h>
#include <data_struct/static_memory_gremlin.h>

// forward.
class system_values_lookup_list;

class OPSYSTEM_CLASS_STYLE system_values : public virtual object_base
{
public:
  system_values(const istring &section_tag);
    // the "manifest_file" must be the path to a file that was created by
    // the value_tagger as part of indexing all value assets defined in the
    // build.  the "section_tag" indicates what kind of asset is being
    // looked up.  the valid values for sections so far are "DEFINE_OUTCOME",
    // "DEFINE_FILTER" and "DEFINE_EVENT", but it is better to use the defined
    // "VALUES" methods below (such as OUTCOME_VALUES()).  outcomes are used by
    // functions to describe how the operation completed.  filters are used to
    // enable different types of logging.  events are sent between information
    // source and sink to indicate the occurrence of an activity.  
  
  virtual ~system_values();

  IMPLEMENT_CLASS_NAME("system_values");

  // these provide symbolic versions of the section tag used in the
  // constructor.  these are preferable to using the string constants
  // directly.
  static const char *OUTCOME_VALUES();
  static const char *FILTER_VALUES();
  static const char *EVENT_VALUES();

  bool use_other_manifest(const istring &manifest_file);
    // if the default manifest is not appropriate, the values can come from
    // the "manifest_file" instead.

  virtual istring text_form() const;
    // shows all items in the table.

  bool lookup(int value, istring &symbolic_name, istring &description,
          istring &file_location);
    // locates a "value", such as one of the enumerated values defined for
    // the type of asset in question.  true is returned if the value is
    // meaningful; also, "symbolic_name" is set to the item's name as used
    // inside the source code (i.e., its enum member's name), the "description"
    // is set to the full textual description of what that value means and
    // "file_location" is set to the name of the source file that defines the
    // value.

  bool lookup(const istring &symbolic_name, int &value, istring &description,
          istring &file_location);
    // similar to the above lookup, but seeks out the "symbolic_name" in order
    // to find the integer "value" and textual "description" for that item.

  int elements() const;
    // returns how many items are listed for the types of values specified.

  bool get(int index, istring &symbolic_name, int &value,
           istring &description, istring &file_location);
    // accesses the "index"th item in the list.

private:
  istring *_tag;  // the name of our section.
  system_values_lookup_list *_list;  // the values and text that we found.
  istring *_file;  // the manifest filename.

  bool open_values();  // retrieves the information from the file.
};

////////////////////////////////////////////////////////////////////////////

// these macros can be used to create program-wide instances of the
// system_values class for each of the supported types of values.  the names
// of the functions are __global_outcome_list(), __global_filter_list() and
// __global_event_list(), each of which return a reference to a set of system
// values of the appropriate type.
//
// note that the function names above can be used anywhere, but the macro
// should only be plugged in once in the body of the application (i.e.,
// do not use the macro in libraries).  this is done automatically for all
// apps using the HOOPLE_STARTUP_CODE or HOOPLE_MAIN macros.

// these macros require the headers:
//   <basis/utility.h>

#define GLOBAL_OUTCOME_VALUES \
  SAFE_STATIC(system_values, __global_outcome_list, \
      (system_values::OUTCOME_VALUES()))

#define GLOBAL_FILTER_VALUES \
  SAFE_STATIC(system_values, __global_filter_list, \
      (system_values::FILTER_VALUES()))

#define GLOBAL_EVENT_VALUES \
  SAFE_STATIC(system_values, __global_event_list, \
      (system_values::EVENT_VALUES()))

// these externs provide a reference that libraries can use, but the macro
// above must be deployed in the application.
extern system_values PWL_EXPO &__global_outcome_list();
extern system_values PWL_EXPO &__global_filter_list();
extern system_values PWL_EXPO &__global_event_list();

#endif

