/*****************************************************************************\
*                                                                             *
*  Name   : timer_driver                                                      *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2003-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "timer_driver.h"

#include <basis/log_base.h>
#include <basis/mutex.h>
#include <basis/portable.h>
#include <basis/utility.h>
#include <data_struct/amorph.cpp>
#include <data_struct/static_memory_gremlin.h>
#include <mechanisms/ithread.h>
#include <mechanisms/time_stamp.h>

#include <signal.h>
#ifdef __UNIX__
  #include <sys/time.h>
#endif
#ifdef __WIN32__
  #include <winsock.h>  // defines timeval.
#endif

//#define DEBUG_TIMER_DRIVER
  // uncomment for noisy code.

#undef LOG
#define LOG(s) CLASS_EMERGENCY_LOG(program_wide_logger(), s)

const int INITIAL_TIMER_GRANULARITY = 14;
  // the timer will support durations of this length or greater initially.
  // later durations will be computed based on the timers waiting.

const int MAX_TIMER_PREDICTION = 140;
  // this is the maximum predictive delay before we wake up again to see if
  // any new timed items have arrived.  this helps us to not wait too long
  // when something's scheduled in between snoozes.

const int PAUSE_TIME = 200;
  // we will pause this many milliseconds if the timer is already occurring
  // when we're trying to get the lock on our list.

const int LONG_TIME = 1 * HOUR_ms;
  // the hook can be postponed a really long time with this when necessary.

////////////////////////////////////////////////////////////////////////////

SAFE_STATIC(timer_driver, timer_driver::global_timer_driver, );

////////////////////////////////////////////////////////////////////////////

#ifdef __UNIX__
const int OUR_SIGNAL = SIGUSR2;

class signalling_thread : public ithread
{
public:
  signalling_thread(int initial_interval) : ithread(initial_interval) {}
  
  void perform_activity(void *ptr) {
    raise(OUR_SIGNAL);
  }

private:
};
#endif

#ifdef __UNIX__
void timer_driver_private_handler(int signal_seen)
#elif defined(__WIN32__)
void __stdcall timer_driver_private_handler(window_handle hwnd, u_int msg,
    UINT_PTR id, u_long time)
#else
  #error No timer method known for this OS.
#endif
{
  #define static_class_name() istring("timer_driver")
  #define func "timer_driver_private_handler"
  int seen = 0;
#ifdef __UNIX__
  seen = signal_seen;
  if (seen != OUR_SIGNAL) {
#elif defined(__WIN32__)
  seen = (int)id;
  if (seen != program_wide_timer().real_timer_id()) {
#else
  if (true) {  // unknown OS.
#endif
#ifdef DEBUG_TIMER_DRIVER
    LOG(isprintf("unknown signal/message %d caught.", seen));
#endif
    return;
  }
  program_wide_timer().handle_system_timer();
  #undef func
  #undef static_class_name
}

////////////////////////////////////////////////////////////////////////////

class driven_object_record
{
public:
  int _duration;  // the interval for timer hits on this object.
  timed_object *_to_invoke;  // the object that will be called back.
  time_stamp _next_hit;  // next time the timer should hit for this object.
  bool _okay_to_invoke;  // true if this object is okay to call timers on.
  bool _handling_timer;  // true if we're handling this object right now.

  driven_object_record(int duration, timed_object *to_invoke)
  : _duration(duration), _to_invoke(to_invoke), _next_hit(duration),
    _okay_to_invoke(true), _handling_timer(false) {}
};

class driven_objects_list
: public amorph<driven_object_record>, public virtual object_base
{
public:
  IMPLEMENT_CLASS_NAME("driven_objects_list");

  int find_obj(timed_object *obj) {
    for (int i = 0; i < elements(); i++) {
      if (borrow(i) && (borrow(i)->_to_invoke == obj))
        return i;
    }
    return common::NOT_FOUND;
  }
};

////////////////////////////////////////////////////////////////////////////

timer_driver::timer_driver()
: _timers(new driven_objects_list),
  _lock(new mutex),
#ifdef __UNIX__
  _prompter(new signalling_thread(INITIAL_TIMER_GRANULARITY)),
#endif
#ifdef __WIN32__
  _real_timer_id(NIL),
#endif
  _in_timer(false)
{
  hookup_OS_timer(INITIAL_TIMER_GRANULARITY);

#ifdef __UNIX__
  // register for the our personal signal.
  signal(OUR_SIGNAL, &timer_driver_private_handler);
  _prompter->start(NIL);
#endif
}

timer_driver::~timer_driver()
{
  #define func "destructor"
#ifdef __UNIX__
  _prompter->stop();

  struct sigaction action;
  action.sa_handler = SIG_DFL;
  action.sa_sigaction = NIL;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  action.sa_restorer = NIL;
  int ret = sigaction(OUR_SIGNAL, &action, NIL);
  if (ret) {
///uhhh
  }
#endif
  unhook_OS_timer();

  // make sure we aren't still in a timer handler when we reset our list.
  while (true) {
    _lock->lock();
    if (_in_timer) {
      _lock->unlock();
      LOG("waiting to acquire timer_driver lock.");
      portable::sleep_ms(PAUSE_TIME);
    } else {
      break;
    }
  }

  _timers->reset();  // clear out the registered functions.
  _lock->unlock();

  WHACK(_timers);
  WHACK(_lock);
#ifdef __UNIX__
  WHACK(_prompter);
#endif

#ifdef DEBUG_TIMER_DRIVER
  LOG("timer_driver is closing down.");
#endif
  #undef func
}

#ifdef __WIN32__
int timer_driver::real_timer_id() { return _real_timer_id; }
#endif

bool timer_driver::zap_timer(timed_object *to_remove)
{
  #define func "zap_timer"
#ifdef DEBUG_TIMER_DRIVER
  if (_in_timer) {
    LOG("hmmm: zapping timer while handling previous timer...!");
  }
#endif
  auto_synchronizer l(*_lock);
  int indy = _timers->find_obj(to_remove);
  if (negative(indy)) return false;  // unknown.
#ifdef DEBUG_TIMER_DRIVER
  LOG(isprintf("zapping timer %x.", to_remove));
#endif
  driven_object_record *reco = _timers->borrow(indy);
  reco->_okay_to_invoke = false;
  if (reco->_handling_timer) {
    // results are not guaranteed if we see this situation.
    LOG(isprintf("Logic Error: timer %x being zapped WHILE BEING HANDLED!",
        to_remove));
  }
  return true;
  #undef func
}

bool timer_driver::set_timer(int duration, timed_object *to_invoke)
{
  #define func "set_timer"
#ifdef DEBUG_TIMER_DRIVER
  if (_in_timer) {
    LOG("hmmm: setting timer while handling previous timer...!");
  }
#endif
#ifdef DEBUG_TIMER_DRIVER
  LOG(isprintf("setting timer %x to %d ms.", to_invoke, duration));
#endif
  auto_synchronizer l(*_lock);
  // find any existing record.
  int indy = _timers->find_obj(to_invoke);
  if (negative(indy)) {
    // add a new record to list.
    _timers->append(new driven_object_record(duration, to_invoke));
  } else {
    // change the existing record.
    driven_object_record *reco = _timers->borrow(indy);
    reco->_duration = duration;
    reco->_okay_to_invoke = true;  // just in case.
  }
  return true;
  #undef func
}

void timer_driver::handle_system_timer()
{
  #define func "handle_system_timer"
  if (_in_timer) {
#ifdef DEBUG_TIMER_DRIVER
    LOG("terrible error: invoked system timer while handling previous timer.");
#endif
    return;
  }
  unhook_OS_timer();

#ifdef DEBUG_TIMER_DRIVER
  LOG("into handling OS timer...");
#endif

  array<driven_object_record *> to_invoke_now;

  {
    // lock the list for a short time, just to put in a stake for the timer
    // flag; no one is allowed to change the list while this is set to true.
    auto_synchronizer l(*_lock);
    _in_timer = true;

    // zip across our list and find out which of the timer functions should be
    // invoked.
    for (int i = 0; i < _timers->elements(); i++) {
      driven_object_record *funky = _timers->borrow(i);
      if (!funky) {
        char *msg = "error in timer list logic!";
        LOG(msg);
#ifdef CATCH_ERRORS
        continuable_error(static_class_name(), func, msg);
#endif
        _timers->zap(i, i);
        i--;  // skip back over dud record.
        continue;
      }
      if (funky->_next_hit <= time_stamp()) {
        // this one needs to be jangled.
        to_invoke_now += funky;
      }
    }
  }

#ifdef DEBUG_TIMER_DRIVER
  istring pointer_dump;
  for (int i = 0; i < to_invoke_now.length(); i++) {
    driven_object_record *funky = to_invoke_now[i];
    pointer_dump += isprintf("%x ", funky->_to_invoke);
  }
  if (pointer_dump.t())
    LOG(istring("activating ") + pointer_dump);
#endif

  // now that we have a list of timer functions, let's call on them.
  for (int i = 0; i < to_invoke_now.length(); i++) {
    driven_object_record *funky = to_invoke_now[i];
    {
      auto_synchronizer l(*_lock);
      if (!funky->_okay_to_invoke) continue;  // skip this guy.
      funky->_handling_timer = true;
    }
    // call the timer function.
    funky->_to_invoke->handle_timer_callback();
    {
      auto_synchronizer l(*_lock);
      funky->_handling_timer = false;
    }
    // reset the time for the next hit.
    funky->_next_hit.reset(funky->_duration);
  }

  // compute the smallest duration before the next guy should fire.
  int next_timer_duration = MAX_TIMER_PREDICTION;
  time_stamp now;  // pick a point in time as reference for all timers.
  for (int i = 0; i < _timers->elements(); i++) {
    driven_object_record *funky = _timers->borrow(i);
    int funky_time = int(funky->_next_hit.value() - now.value());
    // we limit the granularity of timing since we don't want to be raging
    // on the CPU with too small a duration.
    if (funky_time < INITIAL_TIMER_GRANULARITY)
      funky_time = INITIAL_TIMER_GRANULARITY;
    if (funky_time < next_timer_duration)
      next_timer_duration = funky_time;
  }

  {
    // release the timer flag again and do any cleanups that are necessary.
    auto_synchronizer l(*_lock);
    _in_timer = false;
    for (int i = 0; i < _timers->elements(); i++) {
      driven_object_record *funky = _timers->borrow(i);
      if (!funky->_okay_to_invoke) {
        // clean up something that was unhooked.
        _timers->zap(i, i);
        i--;
      }
    }
  }

#ifdef DEBUG_TIMER_DRIVER
  LOG("done handling OS timer.");
#endif

  // set the next expiration time to the smallest next guy.
  reset_OS_timer(next_timer_duration);
  #undef func
}

// the following OS_timer methods do not need to lock the mutex, since they
// are not actually touching the list of timers.

void timer_driver::hookup_OS_timer(int duration)
{
  #define func "hookup_OS_timer"
  if (negative(duration)) {
#ifdef DEBUG_TIMER_DRIVER
    LOG("seeing negative duration for timer!");
#endif
    duration = 1;
  } else if (!duration) {
#ifdef DEBUG_TIMER_DRIVER
    LOG("patching zero duration for timer.");
#endif
    duration = 1;
  }
#ifdef DEBUG_TIMER_DRIVER
  LOG(isprintf("hooking next OS timer in %d ms.", duration));
#endif
#ifdef __UNIX__
  // just make our thread hit after the duration specified.
  _prompter->reschedule(duration);
#elif defined(__WIN32__)
  _real_timer_id = SetTimer(NIL, 0, duration, timer_driver_private_handler);
  if (!_real_timer_id) {
    // failure to set the timer.
    LOG("could not set the interval timer.");
    return;
  }
#endif
  #undef func
}

void timer_driver::unhook_OS_timer()
{
  #define func "unhook_OS_timer"
#ifdef __UNIX__
  // postpone the thread for quite a while so we can take care of business.
  _prompter->reschedule(LONG_TIME);
#elif defined(__WIN32__)
  if (_real_timer_id) KillTimer(NIL, _real_timer_id);
#endif
#ifdef DEBUG_TIMER_DRIVER
  LOG("unhooked OS timer.");
#endif
  #undef func
}

void timer_driver::reset_OS_timer(int next_hit)
{
  #define func "reset_OS_timer"
  unhook_OS_timer();  // stop the timer from running.
  hookup_OS_timer(next_hit);  // restart the timer with the new interval.
  #undef func
}

