#ifndef TIMER_DRIVER_CLASS
#define TIMER_DRIVER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : timer_driver                                                      *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Provides platform-independent timer support.  Multiple objects can be    *
*  hooked to the timer to be called when their interval elapses.              *
*                                                                             *
*******************************************************************************
* Copyright (c) 2003-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

// NOTE: Only one of the timer_driver objects is allowed per program.

#include "opsysdll.h"

#include <basis/object_base.h>

// forward.
class driven_objects_list;
class signalling_thread;

// the timed_object is the base class for objects that can be hooked into
// timer events.

class timed_object
{
public:
  virtual ~timed_object() {}
  virtual void handle_timer_callback() = 0;
    // this method is invoked when the timer period elapses for this object.
};

////////////////////////////////////////////////////////////////////////////

// the timer_driver is to be instantiated once within the program.  it allows
// new timed_objects to be added as needed.

#define program_wide_timer() timer_driver::global_timer_driver()
  // provides access to the singleton timer_driver.  no other timer_driver
  // objects should ever be created, since this single one will service all
  // timer needs within the program.

class OPSYSTEM_CLASS_STYLE timer_driver : public virtual object_base
{
public:
  timer_driver();
  virtual ~timer_driver();

  IMPLEMENT_CLASS_NAME("timer_driver");

  //////////////////////////////////////////////////////////////////////////

  // main methods for controlling timed_objects.

  bool set_timer(int duration, timed_object *to_invoke);
    // sets a timer to call "to_invoke" every "duration" milliseconds.  if the
    // object "to_invoke" already exists, then its duration is changed.

  bool zap_timer(timed_object *to_drop);
    // removes the timer that was established for "to_drop".  do not zap a
    // timer from its own callback!  it could cause synchronization problems.

  //////////////////////////////////////////////////////////////////////////

  // internal methods.

#ifdef __WIN32__
  int real_timer_id();
    //!< provides the timer id for comparison on windows platforms.
#endif

  void handle_system_timer();
    // this is invoked by the OS timer support.  must be called by main thread.

  static timer_driver &global_timer_driver();
    // the first time this is invoked, it creates a program-wide timer driver.

private:
  driven_objects_list *_timers;  // timer hooked objects.
  mutex *_lock;  // protects list of timers.
#ifdef __UNIX__
  signalling_thread *_prompter;  // drives our timers.
#endif
#ifdef __WIN32__
  int _real_timer_id;  // used for storing window timer handle.
#endif
  bool _in_timer;  // true if we're handling the timer right now.

  // these do the low-level system magic required to get a function hooked
  // to a timer.
  void hookup_OS_timer(int duration);
    // hooks us into the timer events at the "duration" specified.
  void reset_OS_timer(int next_hit);
    // changes the interval to "next_hit" so that only that many milliseconds
    // will elapse before the next timer hit.
  void unhook_OS_timer();
    // disconnects us from the timer events.
};

#endif

