/*****************************************************************************\
*                                                                             *
*  Name   : version_ini editing support                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1995-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "byte_filer.h"
#include "filename.h"
#include "ini_config.h"
#include "version_ini.h"

#include <basis/common_templates.h>
#include <basis/version_record.h>
#include <basis/function.h>
#include <basis/portable.h>
#include <basis/utility.h>

#include <sys/stat.h>
#ifdef __WIN32__
  #include <io.h>
#endif

const istring VERSION_SECTION = "version";
  // the section that version entries are stored under in the INI file.

// these are semi-constant fields.
const istring COMPANY_KEY="company";
const istring COPYRIGHT_KEY="copyright";
const istring LEGAL_INFO_KEY="legal_info";
const istring PRODUCT_KEY="product_name";
const istring WEB_SITE_KEY="web_site";

// not used anymore; now matched with the file version's first two digits.
//const version PRODUCT_VERSION(2, 0, 0, 0);
  // the current version of the entire product.

// these are field names in the INI file.
const istring MAJOR = "major";
const istring MINOR = "minor";
const istring REVISION = "revision";
const istring BUILD = "build";
const istring DESCRIPTION = "description";
const istring ROOT = "root";
const istring NAME = "name";
const istring EXTENSION = "extension";
const istring OLE_AUTO = "ole_auto";

// this is the default version INI file name, if no other is specified.
const istring VERSION_INI_FILE = "/version.ini";

#undef LOG
#define LOG(t) CLASS_EMERGENCY_LOG(program_wide_logger(), t)

version_ini::version_ini(const istring &path_name)
: _loaded(false),
  _path_name(new filename(path_name)),
  _ini(new ini_configurator("", ini_configurator::RETURN_ONLY)),
  _held_record(new version_record)
{
  check_name(*_path_name);
  _ini->name(*_path_name);
}

version_ini::~version_ini()
{
  WHACK(_ini);
  WHACK(_path_name);
  WHACK(_held_record);
}

bool version_ini::ole_auto_registering()
{
  istring extension = _ini->load(VERSION_SECTION, OLE_AUTO, "");
  return (extension.lower() != "");
}

bool version_ini::executable()
{
  istring extension = _ini->load(VERSION_SECTION, EXTENSION, "");
  if (extension.lower() == istring("exe")) return true;
  return false;
}

bool version_ini::library() { return !executable(); }

bool version_ini::writable() { return _path_name->is_writable(); }

void version_ini::check_name(filename &to_examine)
{
  // if it's just a directory name, add the file name.
  if (to_examine.is_directory()) {
    to_examine = istring(to_examine) + VERSION_INI_FILE;
    to_examine.canonicalize();
  }

  // add the directory explicitly (if it's not there already) or the ini
  // writer will get it from the windows directory.
  if ( (to_examine.raw()[0] != '.') && (to_examine.dirname().raw() == ".") ) {
    to_examine = istring("./") + to_examine;
    to_examine.canonicalize();
  }
}

bool version_ini::executable(const istring &path_name_in)
{
  filename path_name(path_name_in);
  check_name(path_name);
  ini_configurator temp_ini(path_name, ini_configurator::RETURN_ONLY);
  istring extension = temp_ini.load(VERSION_SECTION, EXTENSION, "");
  extension.to_lower();
  if (extension == istring("exe")) return true;
  return false;
}

bool version_ini::library(const istring &path_name)
{ return !executable(path_name); }

version version_ini::get_version()
{
  if (_loaded) return _held_record->file_version;
  get_record();
  return _held_record->file_version;
}

void version_ini::set_version(const version &to_write, bool write_ini)
{
  _held_record->file_version = to_write;  // copy the version we're given.

  // set the product version appropriately to the file version.
  _held_record->product_version = to_write;
  _held_record->product_version.set_component(version::REVISION, "0");
  _held_record->product_version.set_component(version::BUILD, "0");

  if (!write_ini) return;  // they don't want a change to the file.
  _ini->store(VERSION_SECTION, MAJOR, to_write.get_component(version::MAJOR));
  _ini->store(VERSION_SECTION, MINOR, to_write.get_component(version::MINOR));
  _ini->store(VERSION_SECTION, REVISION, to_write.get_component(version::REVISION));
  _ini->store(VERSION_SECTION, BUILD, to_write.get_component(version::BUILD));
}

version version_ini::read_version_from_ini()
{
  string_array parts;
  parts += _ini->load(VERSION_SECTION, MAJOR, "0");
  parts += _ini->load(VERSION_SECTION, MINOR, "0");
  parts += _ini->load(VERSION_SECTION, REVISION, "0");
  parts += _ini->load(VERSION_SECTION, BUILD, "0");
  return version(parts);
}

version_record &version_ini::access_record() { return *_held_record; }

version_record version_ini::get_record()
{
  if (_loaded) return *_held_record;
  version_record to_return;
  to_return.description = _ini->load(VERSION_SECTION, DESCRIPTION, "");
  to_return.file_version = read_version_from_ini();
  to_return.internal_name = _ini->load(VERSION_SECTION, NAME, "");
  to_return.original_name = _ini->load(VERSION_SECTION, ROOT, "");
  to_return.original_name += ".";

  // the dll type of extension is a hard default.  anything besides the exe
  // ending gets mapped to dll.
  istring extension = _ini->load(VERSION_SECTION, EXTENSION, "");
  extension.to_lower();
  if (extension == "dll") {}
  else if (extension == "exe") {}
  else extension = "dll";
  to_return.original_name += extension;

  to_return.product_version = to_return.file_version;
  to_return.product_version.set_component(version::REVISION, "0");
  to_return.product_version.set_component(version::BUILD, "0");

  to_return.product_name = _ini->load(VERSION_SECTION, PRODUCT_KEY, "");
  to_return.company_name = _ini->load(VERSION_SECTION, COMPANY_KEY, "");
  to_return.copyright = _ini->load(VERSION_SECTION, COPYRIGHT_KEY, "");
  to_return.trademarks = _ini->load(VERSION_SECTION, LEGAL_INFO_KEY, "");
  to_return.web_address = _ini->load(VERSION_SECTION, WEB_SITE_KEY, "");

  // fix any incoming macros from the version record.
  to_return.internal_name.replace("$product_name", software_product_name());

  *_held_record = to_return;
  _loaded = true;
  return to_return;
}

void version_ini::set_record(const version_record &to_write, bool write_ini)
{
  *_held_record = to_write;
  if (write_ini) {
    _ini->store(VERSION_SECTION, DESCRIPTION, to_write.description);
    set_version(to_write.file_version, write_ini);
    _ini->store(VERSION_SECTION, ROOT, to_write.original_name);
    _ini->store(VERSION_SECTION, NAME, to_write.internal_name);
  }
  _loaded = true;  // we consider this to be the real version now.
}

///////////////////////////////////////////////////////////////////////////

const istring version_rc_template = "\
#ifndef NO_VERSION\n#include <winver.h>\n\
  // pull in the token definitions.\n\
#define BI_PLAT_WIN32\n\
  // force 32 bit compile.\n\
1 VERSIONINFO LOADONCALL MOVEABLE\n\
FILEVERSION $file_ver\n\
PRODUCTVERSION $prod_ver\n\
FILEFLAGSMASK 0\n\
FILEFLAGS VS_FFI_FILEFLAGSMASK\n\
#if defined(BI_PLAT_WIN32)\n\
  FILEOS VOS__WINDOWS32\n\
#else\n\
  FILEOS VOS__WINDOWS16\n\
#endif\n\
FILETYPE VFT_APP\n\
BEGIN\n\
  BLOCK \"StringFileInfo\"\n\
  BEGIN\n\
    // Language type = U.S. English(0x0409) and Character Set = Windows, Multilingual(0x04e4)\n\
    BLOCK \"040904E4\"              // Matches VarFileInfo Translation hex value.\n\
    BEGIN\n\
      VALUE \"CompanyName\", \"$company\\000\"\n\
#ifndef _DEBUG\n\
      VALUE \"FileDescription\", \"$file_desc\\000\"\n\
#else\n\
      VALUE \"FileDescription\", \"$file_desc\\000\"\n\
#endif\n\
      VALUE \"FileVersion\", \"$file_txt_ver\\000\"\n\
      VALUE \"InternalName\", \"$internal\\000\"\n\
      VALUE \"LegalCopyright\", \"$copyright\\000\"\n\
      VALUE \"LegalTrademarks\", \"$legal_tm\\000\"\n\
      VALUE \"OriginalFilename\", \"$original_name\\000\"\n\
      VALUE \"ProductName\", \"$prod_name\\000\"\n\
      VALUE \"ProductVersion\", \"$prod_txt_ver\\000\"\n\
      $special_ole_flag\n\
    END\n\
  END\n\
\n\
  BLOCK \"VarFileInfo\"\n\
  BEGIN\n\
    VALUE \"Translation\", 0x0409, 0x04e4 // US English (0x0409) and win32 multilingual (0x04e4)\n\
  END\n\
END\n\
#endif\n";

///////////////////////////////////////////////////////////////////////////

//hmmm: unify this bit and the bit in value_tagger once it resides in a
//      library.  that just happened in august 2006.  time to do this...
bool version_ini::chmod_writeable(const istring &to_mod)
{
  #define func "chmod_writeable"

  filename moddy(to_mod);
  if (!moddy.exists()) return true;
    // if it doesn't exist, you should be able to create it, unless there are
    // problems with the directory rights.

  int chmod_value;
#ifdef __UNIX__
  chmod_value = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
#elif defined(__WIN32__)
  chmod_value = _S_IREAD | _S_IWRITE;
#else
  //unknown.  let's try unix...
  chmod_value = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
#endif
  int chmod_result = chmod(to_mod.s(), chmod_value);
  if (chmod_result) {
    LOG(istring("there was a problem changing permissions on ") + to_mod
        + "; writing the new version might fail.");
    return false;
  }
  return true;
  #undef func
}

// replaces every occurrence of the keyword in "tag" with the "replacement".
#define REPLACE(tag, replacement) \
  found = true; \
  while (found) { \
    found = (!new_version_entry.replace(tag, replacement)); \
  }

bool version_ini::write_rc(const version_record &to_write)
{
  istring new_version_entry(version_rc_template);

  bool found;

  // $file_ver -> w, x, y, z for version of the file.
  REPLACE("$file_ver", to_write.file_version.flex_text_form(version::COMMAS));

  // $prod_ver -> w, x, y, z for version of the product.
  REPLACE("$prod_ver", to_write.product_version.flex_text_form
      (version::COMMAS));

  // $company -> name of company.
  REPLACE("$company", to_write.company_name);

  // $file_desc -> description of file.
  istring description_release = to_write.description;
  REPLACE("$file_desc", description_release);
  istring description_debug = to_write.description
      + istring(" -- Debug Version");
  REPLACE("$file_desc", description_debug);

  // $file_txt_ver -> file version in form w.x.y.z.
  REPLACE("$file_txt_ver", to_write.file_version.flex_text_form(version::DOTS));

  // $internal -> internal name of the library, without extensions?
  REPLACE("$internal", to_write.internal_name);

  // $copyright -> copyright held by us.
  REPLACE("$copyright", to_write.copyright);

  // $legal_tm -> the legal trademarks that must be included, e.g. windows?
  REPLACE("$legal_tm", to_write.trademarks);

  // $original_name -> the file's name before possible renamings.
  REPLACE("$original_name", to_write.original_name);

  // $prod_name -> the name of the product that this belongs to.
  REPLACE("$prod_name", to_write.product_name);

  // $prod_txt_ver -> product version in form w.x.y.z.
  REPLACE("$prod_txt_ver", to_write.product_version
      .flex_text_form(version::DOTS, version::MINOR));

  istring special_filler;  // nothing by default.
  if (ole_auto_registering())
    special_filler = "VALUE \"OLESelfRegister\", \"\\0\"";
  REPLACE("$special_ole_flag", special_filler);

  istring root_part = "/";
  root_part += _ini->load(VERSION_SECTION, ROOT, "");

  istring rc_filename(istring(_path_name->dirname()) + root_part
      + "_version.rc");

  chmod_writeable(rc_filename);  // make sure we can write to the file.

  byte_filer rc_file(rc_filename, "w");
  if (!rc_file.good()) return false;
  rc_file.write((byte *)new_version_entry.s(), new_version_entry.length());
  rc_file.close();
  return true;
}

///////////////////////////////////////////////////////////////////////////

const istring version_header_template = "\
#ifndef $lib_prefix_VERSION_HEADER\n\
#define $lib_prefix_VERSION_HEADER\n\
\n\
/*****************************************************************************\\\n\
*                                                                             *\n\
*  Name   : Version header for $lib_name\n\
*  Author : Automatically generated by version_stamper                        *\n\
*                                                                             *\n\
\\*****************************************************************************/\n\
\n\
#include <basis/version_record.h>\n\
\n\
#define VERSION_FOR_$lib_prefix version($lib_version)\n\
  // this macro is expected to match the most recent version of the\n\
  // $lib_name library.\n\
\n\
#ifdef __WIN32__\n\
\n\
#include <basis/version_checker.h>\n\
\n\
// this macro can be used to check that the current version of the\n\
// $lib_name library is the same version as expected.  to use it, check\n\
// whether it returns true or not.  if not, the version is incorrect.\n\
#define CHECK_$lib_prefix() (version_checker(istring(\"$lib_prefix\")\\\n\
    + istring(\".dll\"), VERSION_FOR_$lib_prefix, \\\n\
      istring(\"Please contact $company_name for the latest DLL and \"\\\n\
        \"Executable files ($web_address).\")).good_version())\n\
\n\
#else\n\
\n\
// null checking for firmware.\n\
\n\
#define CHECK_$lib_prefix() 1\n\
\n\
#endif //__WIN32__\n\
\n\
#endif\n\
\n";

///////////////////////////////////////////////////////////////////////////

bool version_ini::write_code(const version_record &to_write)
{
  istring root = _ini->load(VERSION_SECTION, ROOT, "");
  root.to_upper();  // make upper case for naming sake.
  istring name = _ini->load(VERSION_SECTION, NAME, "");
  // replace the macros here also.
  name.replace("$product_name", software_product_name());

  istring new_version_entry(version_header_template);

  bool found;

  // $lib_prefix -> the first part of the library's name.
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);
  REPLACE("$lib_prefix", root);

  // $lib_name -> the name of the library, as it thinks of itself.
  REPLACE("$lib_name", name);
  REPLACE("$lib_name", name);
  REPLACE("$lib_name", name);

  // $lib_version -> the current version for this library.
  REPLACE("$lib_version", to_write.file_version.flex_text_form(version::COMMAS));

  // $company_name -> the company that produces the library.
  REPLACE("$company_name", to_write.company_name);

  // $web_address -> the web site for the company.  not actually stored.
  REPLACE("$web_address", to_write.web_address);

  istring header_filename(_path_name->dirname().raw() + istring("/version.h"));

  chmod_writeable(header_filename);  // make sure we can write to the file.

  byte_filer header(header_filename, "w");
  if (!header.good()) return false;
  header.write((byte *)new_version_entry.s(), new_version_entry.length());
  header.close();
  return true;
}

