#ifndef VIRTUAL_LIBRARY_CLASS
#define VIRTUAL_LIBRARY_CLASS

// only for ms-windows currently.

/*****************************************************************************\
*                                                                             *
*  Name   : virtual_library                                                   *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Supports definition of a dynamic library that must provide a set of      *
*  "virtual" functions.  Here, the functions are mainly for loading the DLL   *
*  and ensuring that it receives the expected parameters.  Derived classes    *
*  may add more required functionality as needed but should not deviate from  *
*  the creation support defined here.                                         *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/definitions.h>

namespace library_plugins {

  class vl_data_pack { public: virtual ~vl_data_pack() {} };
    // helps to make construction parameter passing more generic.

  class virtual_library
  { 
  public:
    virtual ~virtual_library() {}  // force virtual destructor.

    virtual virtual_library *create_instance(vl_data_pack &parameters) = 0;
      // this is a factory method that is required by all virtual libraries.
      // when the virtual_library is loaded, this function will be sought and
      // called.  the implementation must construct a new instance of the most
      // derived type of virtual_library that it knows about.  the library's
      // implementation should ensure that the "parameters" are of the expected
      // type.  any failure should cause NIL to be returned.
  };

  // this macro must be used in one and only one file in the virtual library's
  // source code.  it installs the instance creator into static scope; this is
  // needed for our lookup process.  the use of the macro is quite simple; if
  // the root class of your virtual library is called SpongeCake, then the
  // macro should be:
  //     DEFINE_VIRTUAL_LIBRARY(SpongeCake);
  // the SpongeCake class must provide the virtual create_instance() function
  // above in order to be a proper virtual_library.  it must also have a blank
  // constructor in order to work with this macro.
  #define DEFINE_VIRTUAL_LIBRARY(library_class_name) \
    extern "C" { \
      virtual_library HOOPLE_DLL_EXPORT_FUNCTION *static_create_instance \
          (vl_data_pack &parameters) \
      { library_class_name blank; return blank.create_instance(parameters); } \
    }

  // string form of the static method defined above.  this is provided so that
  // we don't need to hard-code the name all over the place.
  #define STATIC_CREATE_INSTANCE_NAME "static_create_instance"

} // namespace.

#endif // outer guard

