#ifndef SCHEDULABLE_CLASS
#define SCHEDULABLE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : schedulable                                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    The schedulable object is a base class for objects that can be added     *
*  into a scheduler for processing.  It requires a description of the period  *
*  for/ scheduling (if the object is to be repeatedly scheduled) and it needs *
*  to know the identity (id) of the schedule_actor that will process the      *
*  object during its scheduling runs.                                         *
*    Note: this is a heavy-weight header; it should only be included inside   *
*  implementation files (*.cpp).                                              *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "item_statistics.h"
#include "periodicity.h"

#include <data_struct/unique_id.h>
#include <mechanisms/heartbeat.h>
#include <nodes/safe_node.h>

// the unique scheduling id assigned to items in the schedule.  it must be
// unique within the constraints that you choose to implement, but one id 'X'
// must always refer to the same real object that holds schedule data for 'X'.

class scheduling_id : public unique_int
{
public:
  inline scheduling_id(int initial) : unique_int(initial) {}
};

////////////////////////////////////////////////////////////////////////////

class SCHEDULING_CLASS_STYLE schedulable
: public nodes::safe_node,
  public periodicity,
  public virtual object_base
{
public:
  schedulable(const scheduling_id &item_id, int actor_id,
          const periodicity &when_to_schedule);
    // constructs a schedulable item whose id is "item_id" that will be managed
    // by the actor with "actor_id" and that will be subject to the periodicity
    // in "when_to_schedule".

  virtual ~schedulable();

  IMPLEMENT_CLASS_NAME("schedulable");

  const scheduling_id &item_id() const { return _sched_id; }
    // the scheduling id for this schedulable item.

  int actor_id() const { return _actor_id; }
    // returns the parent that processes this type of schedule item.

  virtual int weight() const { return 1; }
    // provides an estimate of the size required for this schedule item.
    // the implementor should only need to consider those bytes that they
    // store within their derived schedulable objects.  if one doesn't override
    // this, then the schedule's maximum weight parameter is ineffectual.

//deprecate this.  it's really the object base text form.
  virtual istring schedulable_text_form() const;
    // returns a textual description of the data in this object.

  // returns the state of the countdown to the item's expiration.
  const heartbeat &expiration() const { return _expiration_timer; }
  heartbeat &expiration() { return _expiration_timer; }

  // returns the record of this item's behavior.
  const item_statistics &stats() const { return _stats; }
  item_statistics &stats() { return _stats; }

  // used by the scheduler to track the status of this item.
  bool preprocessed() const { return _preprocessed; }
  void preprocessed(bool state) { _preprocessed = state; }

private:
  int _actor_id;  // our parent's id.
  scheduling_id _sched_id;  // the id of this item in the schedule.
  item_statistics _stats;  // record of item's behavior.
  heartbeat _expiration_timer;
    // controls expiration of scheduled items.  when the heartbeat is due, the
    // item expires and the expiration_hook is called.  at that time, the timer
    // can either be reset completely (indicating that the item should have no
    // strikes against it), allowed to slack until next expiration (indicating
    // that one of the strikes is used up) or actually expired (indicating that
    // the item falls out of scheduling permanently).
  bool _preprocessed;  // true if this item has been pre-processed.
};

#endif

