#ifndef SCHEDULER_CLASS
#define SCHEDULER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : scheduler                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Processes a list of "schedulable" items by keeping track of the current  *
*  item to schedule and activating that item when appropriate.                *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "scheduling_common.h"

// forward.
class actor_mapper;
class int_averager;
class mailbox;
class memory_limiter;
class schedulable;
class schedule_actor;
class schedule_driver;
class schedule_iterator;
class scheduling_id;

namespace nodes {
  class safe_list;
  class safe_list_write_iterator;
}

class SCHEDULING_CLASS_STYLE scheduler : public scheduling_common
{
public:
  scheduler(actor_mapper &mapper, int time_per_activation,
          int maximum_weight, int max_actor_weight,
          int snooze_interval = 64);
    // the "mapper" provides the actors that operate on scheduled items.  if
    // there are any synchronization issues related to the schedule_actors,
    // then the mapper must deal with them.  the "time_per_activation" is
    // the amount of time any one item's activation is allowed to last.  the
    // "maximum_weight" is the largest amount of space the schedule is allowed
    // to consume.  if it's zero, then there is no overall size limit.  the
    // "max_actor_weight" is a similar limit on the largest any one
    // schedule_actor's queue is allowed to get.  if it is zero, then no limit
    // is enforced per schedule_actor.  the "snooze_interval" is the interval
    // between runs of the scheduler; this is the fastest that items can be
    // processed once the queue goes empty.

  virtual ~scheduler();

  IMPLEMENT_CLASS_NAME("scheduler");

  // the following are metrics that describe how this scheduler is operating.

  const int_averager &processing_durations() const;
    // the average times taken in process_schedule.  this includes all of the 
    // phases of scheduling (preprocessing, activation, postprocessing,
    // and destruction) plus any administrative overhead incurred by the
    // implementation.
  const int_averager &activation_durations() const;
    // the average times taken in activating the item.  this is just the time
    // spent within our schedulable object's virtual processing? function.
  const int_averager &item_durations() const;
    // the average times taken for a complete scheduling of items, from
    // queue entry to queue exit.
  const int_averager &bandwidths() const;
    // averages the speed of transmission over the items once they have
    // completed.  this is their weight (bytes consumed) divided by their
    // overall duration in the schedule, measured in kilobytes per second.

  int snooze_interval() const { return _snooze_interval; }
    // the pause time between runs, as specified in constructor.

  void reset_stats();
    // clears all of the current duration and speed measurements from the
    // scheduler.  any existing schedulables will still retain their own
    // metrics, but the main scheduler reports will be dumped.

  //////////////////////////////////////////////////////////////////////////

  // the main service routines below (until private:) are all thread safe.

  outcome add(schedulable *to_schedule, bool single_file);
    // places an item "to_schedule" in the scheduling queue.  the item's
    // actor will be invoked to pass the item through various phases of its
    // scheduling.  note that once the schedulable is given to the scheduler,
    // its destruction will be managed by the scheduler; do not destroy it
    // yourself.  NO_SPACE will be returned if the schedule has hit its upper
    // limit.  in cases where "to_schedule" couldn't be added, it is whacked;
    // thus there is _never_ any reason for a caller to touch "to_schedule"
    // after invoking this method.  the "single_file" parameter indicates
    // whether items scheduled for the same actor should be paused until any
    // others for that actor have completed.

  //////////////////////////////////////////////////////////////////////////

  // the background thread is managed by these two functions.
  void start_scheduling();
  void stop_scheduling();

  //////////////////////////////////////////////////////////////////////////

  // NOTE: neither type of remove operation calls any of the removed items'
  // postprocessing function.

  outcome remove_by_id(scheduling_id id);
    // removes an item in the schedule that has the "id" specified.  NOT_FOUND
    // will be returned if that id cannot be found.  this is not safe to call
    // during execution of the schedule_actor's callbacks; use mark_dead()
    // instead.

  outcome remove_by_actor(int actor_id);
    // removes all items in the schedule associated with the actor that has the
    // "actor_id".  this is commonly used to clear any items associated with
    // an actor that is about to be destroyed.

  outcome mark_dead(scheduling_id to_mark);
    // marks the item specified by "to_mark" as a dead item.  on the next
    // scheduling go-round, it will be removed.  this is safe to call even when
    // the item is active.  the item will still have its postprocess method
    // invoked when it leaves the schedule.

  void filter_missing_actors(const int_set &valid_actors);
    // based on the set of "valid_actors", any items whose actor is not listed
    // get dumped.

  //////////////////////////////////////////////////////////////////////////

  void schedule_text_form(string_array &info, int indent = 0);
    // returns a line of information for each object in the schedule.  the
    // "indent" value is the number of spaces prefixed to each entry.

  void pending_text_form(istring &info, int indent = 0);
    // prints the single-file items that are still pending into "info", with
    // a blank line after each entry.  the "info" string is not reset; the
    // information is appeneded to it.

  void memory_text_form(istring &info, int indent = 2);
    // drops all information about schedule memory usage into "info".

  //////////////////////////////////////////////////////////////////////////

  schedule_iterator *open_iterator(common::list_positions where = common::HEAD);
    // provides an iterator with write access on the schedule.  this must be
    // shut down using close_schedule_iterator().

  schedule_iterator *find(scheduling_id to_find);
    // returns an iterator pointing at the item specified by "to_find", if it
    // exists in the schedule.  if not, NIL is returned.

  void close_iterator(schedule_iterator * &to_close);
    // eliminates an iterator and frees the schedule for use again.  the
    // pointer holding the iterator will be reset to nil.

  const schedulable *active_item(schedule_iterator &to_use) const;
    // returns a pointer to the schedulable that is currently active, if one
    // is.  this is a read-only view of that item.  note that one must have
    // already opened an iterator "to_use" before calling this function.

  bool items_scheduled(schedule_iterator &to_use, int actor);
    // returns true if any scheduled items could be found for "actor" in the
    // current schedule.  note that this might include the active_item().
    // also, the iterator "to_use" will be in a random place after this call.

  void zap(schedule_iterator &to_use);
    // eliminates the schedulable at the position specified by "to_use".
    // the postprocess function is not called.

private:
  actor_mapper &_mapper;  // our id to actor translator.
  int _time_per_item;  // length of time to schedule any one activation.
  nodes::safe_list *_schedule;  // the list of items to schedule.
  int_averager *_processing_durations;  // times taken by scheduling process.
  int_averager *_activation_durations;  // times taken by activation itself.
  int_averager *_item_durations;  // times taken between item entry and exit.
  int_averager *_bandwidths;  // average speed of transmission.
  schedulable *_active_item;  // the current item being scheduled.
  memory_limiter *_weight_tracker;  // manages the weight of the schedule.
  mailbox *_pending_sends;  // the queues of paused items held per actor.
  int_set *_timely_to_run;  // actors with items whose time is ripe.
  int_set *_queued_actors;  // actors with pending items.
  int _snooze_interval;  // snoozing time between runs.

  friend class schedule_driver;  // allowed to touch our privates.
  schedule_driver *_driver;
    // drives the schedule processing.  must come after all other data members.

  outcome process_schedule(int schedulings_allowed);
    // cranks through the schedule looking for items to schedule.  if the
    // entire list is checked and none can be found to activate, then the
    // function returns.  otherwise, items will be processed until the
    // "schedulings_allowed" limit is reached.  note that this might involve
    // hitting the same scheduled item multiple times before its finished--each
    // activation counts as a scheduling hit.  note that when an item is being
    // scheduled, it is temporarily removed from the queue and its identifier
    // is reported in now_active(), which will only be valid if you currently
    // have a lock on the schedule.
    // NOTE: only one thread should call this function.  it is not safe to
    // have multiple threads driving the scheduling.
    // NOTE: start_scheduling() can be used to start a thread that will
    // automatically call this method.  then no external driver is needed.

  bool time_to_schedule(schedulable &to_check);
    // returns true if the item "to_check" is ready for scheduling.

  outcome check_expiration(schedulable &to_check, schedule_actor &actor);
    // returns CANCEL if the item "to_check" expires and is not renewed by
    // its parent.  otherwise returns OKAY.

  outcome do_preprocess(schedulable &to_preprocess);
    // checks with the actor for "to_preprocess" to make sure the item should
    // still be scheduled.  OKAY is returned if the item's okay.  any other
    // outcome should cause cancellation of the item?

  void promote_queued_items();
    // moves any pending items into the main schedule if the conditions
    // are appropriate (i.e., no other items for the same actor still running).

  void promote_actor(int actor_id);
    // promotes the queued items for the "actor_id" only.

  bool find_timely_items(int_set &timely);
    // locates items in the schedule that are ready to be scheduled.  if there
    // are no items left, false is returned.

  outcome schedule_item(int &hits, schedulable &to_process);
    // take the item "to_process" and runs it through the stages of
    // preprocessing, activation and postprocessing.  if the outcome is
    // schedule_actor::FINISHED or schedule_actor::CANCEL, then the item has
    // had it and must be removed.  the "hits" is updated if we actually did
    // some scheduling.  if schedule_actor::POSTPONE is returned, then the
    // item should not be scheduled any further at this time.

  void crunch_on_timely_items(int &hits, int hits_allowed, int_set &timely);
    // schedules any items that are members of the "timely" set.  the set is
    // updated to remove items that are no longer ready.

  // forbidden.
  scheduler(const scheduler &to_copy);
  scheduler &operator =(const scheduler &to_copy);
};

////////////////////////////////////////////////////////////////////////////

// the schedule_iterator allows the schedule to be traversed externally.

class SCHEDULING_CLASS_STYLE schedule_iterator
{
public:
  ~schedule_iterator();
    // closes down iterator (and removes writer on list).
    // NOTE: DO NOT CALL THIS YOURSELF.  it must be public for visibility, but
    // it is not intended for normal users of the scheduler object.

  void next();  // go to next item.
  void previous();  // go to previous item.

  schedulable *access();  // access the current element.

  bool is_head();  // is schedule_iterator at the head?
  bool is_tail();  // is schedule_iterator at the tail?

  void jump_head();  // set the schedule_iterator to the head.
  void jump_tail();  // set the schedule_iterator to the tail.

private:
  friend class scheduler;
  const scheduler &_manager;  // our parent.
  nodes::safe_list_write_iterator *_hidden_iterator;  // does the real work.

  schedule_iterator(const scheduler &mgr);
    // opens an iterator on the list (using write access).

  // disallowed functions.
  schedule_iterator(const schedule_iterator &to_copy);
  schedule_iterator &operator =(const schedule_iterator &to_copy);
};

#endif

