#ifndef FILE_TRANSFER_TENTACLE_CLASS
#define FILE_TRANSFER_TENTACLE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : file_transfer_tentacle                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Manages the transferrence of directory trees from one place to another.  *
*                                                                             *
*  Note: this is a fairly heavy-weight header due to the inclusion of the     *
*  file transfer infoton header.  It is better to forward declare the         *
*  objects in both file transfer headers when using the types in other        *
*  headers.                                                                   *
*                                                                             *
*******************************************************************************
* Copyright (c) 2005-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "file_transfer_infoton.h"

#include <octopus/tentacle_helper.h>

// forward.
class directory_tree;
class file_transfer_cleaner;
class file_transfer_status;
class filename_list;
class mutex;
class time_stamp;

class TENTACLES_CLASS_STYLE file_transfer_tentacle
: public tentacle_helper<file_transfer_infoton>
{
public:
  file_transfer_tentacle(bool only_provide_comparisons = false);
    // if "only_provide_comparisons" is true, then no requests for transferring
    // files will be honored.  a requesting entity can find out which files
    // are different or missing,  but not get copies of them.

  virtual ~file_transfer_tentacle();

  IMPLEMENT_CLASS_NAME("file_transfer_tentacle");

  istring text_form() const;
    // returns a string representing the current state of transfers.

  // provides a view of the directory_tree's current state.  be careful!
  // and the unlock call MUST be called when done looking at the tree.
  directory_tree *lock_directory(const istring &source_mapping);
  void unlock_directory();

  // these methods are for the "server" side--the side that has files to offer.

  outcome add_correspondence(const istring &source_mapping,
          const istring &source_root, int refresh_interval);
    // adds a file transfer correspondence.  this is a "source_mapping" which
    // is a short string that is made available to the other side for transfer
    // requests.  when they specify the "source_mapping", it will be translated
    // on this side to the "source_root", which must be a valid filesystem
    // path.  the "refresh_interval" dictates how frequently, in milliseconds,
    // the source will be scanned to update the internal directory tree.  this
    // is done the first time the "source_mapping" is set up also.  if a
    // previous identical "source_mapping" existed, then it is removed and
    // replaced with the information from the new invocation.

  outcome remove_correspondence(const istring &source_mapping);
    // takes out the "source_mapping" which was previously added.  this keeps
    // any transfers from occurring on that name, and will cause aborted
    // transfers if any were still ongoing.

  outcome refresh_now(const istring &source_mapping);
    // refreshes the "source_mapping" right now, regardless of the interval.
    // the mapping must already have been created with add_correspondence().

  bool add_path(const istring &source_mapping, const istring &new_path);
    // inserts the "new_path" into a registered correspondence.  the "source_
    // mapping" must already be registered.

  bool remove_path(const istring &source_mapping, const istring &old_path);
    // deletes the "old_path" out of an existing correspondence.

  // these methods are for the client side--the side that wants to get files.

  outcome register_file_transfer(const octopus_entity &ent,
          const istring &src_root, const istring &dest_root,
          const string_array &include);
    // records a transfer that is going to commence.  the side that wishes
    // to download files must invoke this before starting the transfer.
    // if the "include" list is not empty, then only those files will be
    // transferred.  they have to match the suffix of the files that would
    // have been transferred and wildcards are not currently supported.

  outcome cancel_file_transfer(const octopus_entity &ent,
          const istring &src_root, const istring &dest_root);
    // tosses a previously registered file transfer.  this will be done
    // automatically after a time-out period, but it is better to clean it up
    // as soon as one is finished with the transfer.

  bool status(const octopus_entity &ent, const istring &src,
        const istring &dest, int &total_size, int &total_files,
        int &current_size, int &current_files, bool &done,
        time_stamp &last_active);
    // locates the transfer designated by the "ent", "src" and "dest"
    // parameters and returns the current progress.  files refers to how many
    // files are being transferred and size refers to their combined weight
    // in bytes.  the "done" flag is set if the transfer seems finished.
    // note that this will not set any values until the first reply comes back
    // from the server.

  bool get_differences(const octopus_entity &ent, const istring &src,
        const istring &dest, filename_list &diffs);
    // accesses the list of difference for an ongoing transfer and stores them
    // in "diffs".

  // required tentacle methods...

  virtual outcome reconstitute(const string_array &classifier,
          byte_array &packed_form, infoton * &reformed);
    // recreates a "reformed" infoton from the "classifier" and packed infoton
    // data in "packed_form".  this will only succeed if the classifier's
    // first name is understood here.

  virtual outcome consume(infoton &to_chow, const octopus_request_id &item_id,
          byte_array &transformed);
    // processes the "to_chow" infoton as a file transfer request.

  virtual void expunge(const octopus_entity &to_remove);
    // throws out any transfers occurring for the entity "to_remove".

  // internal use only.

  void periodic_actions();  // drops timed out transfers.

private:
  file_transfer_status *_transfers;  // our record of ongoing transfers.
  file_transfer_status *_correspondences;  // the allowed synonyms for mapping.
  mutex *_lock;  // protects our lists.
  file_transfer_cleaner *_cleaner;  // cleans up dead transfers.
  bool _only_comparisons;  // true if we are only supporting tree compare.

  // these process the request and response infotons that are passed to us.
  outcome handle_tree_compare_request(file_transfer_infoton &req,
          const octopus_request_id &item_id);
  outcome handle_tree_compare_response(file_transfer_infoton &req,
          const octopus_request_id &item_id);
  outcome handle_storage_request(file_transfer_infoton &req,
          const octopus_request_id &item_id);
  outcome handle_storage_response(file_transfer_infoton &req,
          const octopus_request_id &item_id);
};

#endif

