#ifndef STRING_MANIPULATION_CLASS
#define STRING_MANIPULATION_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : string_manipulation                                               *
*  Author : Chris Koeritz                                                     *
*  Author : Brit Minor                                                        *
*  Author : Aaron Buchanan                                                    *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "textual_dll.h"

//! Provides various functions for massaging strings.

class TEXTUAL_CLASS_STYLE string_manipulation
{
public:
  static istring make_random_name(int min = 1, int max = 64);
    //!< creates a random name, where the letters are between 'a' and 'z'.
    /*!< the underscore will also be used occasionally.  the size is random,
    but the minimum size is "min" while the maximum is "max". */

  static bool quote_string(const istring &to_modify, istring &quoted_string);
    //!< Quotes the string "to_modify".
    /*!< Uses single quotes if the string contains double quotes, or uses
    double quotes if the string contains single quotes or no quotes, and
    doesn't even attempt to quote a string that contains both single and
    double quotes.  (by Brit Minor) */

  static istring long_line(char line_item = '/', int repeat = 76);
    //!< produces a long line of "line_item" characters.
    /*!< returns a string of text that is somewhat long compared to an 80 column
    output window and which consists of a single character repeated.  the
    character used and the repeat count are both variable. */

  static istring indentation(int spaces);
    //!< Returns a string made of white space that is "spaces" long.

  static istring &escape_chars(istring &input_string);
    //!< escapes potentially troublesome characters.
    /*!< looks for all instances of backslash(\) and quote(") characters in
    the string and places an additional backslash character in front
    of the existing character.  returns a reference to the passed string. */
  static istring &unescape_chars(istring &input_string);
    //!< reverses escape_chars().
    /*!< looks for all instances of double backslash(\\) or backslash quote(\")
    and removes the extra backslash. */

  static bool substring(const istring &look_in, const istring &after, 
                        const istring &before, istring &found);
    //!< look for the string "after" in "look_in".
    /*!< if found then look for the string "before" starting at the first
    location after "after".  if found then fill found with all characters
    after "after" and before "before".  return false if before or after are
    not found. */

  static void carriage_returns_to_spaces(istring &to_strip);
    //!< converts carriage returns in "to_strip" into spaces.
    /*!< processes the string "to_strip" by replacing all single carriage
    returns with spaces and by turning two or more carriage returns into a
    single CR plus spaces. */

  static void split_lines(const istring &input, istring &output,
          int min_column = 0, int max_column = 79);
    //!< formats blocks of text for a maximum width.
    /*!< processes the "input" text by splitting any lines that are longer
    than the "max_column".  the "min_column" is used to specify how much
    indentation should be included. */

  //////////////////////////////////////////////////////////////////////////

  // numerical manipulation functions:

  static byte char_to_hex(char to_convert);
    //!< Converts a single character into the corresponding hex nibble.
    /*!< If the character is not valid, an arbitrary value is returned. */
  static char hex_to_char(byte to_convert);
    //!< Converts a byte between 0 and 15 into a corresponding hexadecimal character.
  
  static byte_array string_to_hex(const istring &character_form);
    //!< Turns a string form of a set of hex numbers into an array of bytes.
    /*!< This functions takes a string in "character_form" and returns an array
    of bytes that is half as long and which contains the hexadecimal
    interpretation of the string.  This is currently geared to even length
    strings... */
  static istring hex_to_string(const byte_array &byte_form);
    //!< The inverse of string_to_hex prints "byte_form" as text.
    /*!< This function takes an array of bytes and converts them into their
    equivalent hexadecimal character representation. */
  
  //////////////////////////////////////////////////////////////////////////

private:
  static const istring_object &splitter_finding_set();
};

#endif

